!include "MUI.nsh"

!define PRODUCT_NAME "Virtual Volumes"
!define PRODUCT_VER_MAJ "0"
!define PRODUCT_VER_MIN "5"
!define PRODUCT_VER "${PRODUCT_VER_MAJ}.${PRODUCT_VER_MIN}"


Name "${PRODUCT_NAME} ${PRODUCT_VER}"
OutFile "vv-${PRODUCT_VER}-install.exe"

InstallDir "$PROGRAMFILES\Virtual Volumes"

VIProductVersion "${PRODUCT_VER}.0.0"
VIAddVersionKey /LANG=${LANG_ENGLISH} "ProductName" "${PRODUCT_NAME}"
;VIAddVersionKey /LANG=${LANG_ENGLISH} "Comments" ""
VIAddVersionKey /LANG=${LANG_ENGLISH} "CompanyName" "Chrysocome"
;VIAddVersionKey /LANG=${LANG_ENGLISH} "LegalTrademarks" ""
VIAddVersionKey /LANG=${LANG_ENGLISH} "LegalCopyright" "John Newbigin. GPL"
VIAddVersionKey /LANG=${LANG_ENGLISH} "FileDescription" "${PRODUCT_NAME} Installer"
VIAddVersionKey /LANG=${LANG_ENGLISH} "FileVersion" "${PRODUCT_VER}.0.0"

!insertmacro MUI_PAGE_WELCOME
!insertmacro MUI_PAGE_LICENSE "..\doc\Copying.txt"
!insertmacro MUI_PAGE_COMPONENTS
!insertmacro MUI_PAGE_DIRECTORY
!insertmacro MUI_PAGE_INSTFILES
!insertmacro MUI_PAGE_FINISH

!insertmacro MUI_UNPAGE_CONFIRM
!insertmacro MUI_UNPAGE_INSTFILES
!insertmacro MUI_UNPAGE_FINISH

!insertmacro MUI_LANGUAGE "English"




Section "Main Program Files" SecMain
	SetOutPath $INSTDIR
	File "..\bin\vv_cmd.exe"
	File "..\bin\vv.exe"
	File "..\bin\libext2fs.dll"
	File "..\bin\librfs.dll"
	File "..\bin\libfat.dll"
	File "init.vv"
	File "..\doc\readme.txt"

	WriteUninstaller "$INSTDIR\Uninstall.exe"

	WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}" "DisplayName"	"${PRODUCT_NAME}"
	WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}" "UninstallString" "$INSTDIR\Uninstall.exe"
	WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}" "Publisher" 	"Chrysocome"
	WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}" "DisplayVersion" "${PRODUCT_VER}"
	WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}" "VersionMajor" 	"${PRODUCT_VER_MAJ}"
	WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}" "VersionMinor" 	"${PRODUCT_VER_MIN}"
	WriteRegDWORD HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}" "NoModify" 	1
	WriteRegDWORD HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}" "NoRepair" 	1

SectionEnd

Section "Windows 95 Disk Access" SecWin95
	SetOutPath $INSTDIR
	File "..\bin\diskio2.dll"
SectionEnd

Section "Start Menu" StartMenu
	CreateDirectory "$SMPROGRAMS\${PRODUCT_NAME}"
	SetOutPath $INSTDIR
	CreateShortCut "$SMPROGRAMS\${PRODUCT_NAME}\${PRODUCT_NAME}.lnk" "$INSTDIR\vv.exe"
SectionEnd

Section "Configure Firewall" SecFirewall
	; no files in this section
SectionEnd

Section -post
	
  SectionGetFlags ${SecFirewall} $R0
  IntOp $R0 $R0 & 1
  IntCmp $R0 1 "" nofirewall nofirewall

  DetailPrint "Configuring Firewall"
  nsExec::ExecToLog 'netsh firewall add allowedprogram program = "$INSTDIR\vv_cmd.exe" name = "${PRODUCT_NAME}" mode = ENABLE'

 nofirewall:

SectionEnd

Section "Uninstall"
	Delete "$INSTDIR\Uninstall.exe"
	Delete "$INSTDIR\readme.txt"
	Delete "$INSTDIR\init.vv"
	Delete "$INSTDIR\libext2fs.dll"
	Delete "$INSTDIR\librfs.dll"
	Delete "$INSTDIR\libfat.dll"
	Delete "$INSTDIR\vv_cmd.exe"
	Delete "$INSTDIR\vv.exe"
	Delete "$INSTDIR\diskio2.dll"
	RMDir "$INSTDIR"

	Delete "$SMPROGRAMS\${PRODUCT_NAME}\${PRODUCT_NAME}.lnk"
	RMDir "$SMPROGRAMS\${PRODUCT_NAME}"

	DeleteRegKey HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}"

	nsExec::ExecToLog 'netsh firewall delete allowedprogram program = "$INSTDIR\vv_cmd.exe"'

SectionEnd

  LangString DESC_SecMain      ${LANG_ENGLISH} "Install Main Program Files."
  LangString DESC_SecFirewall  ${LANG_ENGLISH} "Add an exception to the windows firewall configuration."
  LangString DESC_SecWin95     ${LANG_ENGLISH} "Install library for disk access on Windows 95/98/ME."
  LangString DESC_SecStartMenu ${LANG_ENGLISH} "Add virtual volumes to the start menu."

!insertmacro MUI_FUNCTION_DESCRIPTION_BEGIN
  !insertmacro MUI_DESCRIPTION_TEXT ${SecMain} $(DESC_SecMain)
  !insertmacro MUI_DESCRIPTION_TEXT ${SecFirewall} $(DESC_SecFirewall)
  !insertmacro MUI_DESCRIPTION_TEXT ${SecWin95} $(DESC_SecWin95)
  !insertmacro MUI_DESCRIPTION_TEXT ${StartMenu} $(DESC_SecStartMenu)
!insertmacro MUI_FUNCTION_DESCRIPTION_END

Function .onInit
start:
  ReadRegStr $R0 HKLM \
  "Software\Microsoft\Windows\CurrentVersion\Uninstall\${PRODUCT_NAME}" \
  "UninstallString"
  StrCmp $R0 "" done
 
  MessageBox MB_OKCANCEL|MB_ICONEXCLAMATION \
  "${PRODUCT_NAME} is already installed. $\n$\nClick `OK` to remove the \
  previous version or `Cancel` to cancel this upgrade." \
  IDOK uninst
  Abort
  
;Run the uninstaller
uninst:
  ClearErrors
  ExecWait '$R0 _?=$INSTDIR' ;Do not copy the uninstaller to a temp file
 
  IfErrors no_remove_uninstaller
    ;You can either use Delete /REBOOTOK in the uninstaller or add some code
    ;here to remove the uninstaller. Use a registry key to check
    ;whether the user has chosen to uninstall. If you are using an uninstaller
    ;components page, make sure all sections are uninstalled.
  no_remove_uninstaller:

  goto start
  
done:
 
FunctionEnd

