unit SevenZipVCL;
(*
   This Unit is under Mozilla Public Licence
    (
     - You can use this Unit for free in free, share and commercial application.
     - Mark clearly in your Readme or Help file that you use this unit/VCL with a link the 
       SevenZipVCL Homepage (http://www.rg-software.de)
     - Any changes of the source must be publised (Just send it to me :-) SevenZipVCL@rg-software.de)
    )


   TsevenZip by Ivo Andonov
   TSevenZipVCL by Rainer Geigenberger (-> http://www.rg-software.de - SevenZipVCL@rg-software.de)

   Thanks to:
    - Marko Kamin
    - Craig Peterson

   Dll Usage:
        For develpoment put the dll into the Windows directory
        Running the Application stand alone you can put the dll into the Application directory	
		
 History:

   Version 0.1
    - Start

   Version 0.2
    - Changed everything to Widestring
    - Added WideStringList_
    - Added TWideStringArray
    - Adding RelativePath works
    - Adding works with Recursive Directory
    - Adding with Wildcards (only *.txt or something)
    - Progress for files (With SetCompleted VCL interaction works)

   Version 0.3
    - Added time reading and writing

   Version 0.4
    - Added Extract
    - Added Test via Extract(True)
    - Set Filetime during extraction
    - Progress during extraction works (small files will not be displayed alone)

   Version 0.5
    - Changes here and there
    - Clean up code

   Version 0.6
    - Changed "Extract all files"
    - Fixed bug in handling directories during extract
    - Attribute to extracted files works now
    - Fixed LastwriteTime during adding
    - Changed some functions

   Version 0.6.1
    - Some minor changes
    - Fixed some bugs
    - Added MaxProgrees to selected files
    - Added some missing GUID - not used now :-)

   Version 0.6.2
    - Compression strength could be set
    - Drive letter to Storepath option included
    - Changes made by Marko Kamin

   Version 0.6.3
    - Changed Archive options to new type Addopts
    - Solid settings works now

   Version 0.6.4
    - Implemented SFX creation

   Version 0.6.4b
    - LZMAStrength added

   Version 0.6.5
    - PPMD method added

   Version 0.6.5a
    - Fixed bug during listing of 0 byte files
   
   Version 0.6.5b
    - Fixed bug during creating files for extract/SFXarchive in Tstreamwriter 

   Version 0.6.5c
    - Fixed bug Creating SFX

   Version 0.6.6
    - Adding and extracting can be canceled

   Version 0.6.6a
    - Adding Extract without path

   Version 0.6.6c
    - Reading SFX
    - Number of files
    - New: IsSFX and SXFOffset
    - New: Function ConvertSFXto7z
    - New: Function Convert7ztoSFX

   Version 0.6.7
    - Added some Widestring function form TNTWare TNT Controls http://www.tntware.com/
    - New: OnExtractOverwrite - Do not work with Messageboxes right now
    - Include Extractoverwrite in Extractoption


*)

(*
   Known Issues / Things ToDo:
   
    - No archive properties during listing
    - With Solid archives filenames and progress during extract comes very late (at the end)
    - No multivolume support
    - No Password support
    - No deleting files from archive

  Please mark all changes with your sign and date e.g. rg 01.01.06
  and send it to me SevenZipVCL@rg-software.de
*)

interface

uses
  windows, SysUtils, Classes, FileCtrl, ActiveX, ComObj;


const
//7z internal consts

//Extract
  //NAskMode
  kExtract = 0;
  kTest    = 1;
  kSkip    = 2;

  //NOperationResult
  kOK                = 0;
  kUnSupportedMethod = 1;
  kDataError         = 2;
  kCRCError          = 3;


  FNAME_MAX32 = 512;

// SevenZIP onMessage Errorcode
  FNoError           = 0;
  FFileNotFound      = 1;
  FDataError         = 2;
  FCRCError          = 3;
  FUnsupportedMethod = 4;
  FJustAnError       = 5;
  FUsercancel        = 6;
  FNoSFXarchive      = 7;
  FSFXModuleError    = 8;

const
  kpidNoProperty = 0;
  kpidHandlerItemIndex = 2;
  kpidPath = 3;
  kpidName = 4;
  kpidExtension = 5;
  kpidIsFolder = 6;
  kpidSize = 7;
  kpidPackedSize = 8;
  kpidAttributes = 9;
  kpidCreationTime = 10;
  kpidLastAccessTime = 11;
  kpidLastWriteTime = 12;
  kpidSolid = 13;
  kpidCommented = 14;
  kpidEncrypted = 15;
  kpidSplitBefore = 16;
  kpidSplitAfter = 17;
  kpidDictionarySize = 18;
  kpidCRC = 19;
  kpidType = 20;
  kpidIsAnti = 21;
  kpidMethod = 22;
  kpidHostOS = 23;
  kpidFileSystem = 24;
  kpidUser = 25;
  kpidGroup = 26;
  kpidBlock = 27;
  kpidComment = 28;
  kpidPosition = 29;

  kpidTotalSize = $1100;
  kpidFreeSpace = $1101;
  kpidClusterSize = $1102;
  kpidVolumeName = $1103;

  kpidLocalName = $1200;
  kpidProvider = $1201;
  kpidUserDefined = $10000;

//----------------------------------------------------------------------------------------------------
//--------------Widestring Classes--------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
type
  TWideStringArray = array of WideString;

  TWideStringList_ = class(TObject)
   private
   public
    WStrings: array of WideString;
    Count : Longword;
    constructor Create;
    procedure Clear;
    procedure AddString(s: WideString);
   end;


type
  TCompressType = (LZMA,PPMD);
  TCompressStrength = (SAVE,FAST,NORMAL,MAXIMUM,ULTRA);
  TLZMAStrength = 0..27;
//rg 17.04.06
  TPPMDMem = 1..31;
  TPPMDSize = 2..32;

//rg 5.4.06
  AddOptsEnum = (AddRecurseDirs, AddSolid, AddStoreOnlyFilename, AddIncludeDriveLetter);
  AddOpts = Set Of AddOptsEnum;

  ExtractOptsEnum = (ExtractNoPath,ExtractOverwrite);
  ExtractOpts = Set Of ExtractOptsEnum;

//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//--------------Start SevenZip Interface-------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------


type
  TInterfacedObject = class(TObject, IUnknown)
  protected
    FRefCount: Integer;
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
  public
    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;
    class function NewInstance: TObject; override;
    property RefCount: Integer read FRefCount;
  end;

const
//  Correct below for 7-Zip 4.23 or 4.29. Comment this line
  {$DEFINE 7z423_}
  {$DEFINE 7z429}

  {$IFDEF 7z423}
  szCLSID_CFormat7z = '{23170F69-40C1-278A-1000-000110050000}';
  szIID_IInArchive = '{23170F69-40C1-278A-0000-000100080000}';
  szIID_IOutArchive = '{23170F69-40C1-278A-0000-000100020000}';
  szIID_ISetProperties = '{23170F69-40C1-278A-0000-000100030000}';
  szIID_IOutStream = '{23170F69-40C1-278A-0000-000000040000}';
  szIID_ISequentialInStream = '{23170F69-40C1-278A-0000-000000010000}';
  szIID_IInStream = '{23170F69-40C1-278A-0000-000000030000}';
  szIID_IStreamGetSize = '{23170F69-40C1-278A-0000-000000060000}';
  szIID_IArchiveOpenCallback = '{23170F69-40C1-278A-0000-000100010000}';
  szIID_IArchiveExtractCallback = '{23170F69-40C1-278A-0000-000100090000}';
  szIID_IArchiveUpdateCallback = '{23170F69-40C1-278A-0000-000100040000}';
  szIID_IProgress = '{23170F69-40C1-278A-0000-000000050000}';
  szIID_ISequentialOutStream = '{23170F69-40C1-278A-0000-000000020000}';
  {$ENDIF}

  {$IFDEF 7z429}
//000
  szIID_IProgress =                         '{23170F69-40C1-278A-0000-000000050000}';
//30
  szIID_ISequentialInStream =               '{23170F69-40C1-278A-0000-000300010000}';
  szIID_ISequentialOutStream =              '{23170F69-40C1-278A-0000-000300020000}';
  szIID_IInStream =                         '{23170F69-40C1-278A-0000-000300030000}';
  szIID_IOutStream =                        '{23170F69-40C1-278A-0000-000300040000}';
  szIID_IStreamGetSize =                    '{23170F69-40C1-278A-0000-000300060000}';
  szIID_IOutStreamFlush =                   '{23170F69-40C1-278A-0000-000300070000}';
//400
  szIID_ICompressProgressInfo =             '{23170F69-40C1-278A-0000-000400040000}';
  szIID_ICompressCoder =                    '{23170F69-40C1-278A-0000-000400050000}';
  szIID_ICompressCoder2 =                   '{23170F69-40C1-278A-0000-000400180000}';
  szIID_ICompressSetCoderProperties =       '{23170F69-40C1-278A-0000-000400200000}';
  szIID_ICompressSetDecoderProperties =     '{23170F69-40C1-278A-0000-000400210000}';
  szIID_ICompressSetDecoderProperties2 =    '{23170F69-40C1-278A-0000-000400220000}';
  szIID_ICompressWriteCoderProperties =     '{23170F69-40C1-278A-0000-000400230000}';
  szIID_ICompressGetInStreamProcessedSize = '{23170F69-40C1-278A-0000-000400240000}';
  szIID_ICompressGetSubStreamSize =         '{23170F69-40C1-278A-0000-000400300000}';
  szIID_ICompressSetInStream =              '{23170F69-40C1-278A-0000-000400310000}';
  szIID_ICompressSetOutStream =             '{23170F69-40C1-278A-0000-000400320000}';
  szIID_ICompressSetInStreamSize =          '{23170F69-40C1-278A-0000-000400330000}';
  szIID_ICompressSetOutStreamSize =         '{23170F69-40C1-278A-0000-000400340000}';
  szIID_ICompressFilter =                   '{23170F69-40C1-278A-0000-000400400000}';
  szIID_ICryptoProperties =                 '{23170F69-40C1-278A-0000-000400800000}';
  szIID_ICryptoSetPassword =                '{23170F69-40C1-278A-0000-000400900000}';
  szIID_ICryptoSetCRC =                     '{23170F69-40C1-278A-0000-000400A00000}';
//500
  szIID_ICryptoGetTextPassword =            '{23170F69-40C1-278A-0000-000500100000}';
  szIID_ICryptoGetTextPassword2 =           '{23170F69-40C1-278A-0000-000400110000}';
//600
  szIID_ISetProperties =                    '{23170F69-40C1-278A-0000-000600030000}';
  szIID_IArchiveOpenCallback =              '{23170F69-40C1-278A-0000-000600100000}';
  szIID_IArchiveExtractCallback =           '{23170F69-40C1-278A-0000-000600200000}';
  szIID_IArchiveOpenVolumeCallback =        '{23170F69-40C1-278A-0000-000600300000}';
  szIID_IInArchiveGetStream =               '{23170F69-40C1-278A-0000-000600400000}';
  szIID_IArchiveOpenSetSubArchiveName =     '{23170F69-40C1-278A-0000-000600500000}';
  szIID_IInArchive =                        '{23170F69-40C1-278A-0000-000600600000}';
  szIID_IArchiveUpdateCallback =            '{23170F69-40C1-278A-0000-000600800000}';
  szIID_IArchiveUpdateCallback2 =           '{23170F69-40C1-278A-0000-000600820000}';
  szIID_IOutArchive =                       '{23170F69-40C1-278A-0000-000600A00000}';

  szCLSID_CFormat7z =                       '{23170F69-40C1-278A-1000-000110070000}';
  {$ENDIF}

  CLSID_CFormat7z: TGUID = szCLSID_CFormat7z;
  IID_IInArchive: TGUID = szIID_IInArchive;
  IID_IOutArchive: TGUID = szIID_IOutArchive;
  IID_ISetProperties: TGUID = szIID_ISetProperties;
  IID_IOutStream: TGUID = szIID_IOutStream;
  IID_ISequentialInStream: TGUID = szIID_ISequentialInStream;
  IID_IInStream: TGUID = szIID_IInStream;
  IID_IStreamGetSize: TGUID = szIID_IStreamGetSize;
  IID_IArchiveOpenCallback: TGUID = szIID_IArchiveOpenCallback;
  IID_IArchiveExtractCallback: TGUID = szIID_IArchiveExtractCallback;
  IID_IArchiveOpenVolumeCallback: TGUID = szIID_IArchiveOpenVolumeCallback;
  IID_IArchiveUpdateCallback: TGUID = szIID_IArchiveUpdateCallback;
  IID_IArchiveUpdateCallback2: TGUID = szIID_IArchiveUpdateCallback2;
  IID_IProgress: TGUID = szIID_IProgress;
  IID_ISequentialOutStream: TGUID = szIID_ISequentialOutStream;


type
  HARC = THandle;

  INDIVIDUALINFO = record
    dwOriginalSize: DWORD;
    dwCompressedSize: DWORD;
    dwCRC: DWORD;
    uFlag: UINT;
    uOSType: UINT;
    wRatio: WORD;
    wDate: WORD;
    wTime: WORD;
    szFileName: array [0..FNAME_MAX32 + 1 - 1] of Char;
    dummy1: array [0..3 - 1] of Char;
    szAttribute: array [0..8 - 1] of Char;
    szMode: array [0..8 - 1] of Char;
  end;

  EXTRACTINGINFO = record
    dwFileSize: DWORD;
    dwWriteSize: DWORD;
    szSourceFileName: array [0..FNAME_MAX32 + 1 - 1] of Char;
    dummy1: array [0..3 - 1] of Char;
    szDestFileName: array [0..FNAME_MAX32 + 1 - 1] of Char;
    dummy: array [0..3 - 1] of Char;
  end;

  PEXTRACTINGINFOEX = ^EXTRACTINGINFOEX;
  EXTRACTINGINFOEX = record
    exinfo: EXTRACTINGINFO;
    dwCompressedSize: DWORD;
    dwCRC: DWORD;
    uOSType: UINT;
    wRatio: WORD;
    wDate: WORD;
    wTime: WORD;
    szAttribute: array [0..8 - 1] of Char;
    szMode: array [0..8 - 1] of Char;
  end;

type
  ISetProperties = interface(IUnknown)
    [szIID_ISetProperties]
    function SetProperties(const names: PWideChar; const values: PPROPVARIANT; numProperties: Integer): Integer; stdcall;
  end;

  ISequentialOutStream = interface(IUnknown)
    [szIID_ISequentialOutStream]
    function Write(const data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
    {$IFDEF 7z423}
    function WritePart(const data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
    {$ENDIF}
  end;

  ISequentialInStream = interface(IUnknown)
    [szIID_ISequentialInStream]
    function Read(var data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
    {$IFDEF 7z423}
    function ReadPart(var data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
    {$ENDIF}
  end;

  IInStream = interface(ISequentialInStream)
    [szIID_IInStream]
    function Seek(offset: Int64; seekOrigin: DWORD;
      newPosition: PInt64): Integer; stdcall;
  end;

  IStreamGetSize = interface(IUnknown)
    [szIID_IStreamGetSize]
    function GetSize(var size: Int64): Integer; stdcall;
  end;

  IArchiveOpenCallback = interface(IUnknown)
    [szIID_IArchiveOpenCallback]
    function SetTotal(const files: Int64; const bytes: Int64): Integer; stdcall;
    function SetCompleted(const files: Int64; const bytes: Int64): Integer; stdcall;
  end;

//by rg
  IArchiveOpenVolumeCallback = interface(IUnknown)
    [szIID_IArchiveOpenVolumeCallback]
    function GetProperty(propID: PROPID; var value: PROPVARIANT): integer; stdcall;
    function GetStream(const name:Widechar; var inStream: IInStream): integer; stdcall;
  end;
//end by rg

  IProgress = interface(IUnknown)
    [szIID_IProgress]
    function SetTotal(total: Int64): Integer; stdcall;
    function SetCompleted(const completeValue: PInt64): Integer; stdcall;
  end;

  IArchiveExtractCallback = interface(IProgress)
    [szIID_IArchiveExtractCallback]
    function GetStream(index: DWORD; out outStream: ISequentialOutStream;  askExtractMode: DWORD): Integer; stdcall;
    // GetStream OUT: S_OK - OK, S_FALSE - skeep this file
    function PrepareOperation(askExtractMode: Integer): Integer; stdcall;
    function SetOperationResult(resultEOperationResult: Integer): Integer; stdcall;
  end;

  IInArchive = interface(IUnknown)
    [szIID_IInArchive]
    function Open(stream: IInStream; const maxCheckStartPosition: PInt64;  openArchiveCallback: IArchiveOpenCallback): Integer; stdcall;
    function Close(): Integer; stdcall;
    function GetNumberOfItems(out numItems: DWORD): Integer; stdcall;
    function GetProperty(index: DWORD; propID: PROPID; var value: PROPVARIANT): Integer; stdcall;
    function Extract(const indices: PDWORD; numItems: DWORD;   testMode: Integer; extractCallback: IArchiveExtractCallback): Integer; stdcall;
    function GetArchiveProperty(propID: PROPID; value: PPROPVARIANT): Integer; stdcall;
    function GetNumberOfProperties(var numProperties: DWORD): Integer; stdcall;
    function GetPropertyInfo(index: DWORD; var name: TBSTR; var propID: PROPID; var varType: {PVARTYPE}Integer): Integer; stdcall;
    function GetNumberOfArchiveProperties(var numProperties): Integer; stdcall;
    function GetArchivePropertyInfo(index: DWORD; name: PBSTR; propID: PPROPID; varType: {PVARTYPE}PInteger): Integer; stdcall;
  end;

  IArchiveUpdateCallback = interface(IProgress)
    [szIID_IArchiveUpdateCallback]
    //function EnumProperties(var enumerator: IEnumSTATPROPSTG): Integer; stdcall;
    function GetUpdateItemInfo(index: DWORD;
      newData: PInteger; // 1 - new data, 0 - old data
      newProperties: PInteger; // 1 - new properties, 0 - old properties
      indexInArchive: PDWORD // -1 if there is no in archive, or if doesn't matter
      ): Integer; stdcall;
    function GetProperty(index: DWORD; propID: PROPID; value: PPROPVARIANT): Integer; stdcall;
    function GetStream(index: DWORD; var inStream: ISequentialInStream): Integer; stdcall;
    function SetOperationResult(operationResult: Integer): Integer; stdcall;
  end;


  IArchiveUpdateCallback2 = interface(IProgress)
    [szIID_IArchiveUpdateCallback2]
    //function EnumProperties(var enumerator: IEnumSTATPROPSTG): Integer; stdcall;
    function GetVolumeSize(index: DWORD; Size:DWord): Integer; stdcall;
    function GetVolumeStream(index: DWORD; var volumeStream: ISequentialInStream): Integer; stdcall;
  end;

  IOutArchive = interface(IUnknown)
    [szIID_IOutArchive]
    function UpdateItems(outStream: ISequentialOutStream; numItems: DWORD; updateCallback: IArchiveUpdateCallback): Integer; stdcall;
    function GetFileTimeType(var _type: DWORD): Integer; stdcall;
  end;

  IOutStream = interface(ISequentialOutStream)
    [szIID_IOutStream]
    function Seek(offset: Int64; seekOrigin: DWORD; newPosition: PInt64): Integer; stdcall;
    function SetSize(newSize: Int64): Integer; stdcall;
  end;

  TMyStreamWriter = class(TInterfacedObject, ISequentialOutStream, IOutStream)
    fOut: Integer;
    fArchiveSFXname : Widestring;
    destructor Destroy; override;
    constructor Create(sz: Widestring; szDate:Tdatetime; FAttr:Cardinal; var outfHandle: integer;CreateSFX:Boolean=false);
    function Write(const data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
    function WritePart(const data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
    function Seek(offset: Int64; seekOrigin: DWORD; newPosition: PInt64): Integer; stdcall;
    function SetSize(newSize: Int64): Integer; stdcall;
  end;

  TMyStreamReader = class(TInterfacedObject, IInStream, IStreamGetSize, ISequentialInStream)
    fIn: Integer;
    constructor Create(sz: Widestring);
    destructor Destroy; override;
    function Seek(offset: Int64; seekOrigin: DWORD; newPosition: PInt64): Integer; stdcall;
    function Read(var data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
    function ReadPart(var data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
    function GetSize(var size: Int64): Integer; stdcall;
  end;

  TSevenZip = class;   // for reference only, implementated later below

  TMyArchiveUpdateCallback = class(TInterfacedObject, IArchiveUpdateCallback, IProgress)
    FSevenzip_int: TSevenzip;
    Files: TWideStringArray;//TStringList;
    Files_size:array of int64;
    Files_Date:array of TFiletime;
    Files_Attr:array of Cardinal;
    FProgressFile : Widestring;
    FProgressFilePos : int64;
    FprogressFileSize : int64;
    FLastPos : int64;
    RootDir : widestring;
//    FIncludeDriveletter:Boolean;
    constructor Create(Owner: TSevenZip);
//    destructor destroy;
    function EnumProperties(var enumerator: IEnumSTATPROPSTG): Integer; stdcall;
    function GetUpdateItemInfo(index: DWORD;
      newData: PInteger; // 1 - new data, 0 - old data
      newProperties: PInteger; // 1 - new properties, 0 - old properties
      indexInArchive: PDWORD // -1 if there is no in archive, or if doesn't matter
      ): Integer; stdcall;
    function GetProperty(index: DWORD; propID: PROPID; value: PPROPVARIANT): Integer; stdcall;
    function GetStream(index: DWORD; var inStream: ISequentialInStream): Integer; stdcall;
    function SetOperationResult(operationResult: Integer): Integer; stdcall;
    function SetTotal(total: Int64): Integer; stdcall;
    function SetCompleted(const completeValue: PInt64): Integer; stdcall;
  end;

  TMyArchiveExtractCallback = class(TInterfacedObject, IArchiveExtractCallback)
    FSevenzip_int : TSevenzip;
    FExtractDirectory : Widestring;
    FProgressFile : Widestring;
    FProgressFilePos : int64;
    FProgressFileSize : int64;
    FLastPos : int64;
    FFilestoextract : int64;
    FLastFileToExt : Boolean;
    FAllFilesExt : Boolean;
    constructor Create(Owner: TSevenZip);
    function GetStream(index: DWORD; out outStream: ISequentialOutStream; askExtractMode: DWORD): Integer; stdcall;
    // GetStream OUT: S_OK - OK, S_FALSE - skeep this file
    function PrepareOperation(askExtractMode: Integer): Integer; stdcall;
    function SetOperationResult(resultEOperationResult: Integer): Integer; stdcall;
    function SetTotal(total: Int64): Integer; stdcall;
    function SetCompleted(const completeValue: PInt64): Integer; stdcall;
  end;

  TMyCallBack = class(TInterfacedObject, IArchiveOpenCallback)
    function SetTotal(const files: Int64; const bytes: Int64): Integer; stdcall;
    function SetCompleted(const files: Int64; const bytes: Int64): Integer; stdcall;
  end;

//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//--------------END SevenZip Interface--------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------


  
  
//----------------------------------------------------------------------------------------------------  
//----------------------------------------------------------------------------------------------------  
//--------------Start SevenZip VCL -------------------------------------------------------------  
//----------------------------------------------------------------------------------------------------  
//----------------------------------------------------------------------------------------------------  

//type
  TListfileEvent = procedure(Sender: TObject; Filename: Widestring; Fileindex,FileSizeU,FileSizeP,Fileattr,Filecrc:cardinal;Filemethod:Widestring ;FileTime:double) of object;
  TExtractfileEvent = procedure(Sender: TObject; Filename: Widestring; Filesize:int64) of object;
  TAddFileEvent = procedure(Sender: TObject; Filename: Widestring; Filesize:int64) of object;
  TPreProgressEvent = procedure(Sender: TObject; MaxProgress: int64) of object;
  TProgressEvent = procedure(Sender: TObject; Filename: Widestring; FilePosArc,FilePosFile: int64) of object;
  TMessageEvent = procedure(Sender: TObject; ErrCode: Integer; Message: string;Filename:Widestring)  of object;
//  TCRC32ErrorEvent = procedure(Sender: TObject; ForFile: string;  FoundCRC, ExpectedCRC: LongWord; var DoExtract: Boolean) of object;
//  TCommentEvent = procedure(Sender: TObject;Comment: string;) of object;
  TSetNewNameEvent = procedure(Sender: TObject; var OldFileName: WideString) of object;
  TExtractOverwrite = procedure(Sender: TObject; FileName: WideString; var DoOverwrite : Boolean) of object;

//type
  TSevenZip = class(TComponent)
  private
    FErrCode: Integer;
    FHandle: HWND;
    FMessage: Widestring;
    FExtrBaseDir: Widestring;
//    FExtrOverwrite:Boolean; //rg in Extractoptions
    FSevenZipFileName: Widestring;

    FComment: Widestring;
    FRootDir: Widestring;

    Ffiles : TWideStringList_;

    { Event variables }
    FOnProgress: TProgressEvent;
    FOnPreProgress: TPreProgressEvent;
    FOnMessage: TMessageEvent;
    FOnlistfile: TlistfileEvent;
    FOnextractfile: TextractfileEvent;
    FOnaddfile: TaddfileEvent;
    FOnSetAddName: TSetNewNameEvent;
    FOnSetExtractName: TSetNewNameEvent;
    FOnExtractOverwite : TExtractOverwrite;

//rg 5.4.06
    FAddOptions : Addopts;
//rg 23.8.06
    FExtractOptions : Extractopts;
//rg 24.8.06
    FNumberOfFiles : Integer;
    FIsSFX : Boolean;
    FSFXOffset : int64;
//rg 6.4.06
    FSFXCreate: Boolean;
    FSFXModule : Widestring;
    FCompresstype: TCompresstype;
    FCompstrength: TCompressStrength;
    FLZMAStrength: TLZMAStrength;
//rg 17.04.06
    FPPMDSize: TPPMDSize;
    FPPMDMem: TPPMDMem;
    FMainCancel : Boolean;
    { Private "helper" functions }

//    procedure LogMessage(var msg: TMessage); message 9999;
    procedure ResetCancel;
    function AppendSlash(sDir: widestring): widestring;
  protected
    inA: IInArchive;
    outA: IOutArchive;
    sp: ISetProperties;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    { Public Properties (run-time only) }
    property Handle: HWND read fHandle write fHandle;
    property ErrCode: Integer read fErrCode write fErrCode;
//rg24.8.2006
    property IsSFX: Boolean read FIsSFX write FIsSFX;
    property SFXOffset: int64 read FSFXOffset write FSFXOffset;
    
    property SevenZipComment: Widestring read Fcomment write FComment;
    property Files: TWideStringList_ read Ffiles write ffiles;

    { Public Methods }
    function Add: integer;
    function Extract(TestArchive:Boolean=False): integer;
    function List: integer;
    procedure Cancel;
    function GetIndexByFilename(FileToExtract:Widestring): integer;
    function SFXCheck(Fn:Widestring): Boolean;
    function ConvertSFXto7z(Fn:Widestring): boolean;
    function Convert7ztoSFX(Fn:Widestring): boolean;


  published
    { Public properties that also show on Object Inspector }
    property AddRootDir: Widestring read FRootDir write FRootDir;
//rg 6.6.06
    property SFXCreate: Boolean read FSFXCreate write FSFXCreate;
    property SFXModule: Widestring read FSFXModule write FSFXModule;
//rg 5.4.06
    property AddOptions: AddOpts read FAddOptions write FAddOptions;
//rg 23.8.06
    property ExtractOptions: ExtractOpts read FExtractOptions write FExtractOptions;
    property ExtrBaseDir: Widestring read FExtrBaseDir write FExtrBaseDir;
    property LZMACompressType: TCompresstype read FCompresstype write FCompresstype;
    property LZMACompressStrength: TCompressStrength read FCompstrength write FCompstrength;
    property LZMAStrength: TLZMAStrength read FLZMAStrength write FLZMAstrength;
//rg 17.04.06
    property LPPMDmem: TPPMDmem read FPPMDmem write FPPMDmem;
    property LPPMDsize: TPPMDsize read FPPMDsize write FPPMDsize;
    property SZFileName: Widestring read FSevenZipFileName write FSevenZipFilename;
    property NumberOfFiles: Integer read FNumberOfFiles write FNumberOfFiles;
    { Events }

    property OnListfile: TlistfileEvent read FOnlistfile write FOnlistfile;
    property OnAddfile: TaddfileEvent read FOnaddfile write FOnaddfile;
    property OnExtractfile: TextractfileEvent read FOnextractfile write FOnextractfile;
    property OnProgress: TProgressEvent read FOnProgress  write FOnProgress;
    property OnPreProgress: TPreProgressEvent read FOnPreProgress  write FOnPreProgress;
    property OnMessage: TMessageEvent read fOnMessage write fOnMessage;
    property OnSetAddName: TSetNewNameEvent read FOnSetAddName write FOnSetAddName;
    property OnSetExtractName: TSetNewNameEvent read FOnSetExtractName write FOnSetExtractName;
    property OnExtractOverwrite: TExtractOverwrite read FOnExtractOverwite write FOnExtractOverwite;
  end;


function CreateObject(const clsid: PGUID; const iid: PGUID; out _out): Integer; stdcall; external '7za.dll';
function PropTypeToString(propType: Integer): string;
function PropIDToString(propID: Integer): string;
//procedure Log(sz: string);
function FileTimeToDateTime(const rFileTime: TFileTime; const Localize: Integer = 0): TDateTime;
procedure SortDWord(var A: array of DWord; iLo, iHi: DWord);


//Unicode procedures
function UppercaseW_(s:WideString):Widestring;
function GetFileSizeandDateTime_Int64(fn: Widestring; var fs:int64; var ft:Tfiletime; var fa:integer): int64;
function FileExistsW(fn: Widestring): Boolean;

procedure Register;

var FMainhandle : HWND; //for debug messages

implementation

uses debug;

//--------------------------------------------------------------------------------------------------
//-------------------Start Twidestringlist_-----------------------------------------------------------
//--------------------------------------------------------------------------------------------------

procedure TWideStringList_.AddString(s: WideString);
var i:Longword;
begin
 i := length(WStrings);
 Setlength(WStrings,i+1);
 WStrings[i] := s;
 Count := i+1;
end;

Procedure TWideStringList_.Clear;
begin
 Setlength(WStrings,0);
 Count := 0;
end;

Constructor TWideStringList_.Create;
begin
 clear;
end;

//--------------------------------------------------------------------------------------------------
//-------------------END Twidestringlist_-------------------------------------------------------------
//--------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------
//-------------------Start UniCode procedures-------------------------------------------------------
//--------------------------------------------------------------------------------------------------

function FileExistsW(fn: Widestring): Boolean;
var
 fs:int64;
 ft:Tfiletime;
 fa:integer;
begin
 Result := (GetFileSizeandDateTime_Int64(fn,fs,ft,fa) > -1);
end;

//START function from TNTControls http://www.tntware.com/
function StrScanWide(const Str: PWideChar; Chr: WideChar): PWideChar;
begin
  Result := Str;
  while Result^ <> Chr do
  begin
    if Result^ = #0 then
    begin
      Result := nil;
      Exit;
    end;
    Inc(Result);
  end;
end;

function LastDelimiterW(const Delimiters, S: WideString): Integer;
var
  P: PWideChar;
begin
  Result := Length(S);
  P := PWideChar(Delimiters);
  while Result > 0 do
  begin
    if (S[Result] <> #0) and (StrScanWide(P, S[Result]) <> nil) then
      Exit;
    Dec(Result);
  end;
end;

function ChangeFileExtW(const FileName, Extension: WideString): WideString;
var
  I: Integer;
begin
  I := LastDelimiterW('.\:',Filename);
  if (I = 0) or (FileName[I] <> '.') then I := MaxInt;
  Result := Copy(FileName, 1, I - 1) + Extension;
end;

function ExtractFilePathW(const FileName: WideString): WideString;
var
  I: Integer;
begin
  I := LastDelimiterW('\:', FileName);
  Result := Copy(FileName, 1, I);
end;

function ExtractFileNameW(const FileName: WideString): WideString;
var
  I: Integer;
begin
  I := LastDelimiterW('\:', FileName);
  Result := Copy(FileName, I + 1, MaxInt);
end;

function ExtractFileExtW(const FileName: WideString): WideString;
var
  I: Integer;
begin
  I := LastDelimiterW('.\:', FileName);
  if (I > 0) and (FileName[I] = '.') then
    Result := Copy(FileName, I, MaxInt) else
    Result := '';
end;
//END function from TNTControls http://www.tntware.com/

function GetFileSizeandDateTime_Int64(fn: Widestring; var fs:int64; var ft:Tfiletime; var fa:integer): int64;
var
  FindData: _Win32_Find_Dataw;
  SearchHandle: THandle;
begin
  Result := 0;
  SearchHandle := FindFirstFilew(PWideChar(fn), FindData);
  if SearchHandle = INVALID_HANDLE_VALUE then
   begin
     Result := -1;
     fs := -1;
     fa := -1;
     ft.dwLowDateTime := 0;
     ft.dwHighDateTime := 0;
     exit;
   end;
  try
    LARGE_INTEGER(Result).LowPart := FindData.nFileSizeLow;
    LARGE_INTEGER(Result).HighPart := FindData.nFileSizeHigh;

    LARGE_INTEGER(fs).LowPart := FindData.nFileSizeLow;
    LARGE_INTEGER(fs).HighPart := FindData.nFileSizeHigh;

    ft.dwLowDateTime  := Finddata.ftLastWriteTime.dwLowDateTime;
    ft.dwHighDateTime := Finddata.ftLastWriteTime.dwHighDateTime;
    fa := Finddata.dwFileAttributes;
  finally
    Windows.FindClose(SearchHandle);
  end;
end;

function UppercaseW_(s:WideString):Widestring;
begin
  Result := S;
  if Length(Result) > 0 then
    CharUpperBuffW(PWideChar(Result), Length(Result));
end;

//--------------------------------------------------------------------------------------------------
//-------------------End UniCode procedures---------------------------------------------------------
//--------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------
//  Start common functions
//------------------------------------------------------------------------------------------------
function TSevenZip.AppendSlash(sDir: widestring): widestring;
begin
  if (sDir <> '') and (sDir[Length(sDir)] <> '\') then
    Result := sDir + '\'
  else
    Result := sDir;
end;

function FileTimeToDateTime(const rFileTime: TFileTime; const Localize: Integer = 0): TDateTime;
var
  dOffset: Double;
  rWork: TFileTime;
begin
  // offset to or from local time
  if Localize > 0 then
    FileTimeToLocalFileTime(rFileTime, rWork)
  else if Localize < 0 then
    LocalFileTimeToFileTime(rFileTime, rWork)
  else begin
    rWork := rFileTime;
  end;

  dOffset := 0.0000001 * ((Int64(rWork.dwHighDateTime) shl 32) or rWork.dwLowDateTime);
  dOffset := dOffset / (60 * 60 * 24);
  Result := EncodeDate(1601, 1, 1) + dOffset;
end;

procedure SortDWord(var A: array of DWord; iLo, iHi: DWord);
var
  Lo, Hi, Mid, T: DWord;
begin
    Lo := iLo;
    Hi := iHi;
    Mid := A[(Lo + Hi) div 2];
    repeat
      while A[Lo] < Mid do Inc(Lo);
      while A[Hi] > Mid do Dec(Hi);
      if Lo <= Hi then
      begin
        T := A[Lo];
        A[Lo] := A[Hi];
        A[Hi] := T;
        Inc(Lo);
        if Hi > 0 then Dec(Hi); //Using DWord and not Integers
      end;
    until Lo > Hi;
    if Hi > iLo then SortDWord(A, iLo, Hi);
    if Lo < iHi then SortDWord(A, Lo, iHi);
end;


//------------------------------------------------------------------------------------------------
//  End common functions
//--------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------
//-------------------Start SevenZip Interface -----------------------------------------------
//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------

function TInterfacedObject.QueryInterface(const IID: TGUID; out Obj): HResult;
const
  E_NOINTERFACE = HResult($80004002);
begin
  if GetInterface(IID, Obj) then
  begin
    Result := 0;
    Log('INTERFACEOK:' + ClassName + ' ' + GUIDToString(IID));
  end else
  begin
    Result := E_NOINTERFACE;
    Log('  NOINTERFACE: ' + ClassName + ' ' + GUIDToString(IID));
  end;
end;

function TInterfacedObject._AddRef: Integer;
begin
  Result := InterlockedIncrement(FRefCount);
end;

function TInterfacedObject._Release: Integer;
begin
  Result := InterlockedDecrement(FRefCount);
  if Result = 0 then
    Destroy;
end;

procedure TInterfacedObject.AfterConstruction;
begin
// Release the constructor's implicit refcount
  InterlockedDecrement(FRefCount);
end;

procedure TInterfacedObject.BeforeDestruction;
begin
  //if RefCount <> 0 then Error(reInvalidPtr);
end;

// Set an implicit refcount so that refcounting
// during construction won't destroy the object.
class function TInterfacedObject.NewInstance: TObject;
begin
  Result := inherited NewInstance;
  TInterfacedObject(Result).FRefCount := 1;
end;

constructor TMyArchiveUpdateCallback.Create(Owner: TSevenZip);
begin
  FSevenzip_int := Owner;
 //Files := TStringList.Create;
end;

function TMyArchiveUpdateCallback.EnumProperties(var enumerator: IEnumSTATPROPSTG): Integer; stdcall;
begin
  Result := E_NOTIMPL;
end;

function TMyArchiveUpdateCallback.GetUpdateItemInfo(index: DWORD;
  newData: PInteger; // 1 - new data, 0 - old data
  newProperties: PInteger; // 1 - new properties, 0 - old properties
  indexInArchive: PDWORD // -1 if there is no in archive, or if doesn't matter
  ): Integer; stdcall;
begin
//  Log(Format('TMyArchiveUpdateCallback.GetUpdateItemInfo(%d)', [index]));
  if newData <> nil then
    newData^ := 1;
  if newProperties <> nil then
    newProperties^ := 1;
  if indexInArchive <> nil then
    indexInArchive^ := DWORD(-1);
  Result := S_OK;
end;

function TMyArchiveUpdateCallback.GetProperty(index: DWORD; propID: PROPID;
  value: PPROPVARIANT): Integer; stdcall;
var sz : WideString;
begin
  Log(Format('TMyArchiveUpdateCallback.GetProperty(%d, %s (%d), %.8x)', [index, PropIDToString(propID), propID, Integer(value)]));
  Result := S_OK;
  case propID of
    //kpidPath (3) VT_BSTR (8)
    kpidPath:
    begin
      value^.vt := VT_BSTR;

//get relative path if wanted
      sz := Files[index];
      if rootdir <> '' then
      begin
        if Uppercasew_(copy(sz,1,length(rootdir))) = rootdir then
          delete(sz,1,length(rootdir));
      end;

//User set filename in archive if wanted
      if assigned(Fsevenzip_int.OnSetAddName) then
        Fsevenzip_int.OnSetAddName(Fsevenzip_int, sz);

//remove drive / Include drive if wanted
      if sz[2] = ':' then
        begin
         if char(sz[1]) in ['A'..'Z','a'..'z'] then
           if (AddIncludeDriveLetter in Fsevenzip_int.FAddOptions) then //include
            delete(sz,2,1)
           else
             delete(sz,1,3);
        end;

//just store filename
      if (AddStoreOnlyFilename in Fsevenzip_int.FAddOptions) then
        sz := ExtractFileNameW(sz);

      value^.bstrVal := StringToOleStr(sz);
    end;
    //kpidAttributes (9) VT_UI4 (19)
    kpidAttributes:
    begin
      value^.vt := VT_UI4;
      value^.ulVal := Files_Attr[index];//filegetattr(files[index]);
    end;
    kpidCreationTime:
    begin
      value^.vt := VT_FILETIME;

      value^.filetime.dwLowDateTime := 0;
      value^.filetime.dwHighDateTime := 0;
    end;
    //kpidLastWriteTime (12) VT_FILETIME (64)
    kpidLastWriteTime:
    begin
      value^.vt := VT_FILETIME;
      value^.filetime.dwLowDateTime := Files_Date[index].dwLowDateTime;;
      value^.filetime.dwHighDateTime := Files_Date[index].dwHighDateTime;
    end;
    kpidIsFolder:
    begin
      value^.vt := VT_BOOL;
      value^.boolVal := False;
    end;
    kpidIsAnti:
    begin
      value^.vt := VT_BOOL;
      value^.boolVal := False;
    end;
    //kpidSize (7) VT_UI8 (21)
    kpidSize:
    begin
      value^.vt := VT_UI8;
      value^.uhVal.QuadPart := Files_size[index];
    end;
  else
    Log('Asking for unknown property');
    Result := S_FALSE;
  end;
end;

function TMyArchiveUpdateCallback.GetStream(index: DWORD;
  var inStream: ISequentialInStream): Integer; stdcall;
begin
  Log('TMyArchiveUpdateCallback.GetStream');
  Fprogressfile := files[index];
  Fprogressfilesize := files_size[index];
  Fprogressfilepos := 0;
  inStream := TMyStreamReader.Create(Files[index]);
  Result := S_OK;
end;

function TMyArchiveUpdateCallback.SetOperationResult(operationResult: Integer): Integer; stdcall;
begin
  Log(Format('TMyArchiveUpdateCallback.SetOperationResult(%d)', [operationResult]));
  Result := S_OK;
end;

function TMyArchiveUpdateCallback.SetTotal(total: Int64): Integer; stdcall;
begin
  Log(Format('TMyArchiveUpdateCallback.SetTotal(%d)', [total]));
  Result := S_OK;
end;

function TMyArchiveUpdateCallback.SetCompleted(const completeValue: PInt64): Integer; stdcall;
begin
/// Progressfile - Newfile
/// Do it here because it works with Multithreaded 7za interaction.
  Log(Format('TMyArchiveUpdateCallback.SetCompleted(%d)', [completeValue^]));

//fileprogress
   if (FProgressFilePos = 0) then
      if assigned(Fsevenzip_int.OnAddFile) then Fsevenzip_int.onAddFile(Fsevenzip_int,FProgressFile,FProgressFileSize);
   FProgressFilePos := FProgressFilePos + (completeValue^ - FLastPos);
   FLastPos := completeValue^;

//full and file progress position
   if assigned(Fsevenzip_int.OnProgress) then Fsevenzip_int.OnProgress(Fsevenzip_int,FProgressFile,completeValue^,FProgressFilePos);

  Result := S_OK;
//rg 24.06
//User cancel operation
  if FSevenzip_int.FMainCancel then
   begin
     if assigned(Fsevenzip_int.onMessage) then Fsevenzip_int.OnMessage(Fsevenzip_int,6,'User canceled operation',FProgressFile);
     Result := S_FALSE;
   end;
end;

function TMyStreamWriter.Write(const data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
var
  psize: DWORD;
  fp: Int64;
//  isize: Int64;
begin
  Log(Format('-> Write(%.8x, %d)', [Integer(data), size]));
  if not WriteFile(fOut, data, size, psize, nil) then
    psize := 0;
  if processedSize <> nil then
    processedSize^ := psize;
  fp := 0;
  fp := FileSeek(fOut, fp, 1);
  Log(Format('<- Write(%d, %d)', [psize, fp]));
  Result := S_OK;
end;

function TMyStreamWriter.WritePart(const data; size: DWORD; processedSize: PDWORD): Integer; stdcall;
var
  psize: DWORD;
  fp: Int64;
//  isize: Int64;
begin
  Log('-> WritePart ' + Format('%.8x', [Integer(data)]) + ' ' + IntToStr(size));
  if not WriteFile(fOut, data, size, psize, nil) then
    psize := 0;
  if processedSize <> nil then
    processedSize^ := psize;
  fp := 0;
  fp := FileSeek(fOut, fp, 1);
  Log('<- WritePart ' + IntToStr(psize) + ' ' + IntToStr(fp));
  Result := S_OK;
end;

function TMyStreamWriter.Seek(offset: Int64; seekOrigin: DWORD; newPosition: PInt64): Integer; stdcall;
var
  fp: Int64;
begin
  Log(Format('TMyStreamWriter.Seek(%d, %d, %.8x)', [offset, seekOrigin, Integer(newPosition)]));
  fp := FileSeek(fOut, offset, seekOrigin);
  if newPosition <> nil then
    newPosition^ := fp;
  Result := S_OK;
end;

function TMyStreamWriter.SetSize(newSize: Int64): Integer; stdcall;
begin
  Log(Format('TMyStreamWriter.SetSize(%d)', [newSize]));
  Result := S_FALSE;
end;

constructor TMyArchiveExtractCallback.Create(Owner: TSevenZip);
begin
  inherited Create;
  FSevenzip_int := Owner;
end;

function TMyArchiveExtractCallback.GetStream(index: DWORD;
  out outStream: ISequentialOutStream; askExtractMode: DWORD): Integer; stdcall;
var
 path : Propvariant;
 size : Propvariant;
 date : Propvariant;
 attr : Propvariant;
   sz, origName : Widestring;
   fe,DoOverwrite : boolean;
   fHnd: integer;
begin
  Log(Format('__TMyArchiveExtractCallback.GetStream(%d, %.8x, %d)', [index, Integer(outStream), askExtractMode]));
  path.vt := VT_EMPTY;
  size.vt := VT_EMPTY;
  date.vt := VT_EMPTY;
  attr.vt := VT_EMPTY;

//Cancel Operation
  if self.FSevenzip_int.FMainCancel then
   begin
    outStream := nil;
    result := S_FALSE;
    exit;
   end;

  Case askExtractMode of
    kExtract:  begin

                 FSevenzip_int.inA.GetProperty(index, kpidPath, path);
                 FSevenzip_int.inA.GetProperty(index, kpidSize, size);
                 FSevenzip_int.inA.GetProperty(index, kpidattributes, attr);
                 FSevenzip_int.inA.GetProperty(index, kpidLastWriteTime, date);

//rg 23.8.06
                 if ExtractNoPath in FSevenzip_int.FExtractOptions then
                   sz := FExtractDirectory + extractfilenameW(path.bstrVal)
                  else
                   sz := FExtractDirectory + path.bstrVal;

                 origName := sz;

                 if assigned(Fsevenzip_int.OnSetExtractName) then
                   Fsevenzip_int.OnSetExtractName(Fsevenzip_int, sz);

                 if FileExistsW(sz) then
                   begin
                     if assigned(Fsevenzip_int.OnExtractOverwrite) then
                         Fsevenzip_int.OnExtractOverwrite(Fsevenzip_int, sz,DoOverwrite);

                     if not DoOverwrite then
                      begin
                       Result := S_OK;
                       outStream := nil;
                       exit;
                      end;
                    end;

                 FProgressFile := sz;
                 FProgressFilePos := 0;
                 FprogressFileSize := size.uhVal.QuadPart;

                 if (attr.uiVal and (1 shl 4)) <> 0 then
                   begin
                    CreateDirectoryw(PWidechar(sz),nil);
                   end
                  else
                   begin
                     FFilestoextract := FFilestoextract - 1;
                     if FFilestoextract = 0 then FLastFileToExt := true;
                     outStream := nil;
                     fe := FileExistsW(sz);

                     if (not fe) or ( fe and (  (ExtractOverwrite in FSevenZip_Int.FExtractOptions) or DoOverwrite) ) then
                     begin
                       forcedirectories(extractfilepathW(sz));
                       outStream := TMyStreamWriter.Create(sz, FileTimeToDateTime(date.filetime,2), attr.lVal, fHnd);
                       if fHnd = -1 then
                       begin
                         outStream := nil;
                         result := S_FALSE;
                         exit;
// did not work here need another place !
 //                         if assigned(Fsevenzip_int.onmessage) then Fsevenzip_int.onmessage(Fsevenzip_int, 2, 'Could not create file', origName);
                       end;
                     end;
                   end;
               end;
    ktest    : begin
                 FSevenzip_int.inA.GetProperty(index, kpidPath, path);
                 FSevenzip_int.inA.GetProperty(index, kpidSize, size);
                 FProgressFile := path.bstrVal;
                 FProgressFilePos := 0;
                 FprogressFileSize := size.uhVal.QuadPart ;
               end;
    kskip    : begin
               end;
  end;
  Result := S_OK;
end;
// GetStream OUT: S_OK - OK, S_FALSE - skeep this file

function TMyArchiveExtractCallback.PrepareOperation(askExtractMode: Integer): Integer; stdcall;
begin
  Log(Format('TMyArchiveExtractCallback.PrepareOperation(%d)', [askExtractMode]));
  Result := S_OK;
end;

function TMyArchiveExtractCallback.SetOperationResult(resultEOperationResult: Integer): Integer; stdcall;
begin
  Result := S_OK;
  Log(Format('TMyArchiveExtractCallback.SetOperationResult(%d)', [resultEOperationResult]));
  case resultEOperationResult of
    kOK                : ;
    kUnSupportedMethod : if assigned(Fsevenzip_int.onmessage) then Fsevenzip_int.onmessage(Fsevenzip_int,4,'Unsupported Method',FProgressFile);
    kDataError         : if assigned(Fsevenzip_int.onmessage) then Fsevenzip_int.onmessage(Fsevenzip_int,2,'Data Error',FProgressFile);
    kCRCError          : if assigned(Fsevenzip_int.onmessage) then Fsevenzip_int.onmessage(Fsevenzip_int,3,'CRC Error',FProgressFile);
  end;

  if FLastFileToExt then FAllFilesExt := true; //no more files to extract, we can stop
end;

function TMyArchiveExtractCallback.SetTotal(total: Int64): Integer; stdcall;
begin
  Log(Format('TMyArchiveExtractCallback.SetTotal(%d)', [total]));
//all filesizes also skipped ones
  if FFilestoextract = 0 then // we extract all files, so we set FMaxProgress here
     if assigned(FSevenzip_int.OnPreProgress) then FSevenzip_int.OnPreProgress(FSevenzip_int,total);
  Result := S_OK;
end;

function TMyArchiveExtractCallback.SetCompleted(const completeValue: PInt64): Integer; stdcall;
begin

   if (FProgressFilePos = 0) then
      if assigned(Fsevenzip_int.OnExtractFile) then Fsevenzip_int.onExtractfile(Fsevenzip_int,FProgressFile,FProgressFileSize);

   FProgressFilePos := FProgressFilePos + (completeValue^ - FLastPos);
   FLastPos := completeValue^;

//full and file progress position
   if assigned(Fsevenzip_int.OnProgress) then Fsevenzip_int.OnProgress(Fsevenzip_int,FProgressFile,completeValue^,FProgressFilePos);

  Log(Format('TMyArchiveExtractCallback.SetCompleted(%d)', [completeValue^]));
  Result := S_OK;

  //have all files extracted. Could stop
  //User cancel operation
  if self.FAllFilesExt then Result := S_FALSE;

  if  Fsevenzip_int.FMainCancel then
   begin
     Result := S_FALSE;
     if assigned(Fsevenzip_int.onMessage) then Fsevenzip_int.OnMessage(Fsevenzip_int,6,'User canceled operation',FProgressFile);
   end;
end;


function TMyCallBack.SetTotal(const files: Int64; const bytes: Int64): Integer; stdcall;
begin
  //frmMain.mmoLog.Lines.Add('SetTotal');
  Result := S_OK;
end;

function TMyCallBack.SetCompleted(const files: Int64; const bytes: Int64): Integer; stdcall;
begin
  //frmMain.mmoLog.Lines.Add('SetCompleter');
  Result := S_OK;
end;

destructor TMyStreamReader.Destroy;
begin
  inherited;
  FileClose(fIn);
end;

destructor TMyStreamWriter.Destroy;
begin
  inherited;
  FileClose(fOut);
  Log('TMyStreamWriter.Destroy');
end;

function TMyStreamReader.GetSize(var size: Int64): Integer; stdcall;
var
  iCurrent: Int64;
begin
  //frmMain.mmoLog.Lines.Add('GetSize');
  iCurrent := 0;
  iCurrent := FileSeek(fIn, iCurrent, 1);
  size := 0;
  size := FileSeek(fIn, Size, 2);
  FileSeek(fIn, iCurrent, 0);
  Result := S_OK;
end;

constructor TMyStreamReader.Create(sz: Widestring);
begin
  inherited Create;
   fIn := CreateFilew(PwideChar(sz), GENERIC_READ, 0, nil, OPEN_EXISTING, 0, 0);
  //frmMain.mmoLog.Lines.Add(IntToStr(fIn));
end;

constructor TMyStreamWriter.Create(sz: Widestring; szDate:Tdatetime; FAttr:Cardinal; var outfHandle: integer;CreateSFX:Boolean=false);
begin
  inherited Create;

// rg 6.4.06
  if CreateSFX then
    begin
     fOut := CreateFilew(PwideChar(sz), GENERIC_READ or GENERIC_WRITE, 0, nil, OPEN_EXISTING, FAttr, 0);
     if fout > 0 then fileseek(fout,0,2);
    end
   else
    fOut := CreateFilew(PwideChar(sz), GENERIC_READ or GENERIC_WRITE, 0, nil, CREATE_ALWAYS {CREATE_ALWAYS}, FAttr, 0);

  filesetdate(fout,DateTimeToFileDate(szdate));

  outfHandle := fOut;

  //frmMain.mmoLog.Lines.Add(IntToStr(fIn));
end;

function TMyStreamReader.Seek(offset: Int64; seekOrigin: DWORD;
  newPosition: PInt64): Integer; stdcall;
begin
  //frmMain.mmoLog.Lines.Add('-> Seek ' + IntToStr(offset) + ' ' + IntToStr(seekOrigin));
  if newPosition = nil then
  begin
    FileSeek(fIn, offset, seekOrigin);
    //frmMain.mmoLog.Lines.Add('<- Seek');
  end else
  begin
    newPosition^ := FileSeek(fIn, offset, seekOrigin);
    //frmMain.mmoLog.Lines.Add('<- Seek ' + IntToStr(newPosition^));
  end;
  Result := S_OK;
end;

function TMyStreamReader.Read(var data; size: DWORD;
  processedSize: PDWORD): Integer; stdcall;
var
  psize: DWORD;                                 
  fp: Int64;
//  isize: Int64;
begin
  //frmMain.mmoLog.Lines.Add('-> Read ' + Format('%.8x', [Integer(data)]) + ' ' + IntToStr(size));
  if not ReadFile(fIn, data, size, psize, nil) then
    psize := 0;
  if processedSize <> nil then
    processedSize^ := psize;
  fp := 0;
  fp := FileSeek(fIn, fp, 1);
  //frmMain.mmoLog.Lines.Add('<- Read ' + IntToStr(psize) + ' ' + IntToStr(fp));
  //Application.ProcessMessages;
  Result := S_OK;
end;

function TMyStreamReader.ReadPart(var data; size: DWORD;
  processedSize: PDWORD): Integer; stdcall;
var
  psize: DWORD;
  fp: Int64;
//  isize: Int64;
begin
  //frmMain.mmoLog.Lines.Add('-> ReadPart ' + IntToStr(size));
  if not ReadFile(fIn, data, size, psize, nil) then
    psize := 0;
  if processedSize <> nil then
    processedSize^ := psize;
  fp := 0;
  fp := FileSeek(fIn, fp, 1);
  //frmMain.mmoLog.Lines.Add('<- ReadPart ' + IntToStr(psize) + ' ' + IntToStr(fp));
  //Application.ProcessMessages;
  Result := S_OK;
end;

// ------------------------------------------------------------------------------------------
//functions for SevenZip
function PropIDToString(propID: Integer): string;
begin
  case propID of
    kpidNoProperty        : Result := 'kpidNoProperty';
    kpidHandlerItemIndex  : Result := 'kpidHandlerItemIndex';
    kpidPath              : Result := 'kpidPath';
    kpidName              : Result := 'kpidName';
    kpidExtension         : Result := 'kpidExtension';
    kpidIsFolder          : Result := 'kpidIsFolder';
    kpidSize              : Result := 'kpidSize';
    kpidPackedSize        : Result := 'kpidPackedSize';
    kpidAttributes        : Result := 'kpidAttributes';
    kpidCreationTime      : Result := 'kpidCreationTime';
    kpidLastAccessTime    : Result := 'kpidLastAccessTime';
    kpidLastWriteTime     : Result := 'kpidLastWriteTime';
    kpidSolid             : Result := 'kpidSolid';
    kpidCommented         : Result := 'kpidCommented';
    kpidEncrypted         : Result := 'kpidEncrypted';
    kpidSplitBefore       : Result := 'kpidSplitBefore';
    kpidSplitAfter        : Result := 'kpidSplitAfter';
    kpidDictionarySize    : Result := 'kpidDictionarySize';
    kpidCRC               : Result := 'kpidCRC';
    kpidType              : Result := 'kpidType';
    kpidIsAnti            : Result := 'kpidIsAnti';
    kpidMethod            : Result := 'kpidMethod';
    kpidHostOS            : Result := 'kpidHostOS';
    kpidFileSystem        : Result := 'kpidFileSystem';
    kpidUser              : Result := 'kpidUser';
    kpidGroup             : Result := 'kpidGroup';
    kpidBlock             : Result := 'kpidBlock';
    kpidComment           : Result := 'kpidComment';
    kpidPosition          : Result := 'kpidPosition';

    kpidTotalSize         : Result := 'kpidTotalSize';
    kpidFreeSpace         : Result := 'kpidFreeSpace';
    kpidClusterSize       : Result := 'kpidClusterSize';
    kpidVolumeName        : Result := 'kpidVolumeName';

    kpidLocalName         : Result := 'kpidLocalName';
    kpidProvider          : Result := 'kpidProvider';
    kpidUserDefined       : Result := 'kpidUserDefined';
  else
    Result := 'unknown';
  end;
end;

function PropTypeToString(propType: Integer): string;
begin
  case propType of
    VT_EMPTY           : Result := 'VT_EMPTY';
    VT_NULL            : Result := 'VT_NULL';
    VT_I2              : Result := 'VT_I2';
    VT_I4              : Result := 'VT_I4';
    VT_R4              : Result := 'VT_R4';
    VT_R8              : Result := 'VT_R8';
    VT_CY              : Result := 'VT_CY';
    VT_DATE            : Result := 'VT_DATE';
    VT_BSTR            : Result := 'VT_BSTR';
    VT_DISPATCH        : Result := 'VT_DISPATCH';
    VT_ERROR           : Result := 'VT_ERROR';
    VT_BOOL            : Result := 'VT_BOOL';
    VT_VARIANT         : Result := 'VT_VARIANT';
    VT_UNKNOWN         : Result := 'VT_UNKNOWN';
    VT_DECIMAL         : Result := 'VT_DECIMAL';
    VT_I1              : Result := 'VT_I1';
    VT_UI1             : Result := 'VT_UI1';
    VT_UI2             : Result := 'VT_UI2';
    VT_UI4             : Result := 'VT_UI4';
    VT_I8              : Result := 'VT_I8';
    VT_UI8             : Result := 'VT_UI8';
    VT_INT             : Result := 'VT_INT';
    VT_UINT            : Result := 'VT_UINT';
    VT_VOID            : Result := 'VT_VOID';
    VT_HRESULT         : Result := 'VT_HRESULT';
    VT_PTR             : Result := 'VT_PTR';
    VT_SAFEARRAY       : Result := 'VT_SAFEARRAY';
    VT_CARRAY          : Result := 'VT_CARRAY';
    VT_USERDEFINED     : Result := 'VT_USERDEFINED';
    VT_LPSTR           : Result := 'VT_LPSTR';
    VT_LPWSTR          : Result := 'VT_LPWSTR';
    VT_FILETIME        : Result := 'VT_FILETIME';
    VT_BLOB            : Result := 'VT_BLOB';
    VT_STREAM          : Result := 'VT_STREAM';
    VT_STORAGE         : Result := 'VT_STORAGE';
    VT_STREAMED_OBJECT : Result := 'VT_STREAMED_OBJECT';
    VT_STORED_OBJECT   : Result := 'VT_STORED_OBJECT';
    VT_BLOB_OBJECT     : Result := 'VT_BLOB_OBJECT';
    VT_CF              : Result := 'VT_CF';
    VT_CLSID           : Result := 'VT_CLSID';
  else
    Result := 'Unknown';
  end;
end;
//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------
//-------------------End SevenZip Interface -------------------------------------------------
//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------
//-----------------START DEBUG ONLY-----------------------------------------------------------------
//--------------------------------------------------------------------------------------------------
// For Debug only
// Add this to your MainForm public
{
 procedure LogMessage(var msg: TMessage); message 9999;
}
{
  procedure TForm1.SevenZip1Message(Sender: TObject; ErrCode: Integer; Message: String);
  begin
   memo1.lines.add(message);
  end;
}
// and to Form.Activate this (or set it when you want with e.g. a Button)
{
 sevenzipvcl.FMainhandle := form1.Handle;
}

{procedure Log(sz: string);
var
  p: PString;
begin
  p := new(PString);
  p^ := sz;
  PostMessage(fMainhandle, 9999, 0, Integer(p));
end;}

//--------------------------------------------------------------------------------------------------
//-----------------END DEBUG ONLY-------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------

//------------------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------------------
//-----------------Start SevenZip VCL-------------------------------------------------------
//------------------------------------------------------------------------------------------------

//------------------------------------------------------------------------------------------------
//constructor destructor
//------------------------------------------------------------------------------------------------
(*
procedure TSevenZip.LogMessage(var msg: TMessage);
begin
  if assigned(onMessage) then OnMessage(Self,0,PString(msg.LParam)^);
  Dispose(PString(msg.LParam));
end;
*)

constructor TSevenZip.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ffiles := TWideStringList_.Create;

  ResetCancel;
  FMainHandle := FHandle;
  FNumberOfFiles := -1;
  FSFXModule := '7z.sfx';

  CreateObject(@CLSID_CFormat7z, @IID_IInArchive, inA);

  CreateObject(@CLSID_CFormat7z, @IID_IOutArchive, outA);

  CreateObject(@CLSID_CFormat7z, @IID_ISetProperties, sp);
end;

destructor TSevenzip.Destroy;
begin
  inherited Destroy;
  ffiles.Clear;
  ffiles.Free;
end;
//------------------------------------------------------------------------------------------------
//End constructor destructor
//------------------------------------------------------------------------------------------------

Procedure TSevenZip.Cancel; // public
begin
 FMainCancel := True;
end;

Procedure TSevenZip.ResetCancel; // private
begin
 FMainCancel := False;
end;
//RG 02.06.2006

function TSevenZip.GetIndexByFilename(FileToExtract:Widestring): integer;
var n,i : integer;
    w:Dword;
    fnameprop: PROPVARIANT;
    fileInArchive: widestring;
    ms: TMyStreamReader;
begin
 try
  Result := -1;
  ms := TMyStreamReader.Create(FSevenZipFileName);
  inA.Close;
  i := inA.Open(ms, nil, nil);
  inA.GetNumberOfItems(w); //1..end
  FileToExtract := UppercaseW_(FileToExtract);
  for n:=0 to w-1 do
   begin
    fnameprop.vt := VT_EMPTY;
    inA.GetProperty(n, kpidPath, fnameprop);
    fileInArchive := UppercaseW_(OleStrToString(fnameprop.bstrVal));
    if (fileInArchive = FileToExtract) then
     begin
       Result := n;
       break;
     end;
   end;
 finally
  ina.close;
 end
end;

//-------------------------------------------------------
//SFX functions
//rg 25.8.2006
function TSevenZip.SFXCheck(Fn:Widestring): Boolean;
const ID_7z : Array[0..5] of byte = (55,122,188,175,39,28);
var Source : Integer;
    buffer : array[0..81920] of byte;
    readbytes,i,j:Dword;
begin
 Result := False;
 if UppercaseW_(extractfileextW(Fn)) <> '.EXE' then exit;

 FSFXoffset := 0;
 FIsSFX := false;

 try
  Source  := CreateFilew(PwideChar(Fn), GENERIC_READ, 0, nil, OPEN_EXISTING, 0, 0);
  fileseek(source,65536,0);
  ReadFile(Source,buffer[1],81920,readbytes,nil);

  for i := 0 to sizeof(buffer)-6 do
    begin
      for j := 0 to 5 do
        begin
          if (buffer[i+j] <> ID_7z[j]) then break;
          if j = 5 then FIsSFX := true;
        end;

     if FIsSFX then
      begin
        FSFXOffset := i-1+65536;
        Result := true;
        break;
      end;
    end;

 finally
  Fileclose(source);
 end;

end;

function TSevenZip.ConvertSFXto7z(Fn:Widestring): boolean;
var Source,Dest : Integer;
    DestFn : Widestring;
    buffer : pointer;
    readbytes,writebytes:Dword;
const
    chunksize = 1024*128;
begin
try
  DestFn := changefileextW(Fn,'.7z');
  Source := CreateFilew(PwideChar(Fn), GENERIC_READ, 0, nil, OPEN_EXISTING, 0, 0);
  Fileseek(Source,SFXoffset,0); //goto 7z data
  Dest := CreateFilew(PwideChar(DestFn), GENERIC_Write, 0, nil, CREATE_ALWAYS, 0, 0);

  GetMem(Buffer,chunksize); { allocate the buffer }

  Repeat
     readbytes := Fileread(Source,buffer^,chunksize);
    writebytes := Filewrite(Dest,buffer^,readbytes);
  until readbytes < chunksize;

finally
 freemem(buffer);
 Fileclose(Source);
 Fileclose(Dest);
end;

end;

function TSevenZip.Convert7ztoSFX(Fn:Widestring): boolean;
var Source,Dest : Integer;
    DestFn : Widestring;
    buffer : pointer;
    readbytes,writebytes:Dword;
const
    chunksize = 1024*128;
begin
 Result := false;
 DestFn := changefileextW(Fn,'.exe');
 if not copyfilew(PWidechar(sfxmodule),PWideChar(DestFn),True) then
  begin
   if assigned(onMessage) then onMessage(self,8,'SFXModule error (Not found)',Fsevenzipfilename);
   exit;
  end;

try
  Source := CreateFilew(PwideChar(Fn), GENERIC_READ, 0, nil, OPEN_EXISTING, 0, 0);
  Dest := CreateFilew(PwideChar(DestFn), GENERIC_Write, 0, nil, OPEN_EXISTING, 0, 0);
  fileseek(Dest,0,2);

  GetMem(Buffer,chunksize); { allocate the buffer }

  Repeat
     readbytes := Fileread(Source,buffer^,chunksize);
    writebytes := Filewrite(Dest,buffer^,readbytes);
  until readbytes < chunksize;

finally
 Result := true;
 freemem(buffer);
 Fileclose(Source);
 Fileclose(Dest);
end;

end;


//SFX functions end
//-------------------------------------------------------

function TSevenZip.List: integer;
var
  ms: TMyStreamReader;
  i: Integer;
  w: DWord;
  name: TBSTR;
  prop: PROPID;
  pType: Integer;
  path: PROPVARIANT;
  size: PROPVARIANT;
  packedsize: PROPVARIANT;
  attr:PROPVARIANT;
  fcrc: PROPVARIANT;
  szMethod: PROPVARIANT;
  sztime : PROPVARIANT;
  szMethod_WS : Widestring;
  blockpid : PROPVARIANT;
begin
try
  FNumberOfFiles := -1;
  if UppercaseW_(extractfileextW(Fsevenzipfilename)) = '.EXE' then
   begin
     if not SFXCheck(Fsevenzipfilename) then
       begin
         Result := -1;
         if assigned(onMessage) then onMessage(self,7,'File is not an 7z SFX archive',Fsevenzipfilename);
         Exit;
       end;
   end;

  ms := TMyStreamReader.Create(Fsevenzipfilename);
  inA.Close;

// 24.08.06 - Matteo Riso - Status: experimental
// 25.08.06 Modified by rg
//
// If we loaded an .EXE file, we could start reading from offset 132096.
// Also supported by newerdll

//  if FIsSFX then ms.Seek(FSFXOffset,0,@FSFXOffset);

// End - MR modification

  i := inA.Open(ms, nil, nil);

  if i <> 0 then
   begin
     Result := -1;
     if assigned(onMessage) then onMessage(self,1,'File not found',Fsevenzipfilename);
     Exit;
   end;

(*
   inA.GetNumberOfArchiveProperties(w);
   for i := 0 to w-1 do
    begin
     path.vt := VT_EMPTY;
     ina.GetArchiveProperty(kpidPath,@path);
     if assigned(onMessage) then onMessage(self,i,path.bstrVal,path.pwszVal);
    end;
*)

  inA.GetNumberOfProperties(w);

  for i := 0 to w - 1 do
  begin
    name := new(TBSTR);
    ptype := 0;
    inA.GetPropertyInfo(i, name, prop, pType);
    if name = nil then
    begin
      Log(Format('%d %s %s (%d) %s (%d)', [i, '', PropIDToString(prop), prop, PropTypeToString(pType), pType]))
    end else
    begin
      log(Format('%d %s %s (%d) %s (%d)', [i, name, PropIDToString(prop), prop, PropTypeToString(pType), pType]));
    end;
  end;


  inA.GetNumberOfItems(w);
  FNumberOfFiles := w;

  i := 0;
  for i := 0 to w-1 do
   begin
       path.vt := VT_EMPTY;
       size.vt := VT_EMPTY;
       packedsize.vt := VT_EMPTY;
       attr.vt := VT_EMPTY;
       fcrc.vt := VT_EMPTY;
       szmethod.vt := VT_EMPTY;
       sztime.vt := VT_EMPTY;
       blockpid.vt := VT_EMPTY;

       inA.GetProperty(i, kpidPath, path);
       inA.GetProperty(i, kpidSize, size);
       inA.GetProperty(i, kpidPackedSize, packedsize);
       inA.GetProperty(i, kpidAttributes, attr);
       inA.GetProperty(i, kpidCRC, fcrc);
       inA.GetProperty(i, kpidMethod, szMethod);
       inA.GetProperty(i, kpidLastWriteTime, sztime);
       inA.GetProperty(i, kpidblock, blockpid);

       try
       if ( ((attr.uiVal and (1 shl 4)) <> 0) or (size.uhVal.QuadPart = 0) ) then szMethod_WS := 'None'  // is a directory or 0byte file
        else  //rg 18.04.06
         szMethod_WS := Widestring(szmethod.bstrVal); //Check for diectoies or 0 byte files, if not an exception happens
       if assigned(onlistfile) then
          onlistfile(self,
            Widestring(path.bstrVal),  //filename 1
            i,                         //fileindex for extracting
            size.uhVal.QuadPart,       //Filesizeunp  2
            packedsize.uhVal.QuadPart, //FilesizeP 3
            (attr.uhVal.QuadPart and not (1 shl 13)), //attr 4 , removes first set bit
            fcrc.uhVal.QuadPart,       //CRC 5
            szMethod_WS,               //method 6
            FileTimeToDateTime(sztime.filetime,2)  //filetime 7
            );
       except
       end;
   end; //for i:= 0

   Result := 0;


finally
   ina.Close;
   ResetCancel;
end;

end;

/////// Added MK 30.03.2006

function TSevenZip.Add: integer;
var
  updateCallback: TMyArchiveUpdateCallback;
  intf: IArchiveUpdateCallback;
  outStream: IOutStream;
  i,a,FileAttr, FtoAdd, fHnd: Integer;
  FMaxProgress:int64;
  FileDT:TFiletime;
  FileS:int64;
  setProperties:ISetProperties ;
  SetP: array[0..10] of PROPVARIANT;
  SetPNames: array[0..2] of PWideChar;

//Get compression strength for adding
     function SevenZipCompressionStrengthInt(cs: TCompressStrength): Cardinal;
     begin
        case cs of
         SAVE: result := 0;
         FAST: result := 3;
         NORMAL: result := 5;
         MAXIMUM: result := 7;
         ULTRA: result := 9;
         else result := 5;
       end;
     end;

//Get directory content and recursive if wanted
     procedure GetDirs(MainDir,Ext : widestring);
     var
      srw : _Win32_Find_Dataw;
      SearchHandle:cardinal;
      s : int64;
     begin
       try
        try
           srw.dwFileAttributes := faanyfile;
           SearchHandle := FindFirstfilew(PWideChar(MainDir + '*.*'),srw);
           if SearchHandle <> INVALID_HANDLE_VALUE then
            begin
             repeat
               if srw.cFileName[0] = '.' then continue;

               if (( srw.dwFileAttributes and fadirectory) <> 0 ) then
                 begin
                   if (AddRecurseDirs in FAddoptions) then getDirs(appendslash(MainDir+srw.cFileName),Ext);
                 end
               else
                 begin
                  if Ext <> '.*' then
                   begin
                    if extractfileextW(srw.cFileName) <> Ext then continue;
                   end;

                  Setlength(updateCallback.Files,FtoAdd+1);
                  Setlength(updateCallback.Files_size,FtoAdd+1);
                  Setlength(updateCallback.Files_Date,FtoAdd+1);
                  Setlength(updateCallback.Files_Attr,FtoAdd+1);

                  updateCallback.Files[FtoAdd] := MainDir + srw.cFileName;

                  LARGE_INTEGER(s).LowPart := srw.nFileSizeLow;
                  LARGE_INTEGER(s).HighPart := srw.nFileSizeHigh;
                  updateCallback.Files_size[FtoAdd] := s;

                  updateCallback.Files_Date[FtoAdd].dwLowDateTime := srw.ftLastWriteTime.dwLowDateTime;
                  updateCallback.Files_Date[FtoAdd].dwHighDateTime := srw.ftLastWriteTime.dwHighDateTime;
                  updateCallback.Files_Attr[FtoAdd] := srw.dwFileAttributes;

                  Fmaxprogress := Fmaxprogress + s;
                  inc(FtoAdd);
                end;
              until not FindNextfilew(SearchHandle,srw) or FMainCancel;
            end;
        except
          on E: Exception do raise;
        end;
       finally
         windows.FindClose(i);
       end;
     end;

begin
try
  updateCallback := TMyArchiveUpdateCallback.Create(self);

//Set Frootdir to uppercase for comparing
//Set AddRootdir for relative path or wholepath
//Set Frootdir to '' to add whole path

  Frootdir := UppercaseW_(FRootdir);
  updatecallback.RootDir := appendslash(FRootDir);

  FMaxProgress := 0;
  FtoAdd := 0;

  for i := 0 to ffiles.Count- 1 do
  begin
    a := pos('*',ffiles.WStrings[i]);
    if a > 0 then
      begin
        GetDirs(appendslash(copy(ffiles.WStrings[i],1,a-1)),copy(ffiles.WStrings[i],a+1,8));
      end
     else
      begin
       if fileexistsw(ffiles.WStrings[i]) then
        begin
         Setlength(updateCallback.Files,FtoAdd+1);
         Setlength(updateCallback.Files_size,FtoAdd+1);
         Setlength(updateCallback.Files_Date,FtoAdd+1);
         Setlength(updateCallback.Files_Attr,FtoAdd+1);

         updateCallback.Files[FtoAdd] := ffiles.Wstrings[i];

         GetFileSizeandDateTime_Int64(ffiles.Wstrings[i],FileS,FileDT,FileAttr);

         updateCallback.Files_size[FtoAdd] := FileS;
         updateCallback.Files_Date[FtoAdd] := FileDT;
         updateCallback.Files_Attr[FtoAdd] := FileAttr;

         FMaxProgress := FMaxProgress + FileS;
         inc(FtoAdd);
        end;
      end;
  end;

//send MaxProgress to App
  if assigned(OnPreProgress) then OnPreProgress(self,FMaxProgress);

// rg 6.4.06
  if ( (FSFXCreate) and (fileexistsw(FSFXModule)) ) then
   begin
    FSevenZipFileName := changefileextW(FSevenZipFileName,'.exe');
    if copyfilew(PWidechar(sfxmodule),PWideChar(FSevenZipFileName),True) then
      begin
        outStream := TMyStreamWriter.Create(FSevenZipFileName, now, FILE_ATTRIBUTE_ARCHIVE, fHnd,true)
      end
     else
      begin
       if assigned(onMessage) then onMessage(self,1,'Could not create SFX','');
      end;
   end
   else
    outStream := TMyStreamWriter.Create(FSevenZipFileName, now, FILE_ATTRIBUTE_ARCHIVE, fHnd);


//_______________
//Setp.vt := VT_EMPTY;
//Set archive options
  if outA.QueryInterface(IID_ISetProperties, setProperties) = S_OK then
  begin
//rg 17.04.06
   case FCompressType of
    LZMA: begin
            i := 0;

        // 7z Profile
            Setp[i].vt := VT_UI4;
            SetPNames[i] := StringToOleStr('X');
            Setp[i].ulVal := SevenZipCompressionStrengthInt(FCompstrength);
            inc(i);

        //Solid
            Setp[i].vt := VT_BSTR;
            SetPNames[i] := StringToOleStr('s');
            if (AddSolid in FAddoptions) then
              Setp[i].bstrVal := SysAllocString('on')
             else
              Setp[i].bstrVal := SysAllocString('off');
            inc(i);

        // directorysize 0..27
        //No need to set if you use CompressionStrength Profiles (Save...Ultra)
            if FLZMAStrength > 0 then
            begin
              Setp[i].vt := VT_UI4;
              SetPNames[i] := StringToOleStr('d');
              Setp[i].ulVal := FLZMAStrength;
              inc(i);
            end;
          end;
    PPMD: begin
            i := 0;

        // PPMD compression
            Setp[i].vt := VT_BSTR;
            SetPNames[i] := StringToOleStr('0');
            Setp[i].bstrVal := SysAllocString('PPMd');
            inc(i);

        //PPMD Size
        //No need to set if you use defaults
            if FPPMDsize > 0 then
            begin
              Setp[i].vt := VT_UI4;
              SetPNames[i] := StringToOleStr('o');
              Setp[i].ulVal := 10;
              inc(i);
            end;

        //PPMD Mem
        //No need to set if you use defaults
            if FPPMDmem > 0 then
            begin
              Setp[i].vt := VT_UI4;
              SetPNames[i] := StringToOleStr('mem');
              Setp[i].ulVal := 30;
              inc(i);
            end;

          end;
   end; //end case

 // set options
   result := setProperties.SetProperties(@SetPNames,@Setp,i);
  end; //if QuerryInterface
//____________________


   intf := updateCallback;

 if FtoAdd > 0 then
   begin
    result := outA.UpdateItems(outStream,length(updateCallback.Files),intf);
   end
  else
   begin
     if assigned(onMessage) then onMessage(self,1,'No files to add','');
     Result := -1;
   end;
finally
  ResetCancel;
end;
end;


function TSevenZip.Extract(TestArchive:Boolean=False): integer;
var
  updateCallback: TMyArchiveExtractCallback;
  ms: TMyStreamReader;
  filesDW: array of DWORD;
  Filestoex,w,ni: DWord;
  i,j: Integer;
  FMaxProgress:int64;
  size: PROPVARIANT;
begin
try
  ms := TMyStreamReader.Create(FSevenZipFileName);
  inA.Close;

// 24.08.06 - Matteo Riso - Status: experimental
// 25.08.06 Modified by rg
//
// If we loaded an .EXE file, we could start reading from offset 132096.
// Also supported by newerdll

//  if FIsSFX then ms.Seek(FSFXOffset,0,@FSFXOffset);

// End - MR modification

  i := inA.Open(ms, nil, nil);

  if i <> 0 then
  begin
    Result := -1;
    if assigned(onMessage) then onMessage(self,1,'File not found',Fsevenzipfilename);
    ms.Free;
    Exit;
  end;

  FMaxProgress := 0;
  inA.GetNumberOfItems(w); //1..end
  dec(w); //Starting with 0..end-1

  j := 0;
  if FFiles.Count > 0 then
  begin
    SetLength(filesDW, ffiles.Count);
    for i := 0 to FFiles.count - 1 do
    begin                     
      ni := strtoint(ffiles.Wstrings[i]);
      if ni > w then
      begin
        if assigned(onMessage) then onMessage(self,5,'Index out of Range','');
        Result := -1;
        exit;
      end;
      size.vt := VT_EMPTY;
      inA.GetProperty(ni, kpidSize, size);
      FMaxProgress := FMaxprogress + size.uhVal.QuadPart;
      filesDW[j] := ni;
      inc(j);
    end;
    Filestoex := length(filesDW);
  end
  else
  begin
// extract all files, FFiles.Count must be 0
   FilestoEx := $FFFFFFFF;
  end;

//set FMaxProgress for selected files
  if FMaxProgress > 0 then
   if assigned(OnPreProgress) then OnPreProgress(self,FMaxProgress);

// filesdw must be sorted asc
  if length(filesdw) > 1 then SortDWord(filesDW,low(filesdw),High(filesdw));

  updatecallback := TMyArchiveExtractCallback.Create(self);
  updatecallback.FExtractDirectory := appendslash(Fextrbasedir);
  updatecallback.FFilestoextract   := ffiles.Count; //with all files ffiles.count = 0, thats ok
  updatecallback.FAllFilesExt      := false;        //Stop extracting if no more files to extract
  updatecallback.FLastFileToExt    := false;        //only 1 more to extact


  result := inA.Extract(@filesDW[0], Filestoex, integer(TestArchive), updatecallback)

//  mmoLog.Lines.Add(Format('IInArchive.Extract: %d', [result]));
finally
  ina.close;
  ResetCancel;
end;
end;



(*
function TSevenZip.Delete: integer;
begin
//
end;

*)
//------------------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------------------
//-----------------End SevenZip VCL---------------------------------------------------------
//------------------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------------------

procedure Register;
begin
  RegisterComponents('Seven Zip', [TSevenZip]);
end;

end.

