#ifndef _FAT_H
#define _FAT_H

// On disk structures are always little endian.  They must be converted
// to CPU endian using a conversion function

#pragma pack(1)

typedef unsigned char BYTE;

typedef unsigned short U16;
typedef union
{
	BYTE data[2];
	U16 native;
} U16L;

#define L16TOH(x) (x.native)

typedef unsigned int U32;
typedef union
{
	BYTE data[4];
	U32 native;
} U32L;


#define L32TOH(x) (x.native)

typedef struct
{
	char BS_jmpBoot[3];
	char BS_OEMName[8];
	U16L BPB_BytsPerSec;
	BYTE BPB_SecPerClus;
	U16L BPB_RsvdSecCnt;
	BYTE BPB_NumFATs;
	U16L BPB_RootEntCnt;
	U16L BPB_TotSec16;
	BYTE BPB_Media;
	U16L BPB_FATSz16;
	U16L BPB_SecPerTrk;
	U16L BPB_NumHeads;
	U32L BPB_HiddSec;
	U32L BPB_TotSec32;
	union 
	{
		struct
		{
			BYTE BS_DrvNum;
			BYTE BS_Reserved1;
			BYTE BS_BootSig;
			U16L BS_VolID;
			char BS_VolLab[11];
			char BS_FilSysType[8];
		} fat;
		struct 
		{
			U32L BPB_FATSz32;
			U16L BPB_ExtFlags;
			U16L BPB_FSVer;
			U32L BPB_RootClus;
			U16L BPB_FSInfo;
			U16L BPB_BkBootSec;
			char BPB_Reserved[12];
			BYTE BS_DrvNum;
			BYTE BS_Reserved1;
			BYTE BS_BootSig;
			U32L BS_VolID;
			char BS_VolLab[11];
			char BS_FilSysType[8];
		} fat32;
	};
	char padding[420];
	U16L signature;
} fat_bpb_t;

#define BS_OEMName_default "MSWIN4.1"
#define SIGNATURE_VALUE	 0xAA55

#define	FAT12			1
#define	FAT16			2
#define	FAT32			3

#define FAT12_EOC 0x0FFF
#define FAT16_EOC 0xFFFF
#define FAT32_EOC 0x0FFFFFFF

#define FAT12_BAD 0x0FF7
#define FAT16_BAD 0xFFF7
#define FAT32_BAD 0x0FFFFFF7

char *fat_versions[] = 
{
	"Unknown",
	"FAT12",
	"FAT16",
	"FAT32"
};

typedef struct
{
	int fd;
	io_channel io;

	fat_bpb_t bpb;
	int version;
	int sector_size;
	int total_sectors;
	int fat_sectors;
	int reserved_sectors;
	int root_sector;
	int root_sectors;
	int data_sector;
	int clusters;
	int cluster_size;
	int eoc;
	int bad;

} fat_fs_t;

	
//-----------------------------------------------------------------------------
// FAT32 File Attributes and Types
//-----------------------------------------------------------------------------
#define FILE_ATTR_READ_ONLY   		0x01
#define FILE_ATTR_HIDDEN 		0x02
#define FILE_ATTR_SYSTEM 		0x04
#define FILE_ATTR_SYSHID		0x06
#define FILE_ATTR_VOLUME_ID 		0x08
#define FILE_ATTR_DIRECTORY		0x10
#define FILE_ATTR_ARCHIVE  		0x20
#define FILE_ATTR_LFN_TEXT		0x0F
#define FILE_HEADER_LAST		0x00
#define FILE_HEADER_DELETED		0xE5
#define FILE_HEADER_KANJI_E5		0x05
#define FILE_TYPE_DIR			0x10
#define FILE_TYPE_FILE			0x20
#define FILE_ORD_LFN_LAST		0x40
#define FILE_ORD_LFN_MAX		20

#define FILE_ATTR_MASK (FILE_ATTR_READ_ONLY | FILE_ATTR_HIDDEN | FILE_ATTR_SYSTEM | FILE_ATTR_VOLUME_ID | FILE_ATTR_DIRECTORY | FILE_ATTR_ARCHIVE)

typedef struct 
{
	BYTE name[11];
	BYTE attr;
	BYTE reserved;
	BYTE CrtTimeTenth;
	U16L CrtTime;
	U16L CrtDate;
	U16L LstAccDate;
	U16L cluster_hi;
	U16L WrtTime;
	U16L WrtDate;
	U16L cluster_lo;
	U32L size;
} fat_dirent_short_t;

typedef struct
{
	BYTE LDIR_Ord;
	U16L LDIR_Name1[5];
	BYTE LDIR_Attr;
	BYTE LDIR_Type;
	BYTE LDIR_Checksum;
	U16L LDIR_Name2[6];
	U16L LDIR_FstClusLO;
	U16L LDIR_Name3[2];
} fat_dirent_long_t;

typedef union
{
	fat_dirent_short_t s;
	fat_dirent_long_t l;
} fat_dirent_t;

typedef struct
{
	char *shortname;
	char *longname;
	U32 cluster;
	U32 mtime; // Wrt Date & Time
	U32 ctime; // Crt Date & Time
	U32 size;
	BYTE attr;
} fat_attr_t;

#pragma pack()

typedef struct
{
	unsigned short name[13];
	BYTE checksum;
	BYTE valid;
} fat_lfn_data_t;

typedef struct
{
	unsigned int directory;
	unsigned int file;
	unsigned int current; // this is the cluster number
	int pointer_cluster; // this is the count of clusters
	int size;
} fat_file_t;

typedef int dir_iterate_func(fat_attr_t *attr, void *priv);

int fatfs_open(char *name, io_manager manager, fat_fs_t **h);
int fatfs_close(fat_fs_t *fs);
int fatfs_dump(fat_fs_t *fs);
int fatfs_list(fat_fs_t *fs, unsigned int cluster, dir_iterate_func *callback, void *priv);
int fatfs_open_file(fat_fs_t *fs, unsigned int directory, unsigned int file, int mode, fat_file_t **h);
int fatfs_read_file(fat_fs_t *fs, fat_file_t *h, unsigned int offset, unsigned int len, char *buffer);
int fatfs_close_file(fat_fs_t *fs, fat_file_t *h);

#endif
