#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "ext2_io.h"
#include "fat.h"

int fatfs_dump(fat_fs_t *fs)
{
	printf("FAT version        %s\n", fat_versions[fs->version]);
	printf("total sectors      %d\n", fs->total_sectors);
	printf("sectors per fat    %d\n", fs->fat_sectors);
	printf("first data sector  %d\n", fs->data_sector);
	printf("number of clusters %d\n", fs->clusters);
	printf("cluster size       %d\n", fs->cluster_size);
	printf("BPB_BytsPerSec %d\n", L16TOH(fs->bpb.BPB_BytsPerSec));
	printf("BPB_SecPerClus %d\n", fs->bpb.BPB_SecPerClus);
	printf("BPB_RsvdSecCnt %d\n", L16TOH(fs->bpb.BPB_RsvdSecCnt));
	printf("BPB_NumFATs    %d\n", fs->bpb.BPB_NumFATs);
	printf("BPB_RootEntCnt %d\n", L16TOH(fs->bpb.BPB_RootEntCnt));
	printf("BPB_TotSec16   %d\n", L16TOH(fs->bpb.BPB_TotSec16));
	printf("BPB_FATSz16    %d\n", L16TOH(fs->bpb.BPB_FATSz16));
	printf("BPB_TotSec32   %d\n", L32TOH(fs->bpb.BPB_TotSec32));
	
	printf("0x%04X\n", L16TOH(fs->bpb.signature));
	
	// what cluster is the root directory?

	// make a function get_fat_entry()
	// make a function read_cluster()
	// make a function follow_cluster_chain()
	// make a function dump_directory()

	return 0;
}


int read_sector(fat_fs_t *fs, char *buffer, int sector, int count)
{
	int r;
	int len;

	if(fs->io)
	{
		return io_channel_read_blk(fs->io, sector, count, buffer);
	}
	else
	{
		r = lseek(fs->fd, sector * fs->sector_size, SEEK_SET);
		if(r == -1)
		{
			perror("seek_sector");
			return 1;
		}
		len = fs->sector_size * count;
		r = read(fs->fd, buffer, len);
		if(r != len)
		{
			perror("read_sector");
			return 1;
		}
		return 0;
	}
}

int get_fat_entry(fat_fs_t *fs, unsigned int cluster)
{
	int fat_sector;
	int fat_offset;
	unsigned int offset;
	int entries;
	unsigned char buffer[512 * 2]; // should be dynamic with size of sector
	unsigned short *wbuffer = (unsigned short *)buffer;
	unsigned int *dbuffer = (unsigned int *)buffer;
	int result;

	if(fs->version == FAT16 || fs->version == FAT32)
	{
		if(fs->version == FAT16)
		{
			offset = cluster * 2;
			entries = fs->sector_size / 2;
		}
		else
		{
			offset = cluster * 4;
			entries = fs->sector_size / 4;
		}

		fat_sector = offset / fs->sector_size;
		fat_sector+= fs->reserved_sectors;
		fat_offset = cluster % entries;
//printf("FAT entry offset = %d entries per buffer = %d\n", offset, entries);
//printf("FAT entry for cluster %8d at sector %8d offset %d\n", cluster, fat_sector, fat_offset);

		if(read_sector(fs, buffer, fat_sector, 1) != 0)
		{
			return 1;
		}

		if(fs->version == FAT16)
		{
			result = wbuffer[fat_offset];
		}
		else
		{
			result = dbuffer[fat_offset] & 0x0FFFFFFF;
		}
	}
	else
	{
		// Sometimes an entry will cross a sector boundry so we might need to read a second sector
		// 12 bits?  Bill has a lot to answer for.
		offset = cluster * 12; // in bits
		fat_sector = offset / (fs->sector_size * 8);
		fat_sector+= fs->reserved_sectors;
		fat_offset = cluster * 12 % (512 * 8); // in bits
//printf("FAT12 entry %d sector %d offset %d %d %d\n", cluster, fat_sector, fat_offset, fat_offset / 8, fat_offset % 8);
		if(fat_offset + 12 > (512 * 8))
		{
//printf("FAT12 entry crosses sector boundry\n");
			if(read_sector(fs, buffer, fat_sector, 2) != 0)
			{
				return 1;
			}
		}
		else
		{
			if(read_sector(fs, buffer, fat_sector, 1) != 0)
			{
				return 1;
			}
		}

		// find the 3 byte entry and pick which 12 bytes
//printf("%02x%02x%02x\n", buffer[fat_offset / 8], buffer[(fat_offset / 8) + 1], buffer[(fat_offset / 8) + 2]);
		if(fat_offset % 8 == 0)
		{
			fat_offset = fat_offset / 8;

			// byte = fat_offset / 8
			// +bits from (fat_offset / 8) + 1

			result = buffer[fat_offset];
			result|= ((buffer[fat_offset + 1] & 0x0F) << 8);
			result&= 0x0FFF;
		}
		else
		{
			fat_offset = fat_offset / 8;
			// bits from fat_offset / 8
			// + byte from (fat_offset / 8) + 1

			result = ((buffer[fat_offset] & 0xF0) >> 4);
			result|= buffer[fat_offset + 1] << 4;
			result&= 0x0FFF;
		}
//printf("FAT12 %8d(0x%04x) => %8d(0x%04x)\n\n", cluster, cluster, result, result);
		

	}
	return result;
}

void *read_cluster(fat_fs_t *fs, int cluster)
{
	void *data;
	int sector;

	data = malloc(fs->cluster_size);
	sector = (cluster - 2) * fs->bpb.BPB_SecPerClus + fs->data_sector;
//	printf("Reading cluster %d from sector %d\n", cluster, sector);

	if(read_sector(fs, data, sector, fs->bpb.BPB_SecPerClus) == 0)
	{
		return data;
	}
	else
	{
		free(data);
		return 0;
	}
}

void free_cluster(void *buffer)
{
	free(buffer);
}

void dump_cluster(fat_fs_t *fs, int cluster, int len)
{
	char *buffer = read_cluster(fs, cluster);
	if(buffer)
	{
		write(1, buffer, len);
		free_cluster(buffer);
	}
}

void dump_cluster_chain(fat_fs_t *fs, int cluster, int len)
{
	int remaining = len;
	int size;

	while(1)
{

	if(remaining > fs->cluster_size)
	{
		size = fs->cluster_size;
	}
	else
	{
		size = remaining;
	}

	dump_cluster(fs, cluster, size);
	remaining -= size;

	cluster = get_fat_entry(fs, cluster);

	if(cluster < 1 || cluster > fs->clusters)
	{
		if(remaining > 0)
		{
			printf("Missing clusters from file (%d bytes reamaining)\n", remaining);
		}
		break;
	}
	if(remaining == 0)
	{
		printf("File is too big for clusters\n");
		break;
	}
}
}

void follow_cluster_chain(fat_fs_t *fs, unsigned int cluster)
{
	unsigned int new_cluster;
	while(1)	
	{
		new_cluster = get_fat_entry(fs, cluster);
		printf("Cluster = %d -> %d\n", cluster, new_cluster);
		cluster = new_cluster;
		if(cluster > fs->clusters || cluster <= 1)
		{
			break;
		}
	}
}


// invalid chars are < 0x20, 22, 2A, 2B, 2C, 2E, 2F, 3A, 3B, 3C, 3D, 3E, 3F, 5B, 5C, 5D, 7C
// if name[0] == 0x05 then substitute with E5
void copy_short_name(fat_dirent_short_t *d, char *name)
{
	int i;
	int j;

	for(i = 0; i < 8; i++)
	{
		name[i] = d->name[i];
		if(name[i] != 0x20)
		{
			j = i + 1;
		}
	}
	if(d->name[8] != 0x20)
	{
		name[j++] = '.';
		name[j++] = d->name[8];
		if(d->name[9] != 0x20)
		{
			name[j++] = d->name[9];
			if(d->name[10] != 0x20)
			{
				name[j++] = d->name[10];
			}
		}
	}
	name[j++] = 0;

	if(name[0] == FILE_HEADER_KANJI_E5)
	{
		name[0] = FILE_HEADER_DELETED;
	}
}

void copy_long_name(fat_lfn_data_t *long_name, int long_name_len, char *longname)
{
	int i;
	int j;
	int l = 0;
	for(i = long_name_len; i > 0; i--)
	{
		for(j = 0; j < 13; j++)
		{
			if(long_name[i - 1].name[j] == 0x0000)
			{
				// skip rest of entry
				continue;
			}
			else if(long_name[i - 1].name[j] == 0xFFFF)
			{
				// skip this work (padding)
			}
			else
			{
				// for now we don't handle unicode...
				longname[l++] = long_name[i - 1].name[j];
			}
		}
	}
	longname[l++] = 0;
}

/*
int list_directory(fat_fs_t *fs, unsigned int cluster)
{
	void *buffer;
	fat_dirent_t *dir;
	int count;
	int d;
	char name[13];
	unsigned int file_cluster;
	int numfiles = 0;

	while(cluster - 2 < fs->clusters)
	{
		buffer = read_cluster(fs, cluster);
		if(buffer)
		{
			dir = (fat_dirent_t *)buffer;
			count = fs->cluster_size / sizeof(fat_dirent_t);
	
			for(d = 0; d < count; d++)
			{
				if(dir[d].s.name[0] == 0x00)
				{
					cluster = 0;
					break;
				}
				else if(dir[d].s.name[0] == 0xE5)
				{
				}
				else if((dir[d].s.attr & FILE_ATTR_MASK) == FILE_ATTR_LFN_TEXT)
				{
					// lfn entry
					
				}
				else
				{
					file_cluster = (L16TOH(dir[d].s.cluster_hi) << 16) | L16TOH(dir[d].s.cluster_lo);
					copy_short_name(dir + d, name);
					printf("0x%02x %-12s %8d %8d\n",dir[d].s.attr, name, L32TOH(dir[d].s.size), file_cluster);
					numfiles++;
	
					if((dir[d].s.attr & FILE_TYPE_DIR) && dir[d].s.name[0] != '.')
					{
						list_directory(fs, file_cluster);
					}
				}
			}
	
			free_cluster(buffer);
		}
		if(cluster > 0)
		{
			cluster = get_fat_entry(fs, cluster);
			printf("Following Cluster -> %d\n", cluster);
		}
	}
	printf("==EOD== %d files\n\n", numfiles);

	return 0;
}
*/

// don't call this directly, call fatfs_list on cluster 0 to get the root
int list_root(fat_fs_t *fs, dir_iterate_func *callback, void *priv)
{
	int i;
	int d;
	int ents = 512 / sizeof(fat_dirent_t);
	char buffer[512];
	fat_dirent_short_t *dir = (fat_dirent_short_t *)buffer;
	char name[13];
	fat_attr_t attr;
	
	if(fs->version != FAT32)
	{
		printf("Root is at %d len %d\n", fs->root_sector, fs->root_sectors);
		for(i = 0; i < fs->root_sectors; i++)
		{
			read_sector(fs, buffer, fs->root_sector + i, 1);
			for(d = 0; d < ents; d++)
			{
				if(dir[d].name[0] == FILE_HEADER_LAST)
				{
					return 0;
				}
				else if(dir[d].name[0] == FILE_HEADER_DELETED)
				{
				}
				else if((dir[d].attr & FILE_ATTR_MASK) == FILE_ATTR_LFN_TEXT)
				{
					// lfn entry
				}
				else if(dir[d].attr & FILE_ATTR_VOLUME_ID)
				{
					// volume name
				}
				else
				{
					attr.shortname = name;
					attr.longname = 0;
					attr.cluster = (L16TOH(dir[d].cluster_hi) << 16) | L16TOH(dir[d].cluster_lo);
					attr.size = L32TOH(dir[d].size);
					attr.attr = dir[d].attr;
					copy_short_name(dir + d, name);
					//printf("0x%02x %-12s %8d %8d\n",dir[d].attr, name, L32TOH(dir[d].size), file_cluster);
					if(callback(&attr, priv) > 0)
					{
						// we don't need to search any more
						return 0;
					}
				}
			}
		}
	}
	return 0;
}

BYTE short_name_checksum(BYTE *name)
{
	int i;
	BYTE sum = 0;
	for(i = 0; i < 11; i++)
	{
		sum = ((sum & 1) ? 0x80 : 0) + (sum >> 1) + name[i];
	}
	return sum;
}

int fatfs_list(fat_fs_t *fs, unsigned int cluster, dir_iterate_func *callback, void *priv)
{
	void *buffer;
	fat_dirent_short_t *dir;
	fat_dirent_long_t *lfn;
	int count;
	int d;
	char name[13];
	fat_attr_t attr;
	fat_lfn_data_t long_name[FILE_ORD_LFN_MAX];
	int long_name_len;
	char longname[512];

	if(cluster == 0)
	{
		// list the root
		if(fs->version == FAT32)
		{
			cluster = L32TOH(fs->bpb.fat32.BPB_RootClus);
		}
		else
		{
			return list_root(fs, callback, priv);
		}
	}


	if(cluster > 0)
	{
		while(cluster - 2 < fs->clusters)
		{
			buffer = read_cluster(fs, cluster);
			if(buffer)
			{
				dir = (fat_dirent_short_t *)buffer;
				count = fs->cluster_size / sizeof(fat_dirent_t);

				// decode this cluster worth of data
				dir = (fat_dirent_short_t *)buffer;
				count = fs->cluster_size / sizeof(fat_dirent_t);
		
				for(d = 0; d < count; d++)
				{
					if((dir[d].attr & FILE_ATTR_MASK) == FILE_ATTR_LFN_TEXT)
					{
						// lfn entry
						//printf("LFN Entry at %d\n", d);
						lfn = (fat_dirent_long_t *)dir + d;
						//printf("Ord = 0x%02x checksum = 0x%02x\n", lfn->LDIR_Ord, lfn->LDIR_Checksum);
						if(lfn->LDIR_Ord & FILE_ORD_LFN_LAST)
						{
							//printf("Start of LFN\n");
							long_name_len = 0;
						}
						if(long_name_len < FILE_ORD_LFN_MAX)
						{
							long_name[long_name_len].valid = 1;
							long_name[long_name_len].name[0]  = L16TOH(lfn->LDIR_Name1[0]);
							long_name[long_name_len].name[1]  = L16TOH(lfn->LDIR_Name1[1]);
							long_name[long_name_len].name[2]  = L16TOH(lfn->LDIR_Name1[2]);
							long_name[long_name_len].name[3]  = L16TOH(lfn->LDIR_Name1[3]);
							long_name[long_name_len].name[4]  = L16TOH(lfn->LDIR_Name1[4]);
							long_name[long_name_len].name[5]  = L16TOH(lfn->LDIR_Name2[0]);
							long_name[long_name_len].name[6]  = L16TOH(lfn->LDIR_Name2[1]);
							long_name[long_name_len].name[7]  = L16TOH(lfn->LDIR_Name2[2]);
							long_name[long_name_len].name[8]  = L16TOH(lfn->LDIR_Name2[3]);
							long_name[long_name_len].name[9]  = L16TOH(lfn->LDIR_Name2[4]);
							long_name[long_name_len].name[10] = L16TOH(lfn->LDIR_Name2[5]);
							long_name[long_name_len].name[11] = L16TOH(lfn->LDIR_Name3[0]);
							long_name[long_name_len].name[12] = L16TOH(lfn->LDIR_Name3[1]);
							long_name[long_name_len].checksum = lfn->LDIR_Checksum;
							if(long_name_len > 0)
							{
								if(long_name[0].checksum != lfn->LDIR_Checksum)
								{
									long_name[0].valid = 0;
								}
							}
							long_name_len++;
						}
						else
						{
							long_name[0].valid = 0;
						}
					}
					else if(dir[d].name[0] == FILE_HEADER_LAST)
					{
						// no more entries
						cluster = 0;
						break;
					}
					else if(dir[d].name[0] == FILE_HEADER_DELETED)
					{
						// this entry is deleted
					}
					else if(dir[d].attr & FILE_ATTR_VOLUME_ID)
					{
						// volume name
					}
					else
					{
						attr.shortname = name;
						attr.longname = 0;
						attr.cluster = (L16TOH(dir[d].cluster_hi) << 16) | L16TOH(dir[d].cluster_lo);
						attr.size = L32TOH(dir[d].size);
						attr.attr = dir[d].attr;
						copy_short_name(dir + d, name);

						if(long_name_len > 0 && long_name[0].valid)
						{
							// validate the LFN entry...
							if(long_name[0].checksum == short_name_checksum(dir[d].name))
							{
								// matches, now UTF encode the string...
								copy_long_name(long_name, long_name_len, longname);
								//printf("Decode LFN to '%s'\n", longname);
								attr.longname = longname;
							}
						}
						//printf("0x%02x %-12s %8d %8d\n",dir[d].attr, name, L32TOH(dir[d].size), file_cluster);
						if(callback(&attr, priv) > 0)
						{
							// we don't need to search any more
							cluster = 0;
							break;
						}
					}
				}

				free_cluster(buffer);
			}
			if(cluster > 0)
			{
				cluster = get_fat_entry(fs, cluster);
				//printf("Following Cluster -> %d\n", cluster);
			}
		}
	}
	return 0;
}


//int open_fs(fat_fs_t *fs, char *name)
int fatfs_open(char *name, io_manager manager, fat_fs_t **h)
{
	int is_open = 0;
	io_channel io = 0;
	int fd = -1;
	fat_fs_t *fs;

	if(manager)
	{
		if(manager->open(name, 0, &io) == 0)
		{
			is_open = 1;
		}
	}
	else
	{
		fd = open(name, O_RDONLY);
		if(fd >= 0)
		{
			is_open = 1;
		}
	}

	if(is_open)
	{
		fs = malloc(sizeof(fat_fs_t));
		*h = fs;
		fs->io = io;
		fs->sector_size = 512;
		if(io)
		{
			io_channel_set_blksize(fs->io, fs->sector_size);
		}
		fs->fd = fd;

		if(read_sector(fs, (void *)&fs->bpb, 0, 1) != 0)
		{
			perror("read");
			return -1;
		}

		if(L16TOH(fs->bpb.signature) != SIGNATURE_VALUE)
		{
			printf("FAT Signature not found\n");
			//close(fs->h);
			return -1;
		}

		// number of sectors
		if(L16TOH(fs->bpb.BPB_TotSec16) > 0)
		{
			fs->total_sectors = L16TOH(fs->bpb.BPB_TotSec16);
		}
		else
		{
			fs->total_sectors = L32TOH(fs->bpb.BPB_TotSec32);
		}

		if(L16TOH(fs->bpb.BPB_FATSz16) > 0)
		{
			fs->fat_sectors = L16TOH(fs->bpb.BPB_FATSz16);
		}
		else
		{
			fs->fat_sectors = L32TOH(fs->bpb.fat32.BPB_FATSz32);
		}

		fs->sector_size = L16TOH(fs->bpb.BPB_BytsPerSec);
		fs->cluster_size = fs->bpb.BPB_SecPerClus * fs->sector_size;
		fs->reserved_sectors = L16TOH(fs->bpb.BPB_RsvdSecCnt);
	
		fs->root_sector = fs->reserved_sectors + (fs->bpb.BPB_NumFATs * fs->fat_sectors);
	
		fs->root_sectors = ((L16TOH(fs->bpb.BPB_RootEntCnt) * 32) + (L16TOH(fs->bpb.BPB_BytsPerSec) - 1)) / L16TOH(fs->bpb.BPB_BytsPerSec);
		fs->data_sector = fs->root_sector + fs->root_sectors;
		fs->clusters = (fs->total_sectors - fs->data_sector) / fs->bpb.BPB_SecPerClus;

		if(fs->clusters < 4085)
		{
			fs->version = FAT12;
			fs->eoc = FAT12_EOC;
			fs->bad = FAT12_BAD;
		}
		else if(fs->clusters < 65525)
		{
			fs->version = FAT16;
			fs->eoc = FAT16_EOC;
			fs->bad = FAT16_BAD;
		}
		else
		{
			fs->version = FAT32;
			fs->eoc = FAT32_EOC;
			fs->bad = FAT32_BAD;
		}
	}
	else
	{
		return -2;
	}

	return 0;
}

int fatfs_close(fat_fs_t *fs)
{
	return 0;
}

int find_dir_entry(fat_attr_t *attr, void *priv)
{
	fat_file_t *f = (fat_file_t *)priv;

	if(attr->cluster == f->file)
	{
		f->size = attr->size;
		return 2;
	}
	return 0;
}

int fatfs_open_file(fat_fs_t *fs, unsigned int directory, unsigned int file, int mode, fat_file_t **h)
{
	fat_file_t *f;

	// we need to know the file size... that only comes from the directory listing

	// we should do lots of sanity checks and link everything to the fs.
	// that can get added later as we implement write support
	// for now, keep it simple and get it working

	f = (fat_file_t *)malloc(sizeof(fat_file_t));

	f->directory = directory;
	f->file = file;
	f->current = file;
	f->pointer_cluster = 0;
	f->size = -1; // lookup the dir...
	fatfs_list(fs, directory, find_dir_entry, f);

	if(f->size != -1)
	{
		// found it
		*h = f;
		return 0;
	}
	else
	{
		free(f);
		return 1;
	}

}

int fatfs_read_file(fat_fs_t *fs, fat_file_t *h, unsigned int offset, unsigned int len, char *buffer)
{
	int remaining = len;
	int start_cluster;
	int cluster_offset;
	int cluster_len;
	char *cluster_buffer;

	//printf("Reading %d -> %d\n", offset, offset + len);

	// sanity and bounds check
	if(offset > h->size)
	{
		return 1;
	}
	if(offset + len > h->size)
	{
		return 1;
	}

	while(remaining > 0)
	{
		start_cluster = offset / fs->cluster_size;

		if(start_cluster < h->pointer_cluster)
		{
			h->pointer_cluster = 0;
			h->current = h->file;
		}


		//printf("Start reading in cluster %d\n", start_cluster);
		// seek forward to the correct cluster
		while(h->pointer_cluster < start_cluster)
		{
			h->current = get_fat_entry(fs, h->current);
			h->pointer_cluster++;
			//printf("We are now at %d cluster %d\n", h->pointer_cluster, h->current);
		}
		cluster_buffer = read_cluster(fs, h->current);
		if(!cluster_buffer)
		{
			return 2;
		}

		cluster_offset = offset - (h->pointer_cluster * fs->cluster_size);
		// adjust end so it does not cross cluster boundry
		if(cluster_offset + remaining <= fs->cluster_size)
		{
			cluster_len = remaining;
		}
		else
		{
			cluster_len = fs->cluster_size - cluster_offset;
		}
		
		// copy the buffer
		memcpy(buffer, cluster_buffer + cluster_offset, cluster_len);
		offset+= cluster_len;
		buffer+= cluster_len;
		remaining-= cluster_len;
		
		free_cluster(cluster_buffer);
	}	
	return 0;
}

int fatfs_close_file(fat_fs_t *fs, fat_file_t *h)
{
	free(h);
	return 0;
}



