unit libreiser3;

interface

uses windows, types, vv_io;

type
   PReiserFS = Pointer;
   PPReiserFS = ^PReiserFS;

   REISERFS_STAT2 = packed record
    sd_mode       : U16;	// file type, permissions */
    sd_reserved   : U16;
    sd_nlink      : U32;	// number of hard links */
    sd_size       : U64;	// file size */
    sd_uid        : U32;	// owner */
    sd_gid        : U32;	// group */
    sd_atime      : U32;	// time of last access */
    sd_mtime      : U32;	// time file was last modified  */
    sd_ctime      : U32;	// time inode (stat data) was last changed (except changes to sd_atime and sd_mtime) */
    sd_blocks     : U32;
	 sd_rdev       : U32;
   end;

   PREISERFS_STAT2 = ^REISERFS_STAT2;

   REISERFS_CPU_KEY = packed record
    k_dir_id   : U32;    // packing locality: by default parent directory object id */
    k_objectid : U32;  // object identifier */
    k_offset   : U64;
    k_type     : U32;
   end;

   PREISERFS_CPU_KEY = ^REISERFS_CPU_KEY;

   REISERFS_READ_INFO = packed record
    Offset : U64;
    Length : U64;
    Complete : U64;
    Buffer : PChar;
   end;

   PREISERFS_READ_INFO = ^REISERFS_READ_INFO;


function reiserfs_open(name : PChar; io_manager : Pio_manager; pfs : PPReiserFS) : errcode_t; cdecl;
function reiserfs_close(fs : PReiserFS) : errcode_t; cdecl;

type dir_iterate_func = function (Name : PChar; Key : PREISERFS_CPU_KEY; priv_data : Pointer) : Integer; cdecl;
function reiserfs_list(fs : PReiserFS; path : PChar; func : dir_iterate_func; priv_data : Pointer) : errcode_t; cdecl;
function reiserfs_list_key(fs : PReiserFS; key : PREISERFS_CPU_KEY; func : dir_iterate_func; priv_data : Pointer) : errcode_t; cdecl;
function reiserfs_stat_key(fs : PReiserFS; key : PREISERFS_CPU_KEY; stat_data : PREISERFS_STAT2) : errcode_t; cdecl;
function reiserfs_read_key(fs : PReiserFS; key : PREISERFS_CPU_KEY; info : PREISERFS_READ_INFO) : errcode_t; cdecl;

function UuidToStr(uuid : String) : String;

implementation

uses sysutils;

function UuidToStr(uuid : String) : String;
var
   i : Integer;
begin
   Result := '';
   for i := 1 to 16 do
   begin
      Result := Result + IntToHex(Ord(uuid[i]), 2);
      if i in [4, 6, 8, 10] then
      begin
         Result := Result + '-';
      end;
   end;
end;


const DLL='librfs.dll';

function reiserfs_open; external DLL;
function reiserfs_close; external DLL;
function reiserfs_list; external DLL;
function reiserfs_list_key; external DLL;
function reiserfs_stat_key; external DLL;
function reiserfs_read_key; external DLL;


end.
