unit NetLib;
// $Header: /home/cso/jnewbigin/cvsroot/vv/src/NetLib.pas,v 1.1 2006/08/19 06:04:37 jnewbigin Exp $

interface

uses WinSock, SysUtils, Windows, Classes, Dialogs;

function recv(s: TSocket; Buf : Pointer; len, flags: Integer): Integer; stdcall;
function send(s: TSocket; Buf : Pointer; len, flags: Integer): Integer; stdcall;

const
   SD_RECEIVE = $00;
   SD_SEND    = $01;
   SD_BOTH    = $02;


type

  TSocketException = class(Exception)
  public
     GetLastError : Integer;
     constructor Create(const Msg: string);
  end;

  TConnectionEvent = procedure (Sender: TObject; Socket : TSocket; address : TSockAddr) of object;

{  Threadable = class
     procedure Thread; virtual;
  end;}

  TNetServer = class //(Threadable)
  private
    WSAData : TWSAData;
    fd      : TSocket;
    addrIn  : TSockAddrIn;
    address : TInAddr;

    AcceptThread      : THandle;
    ConnectionThreads : TList;

    StopAccept    : Boolean;


    OnConnect : TConnectionEvent;
    Backlog    : Integer;

    procedure Thread;

    procedure StartAcceptThread;
  public
    constructor Create(Port : Integer);
    destructor Destroy; override;

    procedure Enable(Event : TConnectionEvent);
    procedure StopAcceptThread;
    function DoAccept : TSocket;

  end;

  TNetEndpoint = class //(Threadable)
  private
    WSAData : TWSAData;
    fd      : TSocket;
    addrIn  : TSockAddrIn;
    address : TInAddr;

    ThreadHandle   : THandle;
    StopThread : Boolean;

    Buffer : String;
    BufferStart : Integer;
    BufferLen : Integer;

  protected
    procedure Thread; virtual;

    procedure FillBuffer;


  public
    constructor CreateServer(fd : TSocket); // accept a server connection
    constructor Create(Host : String; Port : Integer); // create a client
    destructor Destroy; override;

    procedure StartNetThread;
    procedure StopNetThread;

    procedure Shutdown(how : Integer);
    function GetSocket : Integer;

    procedure BlockingWait(p : Pointer; Len : integer);

    procedure SendBuffer(p : Pointer; len : Integer);
    procedure RecvBuffer(p : Pointer; len : Integer);

    procedure SendInteger(i : Integer);
    function  RecvInteger : Integer;

    procedure SendBoolean(b : Boolean);
    function  RecvBoolean : Boolean;

    procedure SendInt64(i : Int64);
    function  RecvInt64 : Int64;

    procedure SendSingle(i : Single);
    function  RecvSingle : Single;

    procedure SendByte(b : Byte);
    function  RecvByte : Byte;

    procedure SendString(S : String);
    function  RecvString : String;

    procedure SendStringZ(S : String);
    function  RecvStringZ : String;

    procedure SendStringN(S : String);
    function  RecvStringN : String;
  end;

function StartThread(P : Pointer) : Integer;


implementation

uses debug;

const
  winsocket = 'wsock32.dll';

function send; external    winsocket name 'send';
function recv; external    winsocket name 'recv';

constructor TSocketException.Create(const Msg: string);
begin
   GetLastError := WSAGetLastError;
   inherited Create('SocketError ' + IntToStr(GetLastError) + ' ' + Msg);
end;

function StartThread(P : Pointer) : Integer;
var
   Obj : TObject;
begin
   // One day I will get this to work with an inherited thread object...

   Result := 1;
   Obj := TObject(P);
   if Obj.ClassType = TNetServer then
   begin
      TNetServer(Obj).Thread;
   end
   else {if Obj.ClassType = NetEndpoint then}
   begin
      TNetEndpoint(Obj).Thread;
   end;

//   Threadable(TObject(P)).Thread;
end;

constructor TNetServer.Create(Port : Integer);
var
   block : u_Long;
begin
   inherited Create;

   WSAStartup($101, WSAData);

   backlog := 1;

   address.S_addr := INADDR_ANY;

   addrIn.sin_family  := AF_INET;
   addrIn.sin_port    := htons(Port);
   addrIn.sin_addr    := address;

   fd := socket(addrIn.sin_family, SOCK_STREAM, 0);
   if fd = -1 then
   begin
      raise TSocketException.Create('calling socket');
   end;

   if bind(fd, addrIn, sizeof(addrIn)) < 0 then
   begin
      raise TSocketException.Create('calling bind');
   end;

   block := 0;
   ioctlsocket(fd, FIONBIO, block);

   // you now need to call Enable to start getting connections...
end;

procedure TNetServer.Enable(Event : TConnectionEvent);
begin
   listen(fd, backlog); // backlog....?

   // spawn a thread to accept connections.....
   if Assigned(Event) then
   begin
      OnConnect := Event;
      StartAcceptThread;
   end;
end;


destructor TNetServer.Destroy;
begin
   WSACleanUp;
   inherited Destroy;
end;

procedure TNetServer.StartAcceptThread;
begin
   if AcceptThread = 0 then
   begin
      StopAccept := False;
      BeginThread(nil, 0, StartThread, self, 0, AcceptThread);
   end;
end;

procedure TNetServer.StopAcceptThread;
begin
   StopAccept := True;
end;

procedure TNetServer.Thread;
begin
   while not StopAccept do
   begin
      DoAccept;
   end;
end;

function TNetServer.DoAccept : TSocket;
var
   addr     : TSockAddr;
   addrlen  : Integer;
   newfd    : TSocket;
   Endpoint : TNetEndpoint;
   optval   : Integer;
   block    : u_long;
begin
   addrlen  := sizeof(addr);
   newfd    := accept(fd, PSOCKADDR(@addr), @addrlen);
   if newfd < 0 then
   begin
      raise TSocketException.Create('calling accept');
   end;
{   optval := 1;
   if setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, @optval, sizeof(optval)) = SOCKET_ERROR then
   begin
      raise SocketException.Create('setsockopt');
   end;}

   // create a new thread


   // here we call the nominated function.....
   if Assigned(OnConnect) then
   begin
      block := 0;
      ioctlsocket(newfd, FIONBIO, block);


      OnConnect(self, newfd, addr);
   end;

//   Endpoint := NetEndpoint.CreateServer(newfd);
//   Endpoint.StartNetThread;
   Result := newfd;

end;

////////////////
//
// endpoint bits
//
////////////////

constructor TNetEndpoint.Create(Host : String; Port : Integer);
var
   HostEnt : PHostEnt;
   p : pointer;
begin
   inherited Create;
   ThreadHandle := 0;
   // we are a client wanting to connect to a server
   WSAStartup($101, WSAData);

   // find the host address

   address.S_addr := inet_addr(PChar(host));
   if address.S_addr = -1 {INADDR_NONE} then
   begin
      HostEnt := gethostbyname(PChar(Host));
      if not assigned(HostEnt) then
      begin
         raise TSocketException.Create('calling gethostbynanm');
      end;
      address.S_addr := longint(plongint(HostEnt^.h_addr_list^)^);
   end;

   addrIn.sin_family  := AF_INET;
   addrIn.sin_port    := htons(Port);
   addrIn.sin_addr    := address;

   fd := socket(addrIn.sin_family, SOCK_STREAM, 0);
   if connect(fd, addrIn, sizeof(addrIn)) < 0 then
   begin
      raise TSocketException.Create('calling connect');
   end;

   SetLength(Buffer, 1024);
   BufferStart := 0;
   BufferLen := 0;
end;

constructor TNetEndpoint.CreateServer(fd : TSocket);
begin
   inherited Create;
   // we are a server, conneted to a client
   self.fd := fd;

   SetLength(Buffer, 1024);
   BufferStart := 0;
   BufferLen := 0;
end;

destructor TNetEndpoint.Destroy;
begin
   closesocket(fd);
{   if ThreadHandle > 0 then
   begin
      StopNetThread;
      Sleep(1000);
      if ThreadHandle > 0 then
      begin
         TerminateThread(ThreadHandle, -1);
      end
   end;}
   inherited Destroy;
//   Debug('endpoint destroyed');
end;

procedure TNetEndpoint.Shutdown(how : Integer);
begin
   winsock.shutdown(fd, how);
end;

procedure TNetEndpoint.Thread;
var
   S : String;
   Len : Integer;
begin
   try
      try
         while not StopThread do
         begin
//            s := RecvString;
//            SendString(S);
//            Form1.Memo1.Lines.Add(S);
            Len := RecvInteger;
            SendInteger(Len);
//            Debug(IntToStr(Len));
         end;
      finally
         ThreadHandle := 0;
      end;
   except
      on E : TSocketException do
      begin
         if E.GetLastError = WSAECONNABORTED then
         begin
            self.Free;
            ExitThread(WSAECONNABORTED);
         end
         else
         begin
            MessageDlg(E.Message, mtError, [mbOK], 0);
            self.Free;
         end;
         ExitThread(E.GetLastError);
      end;
   end;
//   Debug('Thread stop');
   self.Free;
//   ExitThread(0);
end;

procedure TNetEndpoint.StartNetThread;
begin
   if ThreadHandle = 0 then
   begin
      StopThread := False;
      BeginThread(nil, 0, StartThread, self, 0, ThreadHandle);
   end;
end;

procedure TNetEndpoint.StopNetThread;
begin
   StopThread := True;
end;

procedure TNetEndpoint.SendBuffer(p : Pointer; len : Integer);
var
   Sent : Integer;
begin
   Sent := send(fd, p, Len, 0);
   if Sent = -1 then
   begin
      raise TSocketException.Create('send');
   end
   else if Sent < len then
   begin
      raise TSocketException.Create('short send');
   end
   else
   begin
      //Log('sent ' + IntToStr(len) + ' bytes');
   end;
end;

procedure TNetEndpoint.FillBuffer;
var
   recved : Integer;
   Available : Integer;
begin
   if BufferLen = 0 then
   begin
      BufferStart := 0;
   end;
   Available := Length(Buffer) - BufferStart - BufferLen;
   if Available > 0 then
   begin
      recved := recv(fd, PChar(Buffer), Available, 0);
      if recved > 0 then
      begin
         BufferLen := BufferLen + recved;
      end;
   end;

end;

procedure TNetEndpoint.RecvBuffer(p : Pointer; len : Integer);
var
   recved      : Integer;
   Done        : Integer;
   Available   : Integer;
begin
//   BlockingWait(p, len);
   Done := 0;

   if len <= BufferLen then
   begin
      CopyMemory(p, PChar(Buffer) + BufferStart, len);
      BufferStart := BufferStart + len;
      BufferLen := BufferLen - len;
      exit;
   end;

   while Done < len do
   begin
      FillBuffer;
      Available := BufferLen;
      if Available > (len - done) then
      begin
         Available := len - done;
      end;
      CopyMemory(PChar(p) + Done, PChar(Buffer) + BufferStart, Available);
      BufferStart := BufferStart + Available;
      BufferLen := BufferLen - Available;
      Done := Done + Available;

   end;
   {

   recved := recv(fd, p, Len, 0);
   if recved = -1 then
   begin
      raise TSocketException.Create('recv');
   end
   else if recved < len then
   begin
      raise TSocketException.Create('short read');
   end
   else
   begin
      Log('received ' + IntTOStr(len) + ' bytes');
   end;}
end;

procedure TNetEndpoint.BlockingWait(p : Pointer; Len : integer);
var
   aLen : Integer;
begin
   while true do
   begin
      aLen := recv(fd, p, Len, MSG_PEEK);
      if aLen = -1 then
      begin
         raise TSocketException.Create('recv');
      end;
      if aLen >= Len then break;
      sleep(100);
   end;

end;

procedure TNetEndpoint.SendInteger(i : Integer);
begin
   SendBuffer(@i, sizeof(i));
end;

function  TNetEndpoint.RecvInteger : Integer;
begin
   RecvBuffer(@Result, sizeof(Result));
end;

procedure TNetEndpoint.SendBoolean(b : Boolean);
begin
   if b then
   begin
      SendByte(1);
   end
   else
   begin
      SendByte(0);
   end;
end;

function  TNetEndpoint.RecvBoolean : Boolean;
begin
   if RecvByte = 0 then
   begin
      Result := False;
   end
   else
   begin
      Result := True;
   end;
end;

procedure TNetEndpoint.SendInt64(i : Int64);
begin
   SendBuffer(@i, sizeof(i));
end;

function  TNetEndpoint.RecvInt64 : Int64;
begin
   RecvBuffer(@Result, sizeof(Result));
end;

procedure TNetEndpoint.SendSingle(i : Single);
begin
   SendBuffer(@i, sizeof(i));
end;

function  TNetEndpoint.RecvSingle : Single;
begin
   RecvBuffer(@Result, sizeof(Result));
end;

procedure TNetEndpoint.SendByte(b : Byte);
begin
   SendBuffer(@b, sizeof(b));
end;

function TNetEndpoint.RecvByte : Byte;
begin
   RecvBuffer(@Result, sizeof(Result));
end;

procedure TNetEndpoint.SendString(S : String);
begin
   SendInteger(Length(S));
   if Length(S) > 0 then
   begin
      SendBuffer(PChar(S), Length(S));
   end;
end;

function  TNetEndpoint.RecvString : String;
var
   Len : integer;
begin
   Len := RecvInteger;
   SetLength(Result, Len);
   if Len > 0 then
   begin
      RecvBuffer(PChar(Result), Len);
   end;
end;

procedure TNetEndpoint.SendStringZ(S : String);
begin
   S := S + #0;
   SendBuffer(PChar(S), Length(S));
end;

function  TNetEndpoint.RecvStringZ : String;
var
   c : Char;
begin
   Result := '';
   while True do
   begin
      c := Char(RecvByte);
      if c = #0 then break;
      Result := Result + c;
   end;
end;

procedure TNetEndpoint.SendStringN(S : String);
begin
   //Log('Sending ' + S);
   S := S + #13#10;
   SendBuffer(PChar(S), Length(S));
end;

function  TNetEndpoint.RecvStringN : String;
var
   c : Char;
begin
   Result := '';
   while True do
   begin
      c := Char(RecvByte);
      if c = #13 then
      begin
         // ignore
      end
      else if c = #10 then
      begin
         break;
      end
      else
      begin
         Result := Result + c;
      end;
   end;
end;

function TNetEndpoint.GetSocket : Integer;
begin
   Result := fd;
end;

end.
