unit debug;

{$IFDEF FPC}
{$MODE Delphi}
{$ENDIF}

// $Header: /home/cso/jnewbigin/cvsroot/vv/src/debug.pas,v 1.6 2007/03/25 07:21:25 jnewbigin Exp $

interface

uses Windows;

const LOG_ERROR   = 0;
const LOG_NORMAL  = 1;
const LOG_VERBOSE = 2;
const LOG_LOOP    = 3;
const LOG_DUMP    = 4;
const LOG_MAX     = LOG_DUMP;

procedure Log(S : String); overload;
procedure LogW(S : WideString);
procedure Log(S : WideString; Facility : String; Level : Integer = LOG_VERBOSE); overload;
type DebugEvent = procedure (S : String) of object;
type DebugEventW = procedure (S : WideString) of object;
procedure SetDebug(d : DebugEvent);
procedure SetDebugW(d : DebugEventW);
procedure UseWriteln;
procedure SetLogLevel(Facility : String; Level : Integer);

function IsDebuggerPresent : BOOL;

type TWriteLine = class
public
   procedure WriteLine(S : String);
end;

implementation

uses classes;

type TIsDebuggerPresent = function : BOOL; stdcall;

var
   fOnDebug : DebugEvent;
   fOnDebugW : DebugEventW;
   LogLevels : TStringList;
   DefaultLevel : Integer;

procedure Log(S : String);
begin
   if Assigned(fOnDebug) then
   begin
      fOnDebug(s);
   end
   else if Assigned(fOnDebugW) then
   begin
      fOnDebugW(S);
   end;
end;

procedure LogW(S : WideString); overload;
begin
   if Assigned(fOnDebugW) then
   begin
      fOnDebugW(s);
   end
   else if Assigned(fOnDebug) then
   begin
      fOnDebug(S);
   end;
end;

function GetLogLevel(Facility : String) : Integer;
var
   i : Integer;
begin
   Result := DefaultLevel;

   if Assigned(LogLevels) then
   begin
      i := LogLevels.IndexOf(Facility);
      if i >= 0 then
      begin
         Result := Integer(LogLevels.Objects[i]);
      end;
   end;
end;

procedure SetLogLevel(Facility : String; Level : Integer);
var
   i : Integer;
begin
   if Facility = '*' then
   begin
      DefaultLevel := Level;
   end
   else
   begin
      if not Assigned(LogLevels) then
      begin
         LogLevels := TStringList.Create;
      end;
      i := LogLevels.IndexOf(Facility);
      if i >= 0 then
      begin
         LogLevels.Objects[i] := TObject(Level);
      end
      else
      begin
         LogLevels.AddObject(Facility, TObject(Level));
      end;
   end;
end;

// 0 - Error
// 1 - Normal
// 2 - More
// 3 - Loops
// 4 - Dumps
procedure Log(S : WideString; Facility : String; Level : Integer); overload;
var
   LogLevel : Integer;
begin
   // We can have per Facility log levels...
   LogLevel := GetLogLevel(Facility);

   if Level <= LogLevel then
   begin
      LogW('[' + Facility + '] ' + S);
   end;
end;

procedure SetDebug(d : DebugEvent);
begin
   fOnDebug := d;
end;

procedure SetDebugW(d : DebugEventW);
begin
   fOnDebugW := d;
end;

procedure TWriteLine.WriteLine(S : String);
begin
{$IFDEF FPC}
   Writeln(S + #13);
{$ELSE}
   Writeln(S);
{$ENDIF}
end;

procedure UseWriteln;
var
   wl : TWriteLine;
begin
   wl := TWriteLine.Create;
   SetDebug(wl.WriteLine);
end;

function IsDebuggerPresent : BOOL;
var
   hModule : hInst;
//   Error : DWORD;
   JIsDebuggerPresent : TIsDebuggerPresent;
   P : Pointer;
begin
   Result := False;

   // see if we can get IsDebuggerPresent from kernel32.dll

   hModule := GetModuleHandle('kernel32.dll');
   if hModule = 0 then
   begin
      // wininet is not yet loaded...
      hModule := LoadLibrary('kernel32.dll');
      {if hModule = 0 then
      begin
         Error := GetLastError;
         //raise Exception.Create('Error loading Windows Internet Library.  ' +
SysErrorMessage(Error));
      end;}
   end;

   if hModule <> 0 then
   begin
      P := GetProcAddress(hModule, 'IsDebuggerPresent');
      if P = nil then
      begin
         //raise Exception.Create('Could not find procedure ' + ProcName);
      end
      else
      begin
         JIsDebuggerPresent := p;
         Result := JIsDebuggerPresent;
      end;
   end;

end;

initialization
   DefaultLevel := LOG_MAX;


end.

