unit FileDropSource;

interface

uses
  Windows, Classes, ActiveX, ShlObj, ClipBrd;

const
   MAX_NUM_FORMAT = 5;
   FD_PROGRESSUI = $4000;

type
   pFormatList = ^TFormatList;
   TFormatList = array[0..MAX_NUM_FORMAT] of TFormatEtc;

   pMediumList = ^TMediumList;
   TMediumList = array[0..MAX_NUM_FORMAT] of TStgMedium;

   TSendData = class;

   TCreateFileGroup = function : HGLOBAL of object;
   TCreateFileContents = function (nItem : Integer) : TSendData of object;

  TFileGroupDescriptorW2 = record
    cItems: UINT;
    fgd: array[0..0] of TFileDescriptorW;
  end;

  PFileGroupDescriptorW2 = ^TFileGroupDescriptorW2;

   TInterfacedComponent = class(TComponent, IUnknown)
   private
      fRefCount: Integer;
   protected
      function QueryInterface(const IID: TGuid; out Obj): HRESULT; override; StdCall;
      function _AddRef: Integer; StdCall;
      function _Release: Integer; StdCall;
   public
      property RefCount: Integer read fRefCount;
   end;

   TDropSource = class(TInterfacedComponent, IDropSource, IDataObject)
   private
      m_cFormatsAvailable : Integer;
      m_strFormatEtc : TFormatList;
      m_strStgMedium : TMediumList;

     g_cfFileContents        : Integer;
     g_cfFileGroupDescriptor : Integer;
     g_private_type          : Integer; // the private type is to detect drag'n'drop with myself
   protected
      // IDropSource implementation
      function QueryContinueDrag(fEscapePressed: bool; grfKeyState: LongInt): HRESULT; StdCall;
      function GiveFeedback(dwEffect: LongInt): HRESULT; StdCall;

      // IDataObject implementation
      function GetData(const FormatEtcIn: TFormatEtc; out Medium: TStgMedium):HRESULT; StdCall;
      function GetDataHere(const FormatEtc: TFormatEtc; out Medium: TStgMedium):HRESULT; StdCall;
      function QueryGetData(const FormatEtc: TFormatEtc): HRESULT; StdCall;
      function GetCanonicalFormatEtc(const FormatEtc: TFormatEtc; out FormatEtcOut: TFormatEtc): HRESULT; StdCall;
      function SetData(const FormatEtc: TFormatEtc; var Medium: TStgMedium; fRelease: Bool): HRESULT; StdCall;
      function EnumFormatEtc(dwDirection: LongInt; out EnumFormatEtc: IEnumFormatEtc): HRESULT; StdCall;
      function dAdvise(const FormatEtc: TFormatEtc; advf: LongInt; const advsink: IAdviseSink; out dwConnection: LongInt): HRESULT; StdCall;
      function dUnadvise(dwConnection: LongInt): HRESULT; StdCall;
      function EnumdAdvise(out EnumAdvise: IEnumStatData): HRESULT; StdCall;

   public
      fOnCreateFileGroup      : TCreateFileGroup;
      fOnCreateFileContents   : TCreateFileContents;


      constructor Create(aowner: TComponent); override;
      procedure PrepareData;
      function Execute : Integer;//TDragResult;
      function CopyToClipboard: boolean; virtual;
   end;

   TSendData = class(TInterfacedComponent, IStream)
   protected
      Size : Int64;
      Ptr  : Int64;

    function Read(pv: Pointer; cb: Longint; pcbRead: PLongint): HResult; stdcall;
    function Write(pv: Pointer; cb: Longint; pcbWritten: PLongint): HResult; stdcall;

    function Seek(dlibMove: Largeint; dwOrigin: Longint; out libNewPosition: Largeint): HResult; stdcall;
    function SetSize(libNewSize: Largeint): HResult; stdcall;
    function CopyTo(stm: IStream; cb: Largeint; out cbRead: Largeint; out cbWritten: Largeint): HResult; stdcall;
    function Commit(grfCommitFlags: Longint): HResult; stdcall;
    function Revert: HResult; stdcall;
    function LockRegion(libOffset: Largeint; cb: Largeint; dwLockType: Longint): HResult; stdcall;
    function UnlockRegion(libOffset: Largeint; cb: Largeint; dwLockType: Longint): HResult; stdcall;
    function Stat(out statstg: TStatStg; grfStatFlag: Longint): HResult; stdcall;
    function Clone(out stm: IStream): HResult; stdcall;
   public
      constructor Create(aowner: TComponent); override;
      function ReadData(Buffer : PChar; Offset : Int64; Length : Int64) : Boolean; virtual; abstract;
   end;

  // -----------------------------------------------------------------------------
  //			TEnumFormatEtc
  // -----------------------------------------------------------------------------
  { TEnumFormatEtc - format enumerator for TDataObject }


   TEnumFormatEtc = class(TInterfacedObject, IEnumFormatEtc)
   private
      fFormatList  : pFormatList;
      fFormatCount : Integer;
      fIndex       : Integer;
   public
      constructor Create(FormatList: pFormatList; FormatCount, Index: Integer);
    { IEnumFormatEtc }
      function Next(Celt: LongInt; out Elt; pCeltFetched: pLongInt): HRESULT; StdCall;
      function Skip(Celt: LongInt): HRESULT; StdCall;
      function Reset: HRESULT; StdCall;
      function Clone(out Enum: IEnumFormatEtc): HRESULT; StdCall;
  end;

procedure StrLCopyW(Dest : Pointer; Src : WideString; Len : Integer);

implementation

//////////////////////////////
//                          //
//                          //
// TInterfacedComponent     //
//                          //
//                          //
//////////////////////////////

function TInterfacedComponent.QueryInterface(const IID: TGuid; out Obj): HRESULT;
begin
   if GetInterface(IID, Obj) then
   begin
      Result := 0
   end
   else
   begin
      Result := E_NOINTERFACE;
   end;
end;


function TInterfacedComponent._AddRef: Integer;
begin
   Inc(fRefCount);
   Result := fRefCount;
end;

function TInterfacedComponent._Release: Integer;
begin
   Dec(fRefCount);
   if fRefCount = 0 then
   begin
      Destroy;
      result := 0;
      Exit;
   end;
   result := fRefCount;
end;

//////////////////////////////
//                          //
//                          //
// TDropSource              //
//                          //
//                          //
//////////////////////////////


constructor TDropSource.Create(aOwner: TComponent);
begin
   inherited Create(aOwner);
//   DragTypes := [dtCopy]; //default to Copy.
   //To avoid premature release ...
   _AddRef;
   m_cFormatsAvailable := 0;

   // Get the ID's of the formats we need
   g_cfFileContents        := RegisterClipboardFormat(CFSTR_FILECONTENTS);
   g_cfFileGroupDescriptor := RegisterClipboardFormat(CFSTR_FILEDESCRIPTORW);
   g_private_type          := RegisterClipboardFormat(PChar(aowner.Name));

   fOnCreateFileGroup      := nil;
   fOnCreateFileContents   := nil;
end;

// Need to convert to TYMED_ISTREAM (IStream)

procedure TDropSource.PrepareData;
var
   res      : HRESULT;
   effect   : LongInt;
   fe       : FORMATETC;
   stm      : STGMEDIUM;
begin
   ZeroMemory(@stm, sizeof(stm));
   ZeroMemory(@fe, sizeof(fe));

   // File Contents
   stm.tymed := TYMED_ISTREAM;//TYMED_HGLOBAL;
   stm.stm := nil; // IStrear object

	fe.tymed    := stm.tymed;//TYMED_ISTREAM;//TYMED_HGLOBAL;
	fe.ptd      := nil;
	fe.lindex   := -1;
	fe.dwAspect := DVASPECT_CONTENT;
   fe.cfFormat := g_cfFileContents;
   SetData(fe, stm, True);

   // File Group descriptor
   stm.tymed := {TYMED_ISTREAM;//}TYMED_HGLOBAL;
   stm.hGlobal := 0; // data will be provided in DataObject::GetData delayed rendering

	fe.tymed    := stm.tymed;//TYMED_ISTREAM;//TYMED_HGLOBAL;
	fe.ptd      := nil;
	fe.lindex   := -1;
	fe.dwAspect := DVASPECT_CONTENT;
   fe.cfFormat := g_cfFileGroupDescriptor;
   SetData(fe,stm,TRUE);


   // Private data
   stm.tymed := {TYMED_ISTREAM;//}TYMED_HGLOBAL;
   stm.hGlobal := 0; // data will be provided in DataObject::GetData delayed rendering

	fe.tymed    := stm.tymed;//TYMED_ISTREAM;//TYMED_HGLOBAL;
	fe.ptd      := nil;
	fe.lindex   := -1;
	fe.dwAspect := DVASPECT_CONTENT;
   fe.cfFormat := g_private_type;
   SetData(fe,stm,TRUE);

end;
////
function TDropSource.Execute: Integer;
var
   res      : HRESULT;
   effect   : LongInt;
begin
   PrepareData;

   res := DoDragDrop(Self as IDataObject, Self as IDropSource, DROPEFFECT_COPY, effect);
   case res of
      DRAGDROP_S_DROP:
         Result := 1;

      DRAGDROP_S_CANCEL:
         result := 0;//drCancel;

      E_OUTOFMEMORY:
         result := -1;//drOutMemory;
   else
      Result := 0;
   end;
end;

function TDropSource.CopyToClipboard: boolean;
var
   res      : HRESULT;
begin
   PrepareData;

   Res := OleSetClipboard(self as IDataObject);
   if Res = S_OK then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

// IDropSource implementation
function TDropSource.QueryContinueDrag(fEscapePressed: bool; grfKeyState: LongInt): HRESULT; StdCall;
begin
   Result := S_OK; // default

   if fEscapePressed then
   begin
      Result := DRAGDROP_S_CANCEL
   end
   else
   begin
      if (grfKeyState and (MK_LBUTTON or MK_RBUTTON)) = 0 then
      begin
         Result := DRAGDROP_S_DROP;
      end;
   end;
end;

function TDropSource.GiveFeedback(dwEffect: LongInt): HRESULT; StdCall;
begin
   Result := DRAGDROP_S_USEDEFAULTCURSORS;
end;

// IDataObject implementation
function TDropSource.GetCanonicalFormatEtc(const FormatEtc: TFormatEtc; out FormatEtcOut: TFormatEtc): HRESULT;
begin
   Result := DATA_S_SAMEFORMATETC;
end;

function TDropSource.SetData(const FormatEtc: TFormatEtc; var Medium: TStgMedium; fRelease: Bool): HRESULT;
var
   i : Integer;
begin
   if not fRelease then
   begin
      Result := E_FAIL;
      exit;
   end;

   i := 0;
   if m_cFormatsAvailable > 0 then
   begin
      for i := 0 to m_cFormatsAvailable - 1 do
      begin
         if m_strFormatEtc[i].cfFormat = FormatEtc.cfFormat then
         begin
            m_strStgMedium[i] := Medium;//*pSTM;
            m_strFormatEtc[i] := FormatEtc;//*pFE;
         end;
      end;
   end;

   if i < MAX_NUM_FORMAT then
   begin
   	m_strStgMedium[i] := Medium;
   	m_strFormatEtc[i] := FormatEtc;
	   inc(m_cFormatsAvailable);
	   Result := S_OK;
   end
   else
   begin
   	Result := E_OUTOFMEMORY;
   end;
end;


function TDropSource.EnumDAdvise(OUT EnumAdvise: IEnumStatData): HRESULT;
begin
   Result := OLE_E_ADVISENOTSUPPORTED;
end;

function TDropSource.DAdvise(const FormatEtc: TFormatEtc; advf: LongInt; const advSink: IAdviseSink; OUT dwConnection: LongInt): HRESULT;
begin
   Result := E_NOTIMPL;
end;

function TDropSource.DUnadvise(dwConnection: LongInt): HRESULT;
begin
    result := E_NOTIMPL;
end;

function TDropSource.GetData(const FormatEtcIn: TFormatEtc; out Medium: TStgMedium):HRESULT; StdCall;
var
   i : Integer;
   ds : TSendData;
begin
// check for valid arguments......?

	Medium.hGlobal := 0;
   for i := 0 to m_cFormatsAvailable - 1 do
   begin
		if ((FormatEtcIn.tymed and m_strFormatEtc[i].tymed) > 0) and
			(FormatEtcIn.dwAspect = m_strFormatEtc[i].dwAspect) and
			(FormatEtcIn.cfFormat = m_strFormatEtc[i].cfFormat) then
      begin
			Medium.tymed := m_strStgMedium[i].tymed;

			if FormatEtcIn.cfFormat = g_cfFileGroupDescriptor then
			begin
            if Assigned(fOnCreateFileGroup) then
            begin
               Medium.hGlobal := fOnCreateFileGroup;
            end
            else
            begin
               Medium.hGlobal := 0;
            end;

            if Medium.hGlobal <> 0 then
            begin
               Medium.tymed := TYMED_HGLOBAL;
				   Result := S_OK;
               exit;
            end
			end;
			if FormatEtcIn.cfFormat = g_cfFileContents then
			begin
//            OutputDebugString("CImpIDataObject::GetData() - FileContents\n");
				//Medium.hGlobal := CreateFileContents(FormatEtcIn.lindex);
            if Assigned(fOnCreateFileContents) then
            begin
               ds := fOnCreateFileContents(FormatEtcIn.lindex);
               //ds := TSendData.Create(owner);
               //ds.Data := 'This is a test';
               Medium.stm := Pointer(ds as IStream);

               Medium.tymed := TYMED_ISTREAM;
				   Result := S_OK;
               exit;
            end;
			end
		end
	end;
	Result := DV_E_FORMATETC;
end;

  //******************* TDropSource.GetDataHere *************************
function TDropSource.GetDataHere(const FormatEtc: TFormatEtc; out Medium: TStgMedium):HRESULT; StdCall;
begin
   Result := E_NOTIMPL;
end;

function TDropSource.QueryGetData(const FormatEtc: TFormatEtc): HRESULT; StdCall;
var
   i : Integer;
begin
	//
	// Check the aspects we support.  Implementations of this object will only
	// support DVASPECT_CONTENT.
	//
	if (DVASPECT_CONTENT and FormatEtc.dwAspect) = 0 then
   begin
      Result := DV_E_DVASPECT;
      exit;
   end;

	//
	// Now check for an appropriate clipboard format and TYMED.
	//
   for i := 0 to m_cFormatsAvailable - 1 do
   begin
      if (FormatEtc.cfFormat = m_strFormatEtc[i].cfFormat) and
         ((FormatEtc.tymed and m_strFormatEtc[i].tymed) > 0) then
      begin
         Result := S_OK;
         exit;
      end;
   end;

   Result := DV_E_TYMED;
end;

  //******************* TDropSource.EnumFormatEtc *************************
function TDropSource.EnumFormatEtc(dwDirection: LongInt; out EnumFormatEtc:IEnumFormatEtc): HRESULT; StdCall;
begin
   if dwDirection = DATADIR_GET then
   begin
      EnumFormatEtc := TEnumFormatEtc.Create(@m_strFormatEtc, m_cFormatsAvailable, 0);
      Result := S_OK;
   end
   else
   begin
      Result := OLE_S_USEREG; // Should this be Not Implemented?
   end
end;


//////////////////////////////
//                          //
//                          //
// TEnumFormatEtc           //
//                          //
//                          //
//////////////////////////////

constructor TEnumFormatEtc.Create(FormatList: {array of TFormatEtc} pFormatList; FormatCount, Index: Integer);
begin
   inherited Create;
   fFormatList  := FormatList;
   fFormatCount := FormatCount;
   fIndex       := Index;
end;

function TEnumFormatEtc.Next(Celt: LongInt; out Elt; pCeltFetched: pLongInt): HRESULT;
var
   i: Integer;
begin
   i := 0;
   while (i < Celt) and (fIndex < fFormatCount) do
   begin
      TFormatList(Elt)[i] := fFormatList[fIndex];
      Inc(fIndex);
      Inc(i);
   end;
   if pCeltFetched <> nil then
   begin
      pCeltFetched^ := i;
   end;
   if i = Celt then
   begin
      result := S_OK;
   end
   else
   begin
      result := S_FALSE;
   end;
end;

function TEnumFormatEtc.Skip(Celt: LongInt): HRESULT;
begin
   if Celt <= fFormatCount - fIndex then
   begin
      fIndex := fIndex + Celt;
      result := S_OK;
   end
   else
   begin
      fIndex := fFormatCount;
      result := S_FALSE;
   end;
end;

function TEnumFormatEtc.ReSet: HRESULT;
begin
   fIndex := 0;
   result := S_OK;
end;

function TEnumFormatEtc.Clone(OUT Enum: IEnumFormatEtc): HRESULT;
begin
   enum   := TEnumFormatEtc.Create(fFormatList, fFormatCount, fIndex);
   result := S_OK;
end;

(* How to use:

      DragDrop := TDropSource.Create(self);
      DragDrop.fOnCreateFileGroup    := CreateFileGroupDescriptor;
      DragDrop.fOnCreateFileContents := CreateFileContents;
      DragDrop.Execute;
      DragDrop.Free;
*)

(* Sample callback functions.....

function TMainForm.CreateFileGroupDescriptor : HGLOBAL;
var
   nItems   : Integer;
   hg       : HGLOBAL;
   pfgd     : PFILEGROUPDESCRIPTOR;
   i        : integer;
   FileName : String;
begin
   nItems := 30;
//	OutputDebugString("Create group descriptor\n");


   hg := GlobalAlloc({GMEM_FIXED}GMEM_MOVEABLE , sizeof(FILEGROUPDESCRIPTOR) + (nItems - 1) * sizeof(FILEDESCRIPTOR));
   if hg <> 0 then
   begin
      pfgd := PFILEGROUPDESCRIPTOR(GlobalLock(hg));
      if Assigned(pfgd) then
      begin
         i := 0;

         pfgd.cItems := nItems;
//         TCHAR szName[MAX_PATH];

         // add a dir
	      pfgd.fgd[i].dwFlags := FD_ATTRIBUTES;
	      pfgd.fgd[i].dwFileAttributes := FILE_ATTRIBUTE_DIRECTORY;

//         wsprintf(szName, TEXT("mydir"), i);
//         lstrcpyn(pfgd->fgd[i].cFileName, szName, sizeof(pfgd->fgd[i].cFileName));
         FileName := 'mydir';
         StrLCopy(pfgd.fgd[i].cFileName, PChar(FileName), sizeof(pfgd.fgd[i].cFileName));

         // and 2 files
//         for(i = 1; i < nItems; i++)
         for i := 1 to nItems - 1 do
         begin
	         pfgd.fgd[i].dwFlags := FD_ATTRIBUTES;
	         pfgd.fgd[i].dwFileAttributes := FILE_ATTRIBUTE_NORMAL;

//            wsprintf(szName, TEXT("mydir\\myfile%d.txt"), i);
            FileName := 'myfile' + IntToStr(i) + '.txt';
            StrLCopy(pfgd.fgd[i].cFileName, PChar(FileName), sizeof(pfgd.fgd[i].cFileName));
         end;

         GlobalUnlock(hg);
      end;
   end;
   Result := hg;
end;

function TMainForm.CreateFileContents(nItem : Integer) : HGLOBAL;
var
   hg    : HGLOBAL;
   Text  : String;
   psz   : PChar;
   FileSize : DWORD;
begin
//   hg := 0;
//	OutputDebugString("Create file contents\n");
//   char     szText[MAX_PATH];

//   OutputDebugString("Extracting data from ex2fs partition...\n");

//   wsprintfA(szText, "This file is myfile%d.txt created from CFSTR_FILECONTENTS\r\n", nItem);
   Text := 'This is myfile' + IntToStr(nItem) + '.txt';
//   OutputDebugString(szText);

   FileSize := Length(Text) + 1;

   if nItem = 3 then
   begin
      FileSize := 2 * 1024 * 1024;
   end;
   hg := GlobalAlloc({GMEM_FIXED}GMEM_MOVEABLE , FileSize);

   if hg > 0 then
   begin
      psz := PChar(GlobalLock(hg));
      if Assigned(psz) then
      begin
//         lstrcpyA(psz, szText);
         StrCopy(psz, PChar(Text));

         GlobalUnlock(hg);
      end;
   end
   else
   begin
      MessageDlg('Error allocating memory!', mtError, [mbOK], 0);
   end;
   Result := hg;
end;

*)

// Dest is a PWideChar but due to bugs in ShlObj, we only get an ansi type pointer
// just cast it to WideChar and it works...
// len should be in bytes (ie not chars)
procedure StrLCopyW(Dest : Pointer; Src : WideString; Len : Integer);
var
   c : Integer;
   l : Integer;
begin
   c := Len div 2;
   if Length(Src) > c then
   begin
      l := c;
   end
   else
   begin
      l := Length(Src);
   end;

   CopyMemory(Dest, PWideChar(Src), l * sizeof(WideChar));
   PWideChar(Dest)[l] := #0;
end;


function TSendData.Read(pv: Pointer; cb: Longint; pcbRead: PLongint): HResult; stdcall;
var
   Len : Int64;
begin
   Len := cb;
   if (Len + Ptr) > Size then
   begin
      Len := Size - Ptr;
   end;
   if ReadData(pv, Ptr, Len) then
   begin
      pcbRead^ := Len;
      Ptr := Ptr + Len;
      Result := S_OK;
   end
   else
   begin
      Result := S_FALSE;
   end;
end;

function TSendData.Write(pv: Pointer; cb: Longint; pcbWritten: PLongint): HResult; stdcall;
begin
   Result := E_NOTIMPL;
end;

function TSendData.Seek(dlibMove: Largeint; dwOrigin: Longint; out libNewPosition: Largeint): HResult; stdcall;
begin
   if dwOrigin = STREAM_SEEK_SET then
   begin
      Ptr := dlibMove;
      Result := S_OK;
   end
   else if dwOrigin = STREAM_SEEK_CUR then
   begin
      Ptr := Ptr + dlibMove;
      Result := S_OK;
   end
   else if dwOrigin = STREAM_SEEK_END then
   begin
      Ptr := Size - dlibMove;
      Result := S_OK;
   end
   else
   begin
      Result := STG_E_INVALIDFUNCTION;
   end;

   if (Result = S_OK) and ((Ptr < 0) or (Ptr > Size)) then
   begin
      Result := STG_E_INVALIDPOINTER;
   end;

end;

function TSendData.SetSize(libNewSize: Largeint): HResult; stdcall;
begin
   Result := E_NOTIMPL;
end;

function TSendData.CopyTo(stm: IStream; cb: Largeint; out cbRead: Largeint; out cbWritten: Largeint): HResult; stdcall;
begin
   Result := E_NOTIMPL;
end;

function TSendData.Commit(grfCommitFlags: Longint): HResult; stdcall;
begin
   Result := E_NOTIMPL;
end;

function TSendData.Revert: HResult; stdcall;
begin
   Result := E_NOTIMPL;
end;

function TSendData.LockRegion(libOffset: Largeint; cb: Largeint; dwLockType: Longint): HResult; stdcall;
begin
   Result := E_NOTIMPL;
end;

function TSendData.UnlockRegion(libOffset: Largeint; cb: Largeint; dwLockType: Longint): HResult; stdcall;
begin
   Result := E_NOTIMPL;
end;

function TSendData.Stat(out statstg: TStatStg; grfStatFlag: Longint): HResult; stdcall;
begin
   Result := E_NOTIMPL;
end;

function TSendData.Clone(out stm: IStream): HResult; stdcall;
begin
   Result := E_NOTIMPL;
end;

constructor TSendData.Create(aowner: TComponent);
begin
   inherited;
   _AddRef;
end;


initialization
  OleInitialize(nil);

finalization
  OleUninitialize;

end.
