unit FileDropTarget;

interface

uses Windows, Classes, ActiveX, ShlObj, ClipBrd, FileDropSource, TntClasses, winbinfile;

type
   TImportFileContents = function (Window : HWND; Name : WideString; Length : Int64; Data : IStream) : Boolean of object;


   TDropTarget = class(TInterfacedComponent, IDropTarget)
   private
      WindowHandles : TList;
      m_cFormatsAvailable : Integer;
      m_strFormatEtc : TFormatList;
      m_strStgMedium : TMediumList;

     g_cfFileContents        : Integer;
     g_cfFileGroupDescriptor : Integer;
     g_private_type          : Integer; // the private type is to detect drag'n'drop with myself
   protected
      // IDropTarget implementation
    function DragEnter(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint; var dwEffect: Longint): HResult; stdcall;
    function DragOver(grfKeyState: Longint; pt: TPoint; var dwEffect: Longint): HResult; stdcall;
    function DragLeave: HResult; stdcall;
    function Drop(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint; var dwEffect: Longint): HResult; stdcall;

   public
      fOnImportFileContents   : TImportFileContents;



      constructor Create(aowner: TComponent); override;
      procedure AddWindow(h : hwnd);
      function AddFormat(const FormatEtc: TFormatEtc; var Medium: TStgMedium): HRESULT;
      procedure PrepareData;
      function SupportedFormat(dataObj: IDataObject) : Boolean;
      function PrivateFormat(dataObj: IDataObject) : Boolean;
      procedure ConvertDropfilesToList(DropFiles : PDropFiles; List : TTntStringList);
      function ProcessData(Handle : HWND; DataObject: IDataObject) : Boolean;
   end;


   TWinBinData = class(TSendData)
   private
      WinBinFile : TBinaryFile;
   public
      constructor Create(aowner: TComponent); override;
      destructor Destroy; override;
      procedure SetData(Filename : WideString);
      function ReadData(Buffer : PChar; Offset : Int64; Length : Int64) : Boolean; override;
      function GetFileSize : Int64;
   end;


implementation

uses debug, sysutils, filectrl;

constructor TDropTarget.Create(aOwner: TComponent);
begin
   inherited Create(aOwner);
   //To avoid premature release ...
   _AddRef;

   m_cFormatsAvailable := 0;
   // Get the ID's of the formats we need
   g_cfFileContents        := RegisterClipboardFormat(CFSTR_FILECONTENTS);
   g_cfFileGroupDescriptor := RegisterClipboardFormat(CFSTR_FILEDESCRIPTORW);
   g_private_type          := RegisterClipboardFormat(PChar(aowner.Name));

   fOnImportFileContents   := nil;

   WindowHandles := TList.Create;
end;

procedure TDropTarget.AddWindow(h : hwnd);
var
   res : HRESULT;
begin
   Res := RegisterDragDrop(h, self as IDropTarget);
   if Res = S_OK then
   begin
      WindowHandles.Add(Pointer(h));
   end;
end;

function TDropTarget.AddFormat(const FormatEtc: TFormatEtc; var Medium: TStgMedium): HRESULT;
var
   i : Integer;
begin
   i := 0;
   if m_cFormatsAvailable > 0 then
   begin
      for i := 0 to m_cFormatsAvailable - 1 do
      begin
         if m_strFormatEtc[i].cfFormat = FormatEtc.cfFormat then
         begin
            m_strStgMedium[i] := Medium;//*pSTM;
            m_strFormatEtc[i] := FormatEtc;//*pFE;
         end;
      end;
   end;

   if i < MAX_NUM_FORMAT then
   begin
   	m_strStgMedium[i] := Medium;
   	m_strFormatEtc[i] := FormatEtc;
	   inc(m_cFormatsAvailable);
	   Result := S_OK;
   end
   else
   begin
   	Result := E_OUTOFMEMORY;
   end;
end;


procedure TDropTarget.PrepareData;
var
   fe       : FORMATETC;
   stm      : STGMEDIUM;
begin
   ZeroMemory(@stm, sizeof(stm));
   ZeroMemory(@fe, sizeof(fe));

   // File Contents
   stm.tymed := TYMED_ISTREAM;
   stm.stm := nil; // IStrear object

	fe.tymed    := stm.tymed;
	fe.ptd      := nil;
	fe.lindex   := -1;
	fe.dwAspect := DVASPECT_CONTENT;
   fe.cfFormat := g_cfFileContents;
   AddFormat(fe, stm);

   // File Group descriptor
   stm.tymed := TYMED_HGLOBAL;
   stm.hGlobal := 0;

	fe.tymed    := stm.tymed;
	fe.ptd      := nil;
	fe.lindex   := -1;
	fe.dwAspect := DVASPECT_CONTENT;
   fe.cfFormat := g_cfFileGroupDescriptor;
   AddFormat(fe,stm);

   // File names
   stm.tymed := TYMED_HGLOBAL;
   stm.hGlobal := 0;

	fe.tymed    := stm.tymed;
	fe.ptd      := nil;
	fe.lindex   := -1;
	fe.dwAspect := DVASPECT_CONTENT;
   fe.cfFormat := CF_HDROP;
   AddFormat(fe,stm);

end;

function TDropTarget.SupportedFormat(dataObj: IDataObject) : Boolean;
var
   enum     : IEnumFormatEtc;
   Ret      : HRESULT;
   Fetched  : Longint;
   elt      : TFormatList;

   i        : Integer;

begin
   Result := False;

   Ret := dataObj.EnumFormatEtc(DATADIR_GET, enum);

   if Ret = S_OK then
   begin
      while true do
      begin
         Ret := Enum.Next(1, elt, @Fetched);
         if Ret = S_OK then
         begin
            for i := 0 to m_cFormatsAvailable - 1 do
            begin
               if (Elt[0].cfFormat = m_strFormatEtc[i].cfFormat) and (Elt[0].dwAspect  = m_strFormatEtc[i].dwAspect) then
               begin
                  Result := True;
                  exit;
               end;
            end;
         end
         else
         begin
            break;
         end;
      end;
   end;
end;

function TDropTarget.PrivateFormat(dataObj: IDataObject) : Boolean;
var
   Ret      : HRESULT;
   Format : FORMATETC;

begin
   Format.cfFormat   := g_private_type;
   Format.ptd        := nil;
   Format.dwAspect   := DVASPECT_CONTENT;
   Format.lindex     := -1;
   Format.tymed      := TYMED_HGLOBAL;

   Ret := DataObj.QueryGetData(Format);
   if Ret = S_OK then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TDropTarget.DragEnter(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint; var dwEffect: Longint): HResult;
begin
   Log('DragEnter');

   if PrivateFormat(dataObj) then
   begin
      Result := E_UNEXPECTED;
   end
   else if SupportedFormat(dataObj) then
   begin
      Result := S_OK;
   end
   else
   begin
      Result := E_UNEXPECTED;
   end;

end;

function TDropTarget.DragOver(grfKeyState: Longint; pt: TPoint; var dwEffect: Longint): HResult;
begin
   Result := S_OK;
end;

function TDropTarget.DragLeave: HResult;
begin
   Log('DragLeave');

   Result := S_OK;
end;

function TDropTarget.Drop(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint; var dwEffect: Longint): HResult;
var
   rect  : TRect;
   i     : Integer;
   h     : HWND;
begin
   // work out which window we are in
   for i := 0 to WindowHandles.Count - 1 do
   begin
      h := HWND(WindowHandles[i]);
      if GetWindowRect(h, rect) then
      begin
         if    (pt.x >= rect.Left) and (pt.x <= rect.Right)
           and (pt.y >= rect.Top)  and (pt.y <= rect.Bottom) then
         begin
            break;
         end;
      end
      else
      begin
         h := 0;
      end;
   end;
   Log('h=' + IntToStr(h));
   ProcessData(h, dataObj);
   Result := S_OK;
end;

procedure TDropTarget.ConvertDropfilesToList(DropFiles : PDropFiles; List : TTntStringList);
var
   Start : PChar;
   WS : WideString;
   S : String;
begin
   Start := PChar(DropFiles) + DropFiles.pFiles;
   while Start[0] <> #0 do
   begin
      if DropFiles.fWide then
      begin
         WS := WideString((PWideChar(Start)));
         List.Add(WS);
         Start := Start + ((Length(WS) + 1) * sizeof(WideChar));
      end
      else
      begin
         S := String(Start);
         List.Add(S);
         Start := Start + Length(S) + 1;
      end;
   end;
end;


function TDropTarget.ProcessData(Handle : HWND; DataObject: IDataObject) : Boolean;
var
   Ret         : HRESULT;
   Format      : TFormatEtc;

   Medium      : TStgMedium;
   Medium2     : TStgMedium;
   DropFiles   : PDropFiles;
   FileList    : TTntStringList;

   SearchHandle : THandle;
   Search      : WideString;
   FindData    : TWIN32FindDataW;
   BaseDir     : WideString;
   Tail        : WideString;

   i           : Integer;
   FileGroup   : PFileGroupDescriptorW2;
   IsDir       : Boolean;

   FileData : TWinBinData;
   Temp     : LARGE_INTEGER;


begin
   if Assigned(DataObject) then
   begin
      Log('Got DataObject');
      Format.cfFormat   := CF_HDROP;
      Format.ptd        := nil;
      Format.dwAspect   := DVASPECT_CONTENT;
      Format.lindex     := -1;
      Format.tymed      := TYMED_HGLOBAL;

      Ret := DataObject.QueryGetData(Format);
      if Ret = S_OK then
      begin
         Ret := DataObject.GetData(Format, Medium);
         if Ret = S_OK then
         begin
            DropFiles := GlobalLock(Medium.hGlobal);
            FileList := TTntStringList.Create;
            ConvertDropfilesToList(DropFiles, FileList);
            if Assigned(fOnImportFileContents) then
            begin
               while FileList.Count > 0 do
               begin
                  if Length(BaseDir) = 0 then
                  begin
                     BaseDir := ExtractFilePath(FileList[0]);
                     Log('Base directory is ' + BaseDir);
                  end;
                  Tail := Copy(FileList[0], Length(BaseDir) + 1, Length(FileList[0]));
                  if DirectoryExists(FileList[0]) then
                  begin
                     // directory
                     if not fOnImportFileContents(Handle, Tail, 0, nil) then
                     begin
                        break;
                     end;

                     // Expand the directory
                     Search := FileList[0] + '\*.*';
                     SearchHandle := FindFirstFileW(PWideChar(Search), FindData);
                     while SearchHandle <> INVALID_HANDLE_VALUE do
                     begin
                        Search := FindData.cFileName;
                        if Search = '.' then
                        begin
                        end
                        else if Search = '..' then
                        begin
                        end
                        else
                        begin
                           FileList.Add(FileList[0] + '\' + Search);
                        end;
                        if not FindNextFileW(SearchHandle, FindData) then
                        begin
                           Windows.FindClose(SearchHandle);
                           SearchHandle := INVALID_HANDLE_VALUE;
                        end;
                     end;
                  end
                  else
                  begin
                     // file
                     // create a winbinfile senddata object
                     FileData := TWinBinData.Create(nil);
                     FileData.SetData(FileList[0]);
                     if not fOnImportFileContents(Handle, Tail, FileData.GetFileSize, FileData) then
                     begin
                        FileData.Free;
                        break;
                     end;
                     FileData.Free;
                  end;
                  FileList.Delete(0);
               end;
            end;
            FileList.Free;
            GlobalUnlock(Medium.hGlobal);
         end;
      end
      else
      begin
         // Format must be g_cfFileGroupDescriptor
         Format.cfFormat := g_cfFileGroupDescriptor;
         Ret := DataObject.QueryGetData(Format);
         if Ret = S_OK then
         begin
            Ret := DataObject.GetData(Format, Medium);
            if Ret = S_OK then
            begin
               FileGroup := GlobalLock(Medium.hGlobal);
               for i := 0 to FileGroup.cItems - 1 do
               begin
                  // Get the data for this file
                  IsDir := False;
                  if FileGroup.fgd[i].dwFlags and FD_ATTRIBUTES = FD_ATTRIBUTES then
                  begin
                     if FileGroup.fgd[i].dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = FILE_ATTRIBUTE_DIRECTORY then
                     begin
                        IsDir := True;
                     end;
                  end;

                  if IsDir then
                  begin
                     if not fOnImportFileContents(Handle, FileGroup.fgd[i].cFileName, 0, nil) then
                     begin
                        break;
                     end;
                  end
                  else
                  begin
                     Format.tymed := TYMED_ISTREAM;
                     Format.cfFormat := g_cfFileContents;
                     Format.lindex := i;
                     Ret := DataObject.GetData(Format, Medium2);
                     if Ret = S_OK then
                     begin
                        Temp.LowPart := FileGroup.fgd[i].nFileSizeLow;
                        Temp.HighPart := FileGroup.fgd[i].nFileSizeHigh;
                        if Medium2.tymed = Format.tymed then
                        begin
                           if not fOnImportFileContents(Handle, FileGroup.fgd[i].cFileName, Temp.QuadPart, IStream(Medium2.stm)) then
                           begin
                              break;
                           end;
                        end;
                     end;

                  end;
               end;
               GlobalUnlock(Medium.hGlobal);
            end;
         end;
      end;
   end;
   Result := True;
end;

// TMirrorData is used to send data from vv to a droptarget

constructor TWinBinData.Create(aowner: TComponent);
begin
   inherited;

   WinBinFile := nil;

end;

destructor TWinBinData.Destroy;
begin
   if Assigned(WinBinFile) then
   begin
      WinBinFile.Free;
   end;

   inherited;
end;

procedure TWinBinData.SetData(Filename : WideString);
begin
   WinBinFile := TBinaryFile.Create;
   WinBinFile.Assign(Filename);
   WinBinFile.Open(OPEN_READ_ONLY);
   Size := WinBinFile.FileSize;
end;

function TWinBinData.ReadData(Buffer : PChar; Offset : Int64; Length : Int64) : Boolean;
begin
   if Length = 0 then
   begin
      Result := True;
      exit;
   end;

   WinBinFile.Seek(Offset);
   if WinBinFile.BlockRead2(Buffer, Length) = Length then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TWinBinData.GetFileSize : Int64;
begin
   Result := WinBinFile.FileSize;
end;



end.
