unit explore;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  TntMenus, Menus, ComCtrls, ExtCtrls, TntComCtrls, StdCtrls, NetLib,
  ImgList, WinBinFile, ShellAPI, ActiveX, ShlObj, ClipBrd, FileDropSource,
  TntSystem, TntClasses, users, FileDropTarget;

type
  TMNode = class(TObject)
  public
   constructor Create;
   destructor Destroy; override;
   procedure AddChild(Child : TMNode);
   function GetChild(Index : Integer) : TMNode;
   procedure PruneChildren;

   function GetPath : String; // Same as MNodeToPath

   procedure LoadPasswd;
   function GetUser : String;
   function GetGroup : String;

   function CreateFile(Endpoint : TNetEndpoint; Name : WideString) : Boolean;
   function CreateDir(Endpoint : TNetEndpoint; Name : WideString) : Boolean;
   function OpenData(Endpoint : TNetEndpoint) : Boolean;
   function CloseData : Boolean;
   function ReadData(Buffer : PChar; Offset : Int64; Length : Int64) : Boolean;
   function WriteData(Buffer : PChar; Offset : Int64; Length : Int64) : Boolean;

  protected
   Parent      : TMNode;
   DataLoaded  : Boolean;
   Children    : TList;
   Utf8        : String;     // UTF8 name
   Utf16       : WideString; // UTF16 name
   Size        : Int64;
   FileType    : Char;
   TreeNode    : TTntTreeNode;

   // This is the passwd node that belongs to this tree.
   // if Passwd is null then search the parent until you find one.
   PasswdFileName : String;
   GroupFileName  : String;
   Passwd         : TUserList;

   Endpoint : TNetEndpoint;   
   MHandle  : String;

   // Attributes.  Should this be a stringlist?
   uid   : String;
   gid   : String;
   mode  : String;
   acl   : String;
   atime : String;
   mtime : String;
   ctime : String;

  end;

   TMirrorData = class(TSendData)
   private
      EndPoint : TNetEndpoint;
      MNode    : TMNode;
   public
      constructor Create(aowner: TComponent); override;
      destructor Destroy; override;
      procedure SetData(fEndpoint : TNetEndpoint; fMNode : TMNode);
      function ReadData(Buffer : PChar; Offset : Int64; Length : Int64) : Boolean; override;
   end;

  TExploreForm = class(TForm)
    ToolbarPanel: TPanel;
    TntMainMenu1: TTntMainMenu;
    File1: TTntMenuItem;
    Exit1: TTntMenuItem;
    TntPopupMenu1: TTntPopupMenu;
    Lockthetoolbar1: TTntMenuItem;
    Panel1: TPanel;
    Splitter1: TSplitter;
    Panel2: TPanel;
    TntTreeView1: TTntTreeView;
    TntListView1: TTntListView;
    TntStatusBar1: TTntStatusBar;
    Memo1: TMemo;
    SmallImageList: TImageList;
    LargeImageList: TImageList;
    TreePopupMenu: TTntPopupMenu;
    Mount1: TTntMenuItem;
    Copy1: TTntMenuItem;
    Paste1: TTntMenuItem;
    ListPopupMenu: TTntPopupMenu;
    L1: TTntMenuItem;
    Edit1: TTntMenuItem;
    View1: TTntMenuItem;
    Help1: TTntMenuItem;
    About1: TTntMenuItem;
    Toolbar1: TTntMenuItem;
    Statusbar1: TTntMenuItem;
    Debug1: TTntMenuItem;
    N1: TTntMenuItem;
    LargeIcons1: TTntMenuItem;
    SmallIcons1: TTntMenuItem;
    List1: TTntMenuItem;
    Details1: TTntMenuItem;
    N2: TTntMenuItem;
    ArrangeIconsby1: TTntMenuItem;
    Name1: TTntMenuItem;
    N3: TTntMenuItem;
    Refresh1: TTntMenuItem;
    Options1: TTntMenuItem;
    Size1: TTntMenuItem;
    Type1: TTntMenuItem;
    Date1: TTntMenuItem;
    User1: TTntMenuItem;
    Group1: TTntMenuItem;
    Attributes1: TTntMenuItem;
    N4: TTntMenuItem;
    ReverseOrder1: TTntMenuItem;
    N5: TTntMenuItem;
    CaseSensitive1: TTntMenuItem;
    procedure Exit1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure TntTreeView1Expanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    procedure TntTreeView1Expanded(Sender: TObject; Node: TTreeNode);
    procedure TntTreeView1Change(Sender: TObject; Node: TTreeNode);
    procedure TntListView1DblClick(Sender: TObject);
    procedure TntListView1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure TntTreeView1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure TntListView1CustomDrawSubItem(Sender: TCustomListView;
      Item: TListItem; SubItem: Integer; State: TCustomDrawState;
      var DefaultDraw: Boolean);
    procedure TntListView1ColumnDragged(Sender: TObject);
    procedure Mount1Click(Sender: TObject);
    procedure Copy1Click(Sender: TObject);
    procedure TreePopupMenuPopup(Sender: TObject);
    procedure Paste1Click(Sender: TObject);
    procedure TntListView1ColumnClick(Sender: TObject;
      Column: TListColumn);
    procedure TntListView1Compare(Sender: TObject; Item1, Item2: TListItem;
      Data: Integer; var Compare: Integer);
    procedure TntListView1ColumnRightClick(Sender: TObject;
      Column: TListColumn; Point: TPoint);
    procedure L1Click(Sender: TObject);
    procedure ReverseOrder1Click(Sender: TObject);
    procedure CaseSensitive1Click(Sender: TObject);
    procedure Name1Click(Sender: TObject);
    procedure LargeIcons1Click(Sender: TObject);
    procedure SmallIcons1Click(Sender: TObject);
    procedure List1Click(Sender: TObject);
    procedure Details1Click(Sender: TObject);
    procedure Toolbar1Click(Sender: TObject);
    procedure Statusbar1Click(Sender: TObject);
    procedure Debug1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
  private
    { Private declarations }
    CurrentListedMNode : TMNode;
  public
   EndPoint : TNetEndpoint;

   ShowDirectoriesInList : Boolean; // move to main options
   ShowDotFiles   : Boolean;
   ViewerProg     : String;
   ShutdownServer : Boolean;
   SortColumn     : Integer;
   SortBackwards  : Boolean;
   SortCase       : Boolean;
   ShowToolbar    : Boolean;
   ShowStatusbar  : Boolean;
   ShowDebug      : Boolean;

   DebugSplitter  : TSplitter;
   // Drag & Drop
   // These can persist if they are on the clipboard...
   StartDragging : Boolean;
   DragDrop : TDropSource;
   FilesToDrop : TTntStringList;
   FilesToDropExpanded : Boolean;
   // end Drag & Drop
   DropTarget : TDropTarget;
   // end paste

   procedure Debug(S : String);
   function NodeToPath(Node : TTreeNode) : String;
   function MNodeToPath(MNode : TMNode) : String;
   procedure UpdatePath(Node : TTntTreeNode; Path : String);
   procedure AddToTree(Parent : TTntTreeNode; MNode : TMNode);
   procedure ExpandTreeNode(Node: TTntTreeNode);
   procedure SyncListView(PMNode : TMNode);
   procedure ReSyncListView;
   function ShowFileInList(MNode : TMNode) : Boolean;
   function ShowFileInTree(MNode : TMNode) : Boolean;
   procedure SendMountRequest(MNode : TMNode; Mount : String);
   procedure SendLoopbackRequest(MNode : TMNode);
   function SaveFile(Path : String) : String;
   procedure OpenFile(FileName : String);

   function MangleFileName(OldName : WideString) : WideString;
   procedure StartListDrag(Copy : Boolean);
   procedure StartTreeDrag(Copy : Boolean);
   function CreateFileGroupDescriptor : HGLOBAL;
   function CreateFileContents(Index : Integer) : TSendData;
   function ImportFileContents(Window : HWND; Name : WideString; Length : Int64; Data : IStream) : Boolean;

   function SupportClipboardFormat : Boolean;

   procedure ConvertDropfilesToList(DropFiles : PDropFiles; List : TTntStringList);

   procedure SyncViewMenu;

    { Public declarations }
  end;

var
  ExploreForm: TExploreForm;

implementation

uses debug, strings, GuiUtils, mount, about, version;

{$R *.DFM}

{
   Connect via socket to vv-cmd
   List Directory
   Stat File

   Do we want a worker thread?

   Don't store data in the GUI!
   Is the data stored on the server or do we need our own copy ?

   1 GUI - Tnt Unicode components
   2 In-memory data - Mirror of server vnode (TMNode)
   3 Worker thread ?
}

constructor TMNode.Create;
begin
   Children := TList.Create;
end;

destructor TMNode.Destroy;
begin
   Children.Free;
end;

procedure TMNode.AddChild(Child : TMNode);
begin
   Children.Add(Child);
end;

function TMNode.GetChild(Index : Integer) : TMNode;
begin
   Result := TMNode(Children[Index]);
end;

procedure TMNode.PruneChildren;
var
   i : Integer;
begin
   for i := 0 to Children.Count - 1 do
   begin
      GetChild(i).PruneChildren;
      GetChild(i).Free;
   end;
   Children.Clear;
end;


procedure TMNode.LoadPasswd;
begin
   Passwd := TUserList.Create;
   if PasswdFileName <> '' then
   begin
      Passwd.ParsePasswd(PasswdFileName);
   end;
   if GroupFileName <> '' then
   begin
      Passwd.ParseGroup(GroupFileName);
   end;
end;

function TMNode.GetUser : String;
var
   Converter : TMNode;
begin
   Converter := self;

   while not Assigned(Converter.Passwd) and Assigned(Converter.Parent) do
   begin
      Converter := Converter.Parent;
   end;

   if Assigned(Converter.Passwd) and (Length(UID) > 0) then
   begin
      try
         Result := Converter.Passwd.GetName(StrToInt(UID));
      except
         Result := UID;
      end;
   end
   else
   begin
      Result := UID;
   end;
end;

function TMNode.GetGroup : String;
var
   Converter : TMNode;
begin
   Converter := self;

   while not Assigned(Converter.Passwd) and Assigned(Converter.Parent) do
   begin
      Converter := Converter.Parent;
   end;

   if Assigned(Converter.Passwd) and (Length(GID) > 0) then
   begin
      try
         Result := Converter.Passwd.GetGroup(StrToInt(GID));
      except
         Result := GID;
      end;
   end
   else
   begin
      Result := GID;
   end;
end;

function TMNode.CreateFile(Endpoint : TNetEndpoint; Name : WideString) : Boolean;
var
   S : String;
   Error : String;
begin
   Result := False;
   EndPoint.SendStringN('create');
   Log(EndPoint.RecvStringN);
   EndPoint.SendStringN(GetPath + '/' + Name);
   while true do
   begin
      S := EndPoint.RecvStringN;
      Log(S);
      if StartsWith(S, 'E=', Error) then
      begin
         if Error = 'Done' then
         begin
            Result := True;
         end;
         break;
      end;
   end;
end;

function TMNode.CreateDir(Endpoint : TNetEndpoint; Name : WideString) : Boolean;
begin
   Result := False;
end;

function TMNode.OpenData(Endpoint : TNetEndpoint) : Boolean;
var
   S     : String;
   Error : String;
begin
   Self.Endpoint := Endpoint;
   Result := False;
   EndPoint.SendStringN('open');
   Log(EndPoint.RecvStringN);
   EndPoint.SendStringN(GetPath);
   while true do
   begin
      S := EndPoint.RecvStringN;
      if StartsWith(S, 'H=', MHandle) then
      begin
         Log('Opened file and got handle ' + MHandle);
         Result := True;
      end
      else if StartsWith(S, 'E=', Error) then
      begin
         break;
      end;
   end;
end;

function TMNode.CloseData : Boolean;
var
   S     : String;
   Error : String;
begin
   if Length(MHandle) > 0 then
   begin
      Endpoint.SendStringN('close');
      Endpoint.RecvStringN;
      EndPoint.SendStringN(MHandle);
      S := EndPoint.RecvStringN;
      if StartsWith(S, 'E=', Error) then
      begin
         // Don't care
      end;
   end;
   Result := True;
end;

function TMNode.ReadData(Buffer : PChar; Offset : Int64; Length : Int64) : Boolean;
var
   S     : String;
   Error : String;
   L     : Integer;
begin
   Result := False;
   EndPoint.SendStringN('read');
   EndPoint.SendStringN(MHandle);
   EndPoint.SendStringN(IntToStr(Offset));
   EndPoint.SendStringN(IntToStr(Length));

   while true do
   begin
      S := EndPoint.RecvStringN;
      if StartsWith(S, 'E=', Error) then
      begin
         if Error = 'Binary' then
         begin
            while true do
            begin
               L := EndPoint.RecvInteger;
               if L = 0 then
               begin
                  break;
               end;
               if L > Length then
               begin
                  // this is some kind of error
                  break;
               end;
               EndPoint.RecvBuffer(Buffer, L);
               Result := True; // we need to advance the buffer because the data might come in multiple chunks
            end;
         end
         else
         begin
            break;
         end;
      end;
   end;
end;

function TMNode.WriteData(Buffer : PChar; Offset : Int64; Length : Int64) : Boolean;
var
   S     : String;
   Error : String;
begin
   Result := False;
   EndPoint.SendStringN('write');
   S := EndPoint.RecvStringN;
   EndPoint.SendStringN(MHandle);
   EndPoint.SendStringN(IntToStr(Offset));
   EndPoint.SendStringN(IntToStr(Length));

   EndPoint.SendBuffer(Buffer, Length);

   while true do
   begin
      S := EndPoint.RecvStringN;
      if StartsWith(S, 'E=', Error) then
      begin
         if Error = 'Done' then
         begin
            Result := True;
         end;
         break;
      end;
   end;
end;


procedure TExploreForm.Debug(S : String);
begin
   Memo1.Lines.Add(S);
end;

procedure TExploreForm.Exit1Click(Sender: TObject);
begin
   Close;
end;

procedure TExploreForm.FormCreate(Sender: TObject);
var
   Root              : TMNode;
   StartupInfo       : TStartupInfo;
   ProcessInformation : TProcessInformation;
begin
   TntStatusBar1.Panels[0].Text := 'Virtual Volumes ' + AppVersion;
   ShowDotFiles := False;
   ShowDirectoriesInList := True;
   //ViewerProg := 'C:\Program Files\Vim\vim64\gvim.exe';
   ViewerProg := 'notepad';

   // defaults which are like windows
   SortColumn     := 2; // by filetype
   SortBackwards  := False;
   SortCase       := False;
   ShowToolbar    := True;
   ShowStatusbar  := True;
   ShowDebug      := True;


   CurrentListedMNode := nil;

   Panel2.Align := alClient;

   if ShowDebug then
   begin
      Memo1.Align := alBottom;
      DebugSplitter := TSplitter.Create(Panel2);
      DebugSplitter.Parent := Panel2;
      DebugSplitter.Align := alBottom;
   end;

   TntTreeView1.Align := alClient;
   TntListView1.Align := alClient;

   SetDebug(Debug);

   Show;

   try
      EndPoint := TNetEndpoint.Create('localhost', 9000);
   except
      on e : TSocketException do
      begin
         if e.GetLastError = 10061 then
         begin
            // start vv_cmd and try again...
            FillMemory(@StartupInfo, sizeof(StartupInfo), 0);
            if IsDebuggerPresent then
            begin
               StartupInfo.wShowWindow := SW_MINIMIZE;
            end
            else
            begin
               StartupInfo.wShowWindow := SW_HIDE;
            end;
            StartupInfo.dwFillAttribute := FOREGROUND_RED or BACKGROUND_RED or BACKGROUND_GREEN or BACKGROUND_BLUE;
            StartupInfo.dwFlags := STARTF_USESHOWWINDOW or STARTF_USEFILLATTRIBUTE;
            if CreateProcess(nil, 'vv_cmd.exe server exit', nil, nil, False, 0, nil, nil, StartupInfo, ProcessInformation) then
            begin
               Sleep(5 * 1000);
               ShutdownServer := True;
            end
            else
            begin
               Log('CreateProcess failed error = ' + IntToStr(GetLastError));
            end;
            EndPoint := TNetEndpoint.Create('localhost', 9000);
         end
         else
         begin
            raise;
         end;
      end;
   end;
   Log(EndPoint.RecvStringN);

   Root := TMNode.Create;
   Root.TreeNode := TntTreeView1.Items[0];
   TntTreeView1.Items[0].Data := Root;
   TntTreeView1.Items[0].HasChildren := True;

   // load the passwd file...
   try
      Root.PasswdFileName := SaveFile('/etc/passwd');
      Root.GroupFileName  := SaveFile('/etc/group');
      Root.LoadPasswd;
   except
      Root.Passwd := nil;
   end;

   DropTarget := TDropTarget.Create(self);
   DropTarget.PrepareData;
   DropTarget.fOnImportFileContents := ImportFileContents;
   DropTarget.AddWindow(TntTreeView1.Handle);
   DropTarget.AddWindow(TntListView1.Handle);

   SyncViewMenu;     

end;

procedure TExploreForm.AddToTree(Parent : TTntTreeNode; MNode : TMNode);
var
   Node : TTntTreeNode;
begin
   if ShowFileInTree(MNode) then
   begin
      if Assigned(Parent) then
      begin
         Node := TntTreeView1.Items.AddChild(Parent, MNode.Utf16);
      end
      else
      begin
         Node := TntTreeView1.Items.Add(nil, MNode.Utf16);
      end;
      Node.Data := MNode;
      MNode.TreeNode := Node;
      if MNode.FileType = 'D' then
      begin
         Node.ImageIndex := 1;
         Node.SelectedIndex := 1;
      end;
      Node.HasChildren := True; // we can look this up...
   end;
end;

procedure TExploreForm.UpdatePath(Node : TTntTreeNode; Path : String);
var
   S     : String;
   Error : String;
   FileType : String;
   Size     : String;
   UID      : String;
   GID      : String;
   Name     : String;
   MNode    : TMNode;
   PMNode   : TMNode;
   ctime    : String;
   mtime    : String;
   atime    : String;
   mode     : String;
begin
   //Node.Data
   PMNode := nil;
   if Assigned(Node) then
   begin
      PMNode := Node.Data;
   end;
   if not Assigned(PMNode) then
   begin
      Log('No parent MNode');
      exit;
   end;
   EndPoint.SendStringN('ls');
   Log(EndPoint.RecvStringN);
   EndPoint.SendStringN(Path);
   while true do
   begin
      S := EndPoint.RecvStringN;
      //Log(s);
      if StartsWith(S, 'T=', FileType) then
      begin
      end
      else if StartsWith(S, 'S=', Size) then
      begin
      end
      else if StartsWith(S, 'U=', UID) then
      begin
      end
      else if StartsWith(S, 'G=', GID) then
      begin
      end
      else if StartsWith(S, 'C=', ctime) then
      begin
      end
      else if StartsWith(S, 'M=', mtime) then
      begin
      end
      else if StartsWith(S, 'A=', atime) then
      begin
      end
      else if StartsWith(S, 'F=', mode) then
      begin
      end
      else if StartsWith(S, 'N=', Name) then
      begin
         MNode := TMNode.Create;
         MNode.Utf8 := Name;
         MNode.Utf16 := UTF8ToWideString(Name);
         MNode.Size := StrToInt64Def(Size, -1);
         MNode.FileType := FileType[1];
         MNode.uid := UID;
         MNode.gid := GID;
         MNode.ctime := ctime;
         MNode.mtime := mtime;
         MNode.atime := atime;
         MNode.mode  := mode;

         PMNode.AddChild(MNode);
         MNode.Parent := PMNode;
         AddToTree(Node, MNode);

         FileType := '';
         Size     := '';
         UID      := '';
         GID      := '';
         ctime    := '';
         mtime    := '';
         atime    := '';
         mode     := '';
      end
      else if StartsWith(S, 'E=', Error) then
      begin
         break;
      end;
   end;
   PMNode.DataLoaded := True;
end;

function TExploreForm.NodeToPath(Node : TTreeNode) : String;
begin
   Result := '';
   while Assigned(Node) do
   begin
      if Node.Text = '/' then
      begin
         Result := Node.Text + Result;
      end
      else
      begin
         Result := Node.Text + '/' + Result;
      end;
      Node := Node.Parent;
   end;
end;

function TExploreForm.MNodeToPath(MNode : TMNode) : String;
begin
   Result := '';
   while Assigned(MNode) do
   begin
      if MNode.Utf8 = '/' then
      begin
         Result := MNode.Utf8 + Result;
      end
      else
      begin
         if Result = '' then
         begin
            Result := MNode.Utf8;
         end
         else
         begin
            Result := MNode.Utf8 + '/' + Result;
         end;
      end;
      MNode := MNode.Parent;
   end;
end;

function TMNode.GetPath : String;
var
   MNode : TMNode;
begin
   MNode := self;

   Result := '';
   while Assigned(MNode) do
   begin
      if MNode.Utf8 = '/' then
      begin
         Result := MNode.Utf8 + Result;
      end
      else
      begin
         if Result = '' then
         begin
            Result := MNode.Utf8;
         end
         else
         begin
            Result := MNode.Utf8 + '/' + Result;
         end;
      end;
      MNode := MNode.Parent;
   end;
end;

procedure TExploreForm.TntTreeView1Expanding(Sender: TObject; Node: TTreeNode;
  var AllowExpansion: Boolean);
begin
   ExpandTreeNode(TTntTreeNode(Node));
end;

procedure TExploreForm.TntTreeView1Expanded(Sender: TObject;
  Node: TTreeNode);
begin
   SyncListView(Node.Data);
end;

procedure TExploreForm.ExpandTreeNode(Node: TTntTreeNode);
var
   Path : String;
   MNode : TMNode;
begin
   MNode := Node.Data;
   if not MNode.DataLoaded then
   begin
      TntTreeView1.Items.BeginUpdate;
      StartWait;
      try
         Path := MNode.GetPath;
         Log(Path);
         UpdatePath(TTntTreeNode(Node), Path);
         if Node.Count = 0 then
         begin
            Node.HasChildren := False;
         end;
         MNode.DataLoaded := True;
      finally
         EndWait;
         TntTreeView1.Items.EndUpdate;
      end;
   end;
end;


function TExploreForm.ShowFileInList(MNode : TMNode) : Boolean;
begin
   Result := False;
   if MNode.Utf8 = '.' then
   begin
      // do nothing
   end
   else if MNode.Utf8 = '..' then
   begin
      // do nothing
   end
   else if MNode.FileType in ['F','S','C','B'] then
   begin
      Result := True;
   end
   else if ShowDirectoriesInList and (MNode.FileType in ['D','M']) then
   begin
      Result := True;
   end;

   if Result and not ShowDotFiles then
   begin
      if MNode.Utf8[1] = '.' then
      begin
         Result := False;
      end;
   end;
end;

function TExploreForm.ShowFileInTree(MNode : TMNode) : Boolean;
begin

   Result := False;
   if MNode.Utf8 = '.' then
   begin
      // do nothing
   end
   else if MNode.Utf8 = '..' then
   begin
      // do nothing
   end
   else if MNode.FileType in ['D','M'] then
   begin
      Result := True;
   end;

   if Result and not ShowDotFiles then
   begin
      if MNode.Utf8[1] = '.' then
      begin
         Result := False;
      end;
   end;
end;

procedure TExploreForm.ReSyncListView;
var
   PMNode : TMNode;
begin
   PMNode := CurrentListedMNode;
   CurrentListedMNode := nil;
   SyncListView(PMNode);
end;

procedure TExploreForm.SyncListView(PMNode : TMNode);
var
   Item  : TTntListItem;
   i     : Integer;
   MNode : TMNode;
begin
   if CurrentListedMNode <> PMNode then
   begin
      TntListView1.Items.BeginUpdate;
      StartWait;
      try
         TntListView1.Items.Clear;

         if Assigned(PMNode) then
         begin
            ExpandTreeNode(PMNode.TreeNode);
            for i := 0 to PMNode.Children.Count - 1 do
            begin
               MNode := PMNode.GetChild(i);
               if ShowFileInList(MNode) then
               begin
                  Item := TntListView1.Items.Add;
                  Item.Data := MNode;
                  Item.Caption := MNode.Utf16;
                  if MNode.Size >= 0 then
                  begin
                     Item.SubItems.Add(IntToStr(MNode.Size));
                  end
                  else
                  begin
                     Item.SubItems.Add('');
                  end;

                  // Type
                  if MNode.FileType = 'F' then
                  begin
                     Item.ImageIndex := 2;
                     Item.SubItems.Add('File');
                  end
                  else if MNode.FileType = 'D' then
                  begin
                     Item.ImageIndex := 1;
                     Item.SubItems.Add('Directory');
                  end
                  else if MNode.FileType = 'M' then
                  begin
                     Item.ImageIndex := 0;
                     Item.SubItems.Add('Mount Point');
                     // load passwd & group ?
                     try
                        MNode.PasswdFileName := SaveFile(MNode.GetPath + '/etc/passwd');
                        MNode.GroupFileName  := SaveFile(MNode.GetPath + '/etc/group');
                        MNode.LoadPasswd;
                     except
                        MNode.Passwd := nil;
                     end;
                  end
                  else if MNode.FileType = 'B' then
                  begin
                     Item.ImageIndex := 0;
                     Item.SubItems.Add('Block Device');
                  end
                  else
                  begin
                     Item.ImageIndex := 2;
                     Item.SubItems.Add('Unknown');
                  end;

                  // Modified
                  Item.SubItems.Add(MNode.mtime);

                  // User
                  Item.SubItems.Add(MNode.GetUser);

                  // Group
                  Item.SubItems.Add(MNode.GetGroup);

                  // Attributes
                  Item.SubItems.Add(MNode.Mode);
               end;
            end;
         end;
         CurrentListedMNode := PMNode;
      finally
         EndWait;
         TntListView1.Items.EndUpdate;
      end;
   end;
end;

procedure TExploreForm.TntTreeView1Change(Sender: TObject;
  Node: TTreeNode);
begin
   SyncListView(TntTreeView1.Selected.Data);
end;

procedure TExploreForm.OpenFile(FileName : String);
var
   r : Integer;
   StartupInfo : TStartupInfo;
   ProcessInfo : TProcessInformation;
begin
   r := ShellExecute(Handle, nil, PChar(FileName), nil, nil, SW_SHOWDEFAULT);
   if R <= 32 then
   begin
      // there was an error
      if r = SE_ERR_NOASSOC then
      begin
         // run notepad?
         ZeroMemory(@StartupInfo, sizeof(StartupInfo));
         FileName := ViewerProg + ' ' + FileName;
         if not CreateProcess(nil, PChar(FileName), nil, nil, False, CREATE_NEW_PROCESS_GROUP, nil, nil, StartupInfo, ProcessInfo) then
         begin
            MessageDlg('Error ' + IntToStr(GetLastError) + ' viewing file', mtError, [mbOK], 0);
         end;
      end
      else
      begin
         MessageDlg('Error ' + IntToStr(R) + ' viewing file', mtError, [mbOK], 0);
      end;
   end;
end;

procedure TExploreForm.TntListView1DblClick(Sender: TObject);
var
   Item  : TTntListItem;
   MNode : TMNode;
   TempFile : String;
begin
   if TntListView1.SelCount = 1 then
   begin
      Item := TntListView1.Selected;
      MNode := Item.Data;
      if Assigned(MNode.TreeNode) then
      begin
         TntTreeView1.Selected := MNode.TreeNode;
      end
      else
      begin
         if MNode.FileType = 'F' then
         begin
            // can we get the data...
            TempFile := SaveFile(MNodeToPath(MNode));
            OpenFile(TempFile);
         end;
      end;

   end;
end;


function TExploreForm.SaveFile(Path : String) : String;
var
   Length   : Integer;
   Buffer   : String;
   Error    : String;
   Output   : TBinaryFile;
begin
   EndPoint.SendStringN('get');
   Log(EndPoint.RecvStringN);
   EndPoint.SendStringN(Path);
   while true do
   begin
      Buffer := EndPoint.RecvStringN;
      if StartsWith(Buffer, 'E=', Error) then
      begin
         if Error = 'Binary' then
         begin
            Output := TBinaryFile.Create;
            Output.CreateTemp('vv', ExtractFileExt(Path));
            Result := Output.GetFileName;
            while true do
            begin
               Length := EndPoint.RecvInteger;
               if Length = 0 then
               begin
                  break;
               end;
               SetLength(Buffer, Length);
               EndPoint.RecvBuffer(PChar(Buffer), Length);
               Output.BlockWrite2(PChar(Buffer), Length);
               //Log(Buffer);
            end;
            Output.Close;
            Output.Free;
         end
         else
         begin
            break;
         end;
      end;
   end;

end;

// Drag & Drop Source implementation

function TExploreForm.MangleFileName(OldName : WideString) : WideString;
var
   i : Integer;
begin
   Result := OldName;
   for i := 1 to Length(Result) do
   begin
      if Result[i] = ':' then
      begin
         Result[i] := '=';
      end
      else if Result[i] = '\' then
      begin
         Result[i] := '_';
      end;
   end;
end;

function TExploreForm.CreateFileGroupDescriptor : HGLOBAL;
var
   nItems   : Integer;
   Size     : Integer;
   hg       : HGLOBAL;
   pfgd     : PFileGroupDescriptorW2;
   i        : integer;
   MNode    : TMNode;
   MName    : WideString;
   Temp     : LARGE_INTEGER;
   Child    : TMNode;
begin
   // Expand FilesToDrop...
   if not FilesToDropExpanded then
   begin
      Log('Expanding FilesToDrop');
      i := 0;
      while i < FilesToDrop.Count do
      begin
         MNode := TMNode(FilesToDrop.Objects[i]);
         if MNode.FileType in ['D', 'M'] then
         begin
            if Assigned(MNode.TreeNode) then
            begin
               ExpandTreeNode(MNode.TreeNode);
               //FilesToDrop[i] := FilestoDrop[i] + '\';
               // make sure we have the data...
               for nItems := 0 to MNode.Children.Count - 1 do
               begin
                  Child := TMNode(MNode.Children[nItems]);
                  if Child.Utf8 = '.' then
                  begin
                  end
                  else if Child.Utf8 = '..' then
                  begin
                  end
                  else
                  begin
                     Log('Adding ' + Child.Utf16);
                     MName := FilesToDrop[i] + '\' + MangleFileName(Child.Utf16);
                     FilesToDrop.AddObject(MName, Child);
                  end;
               end;
            end;
         end;
         i := i + 1;
      end;
      FilesToDropExpanded := True;
   end;


   //
   nItems := FilesToDrop.Count;

   Size := sizeof(FILEGROUPDESCRIPTORW) + nItems * sizeof(FILEDESCRIPTORW);
   hg := GlobalAlloc(GMEM_MOVEABLE or GMEM_SHARE, Size);
   if hg <> 0 then
   begin
      pfgd := PFileGroupDescriptorW2(GlobalLock(hg));
      if Assigned(pfgd) then
      begin
         FillMemory(pfgd, Size, 0);
         pfgd.cItems := nItems;
         for i := 0 to nItems - 1 do
         begin
            MNode := TMNode(FilesToDrop.Objects[i]);
            MName := FilesToDrop[i];

            if MNode.FileType in ['F'] then
            begin
               pfgd.fgd[i].dwFlags := FD_ATTRIBUTES;
               pfgd.fgd[i].dwFileAttributes := FILE_ATTRIBUTE_NORMAL;

               pfgd.fgd[i].dwFlags :=  pfgd.fgd[i].dwFlags or FD_FILESIZE;
               Temp.QuadPart := MNode.Size;
               pfgd.fgd[i].nFileSizeLow := Temp.LowPart;
               pfgd.fgd[i].nFileSizeHigh := Temp.HighPart;

               pfgd.fgd[i].dwFlags := pfgd.fgd[i].dwFlags or FD_PROGRESSUI;

               StrLCopyW(@pfgd.fgd[i].cFileName[0], MName, sizeof(pfgd.fgd[i].cFileName));
            end
            else if MNode.FileType in ['D', 'M'] then
            begin
               pfgd.fgd[i].dwFlags := FD_ATTRIBUTES;
               pfgd.fgd[i].dwFileAttributes := FILE_ATTRIBUTE_DIRECTORY;

               pfgd.fgd[i].dwFlags := pfgd.fgd[i].dwFlags or FD_PROGRESSUI;

               StrLCopyW(@pfgd.fgd[i].cFileName[0], MName, sizeof(pfgd.fgd[i].cFileName));
            end
            else
            begin
            end
         end;

         GlobalUnlock(hg);
      end;
   end;
   Result := hg;
end;

function TExploreForm.CreateFileContents(Index : Integer) : TSendData;
var
   Data : TMirrorData;
begin
   Data := TMirrorData.Create(self);
   Data.SetData(Endpoint, TMNode(FilesToDrop.Objects[Index]));
   Result := Data;
end;

procedure TExploreForm.StartListDrag;
var
   MNode : TMNode;
   i : Integer;
begin
   FilesToDrop := TTntStringList.Create;
   FilesToDropExpanded := False;
   try

      for i := 0 to TntListView1.Items.Count - 1 do
      begin
         if TntListView1.Items[i].Selected then
         begin
            MNode := TntListView1.Items[i].Data;
            FilesToDrop.AddObject(MNode.Utf16, MNode);
         end
      end;

      DragDrop := TDropSource.Create(self);
      try
         DragDrop.fOnCreateFileGroup    := CreateFileGroupDescriptor;
         DragDrop.fOnCreateFileContents := CreateFileContents;
         if Copy then
         begin
            DragDrop.CopyToClipboard;
         end
         else
         begin
            DragDrop.Execute;
         end;
      finally
         if not Copy then
         begin
            DragDrop.Free;
            DragDrop := nil;
         end;
      end;
   finally
      if not Copy then
      begin
         FilesToDrop.Free;
         FilesToDrop := nil;
      end;
   end;
end;

// Copy = False : Drag & Drop
// Copy = True  : Edit -> Copy
procedure TExploreForm.StartTreeDrag(Copy : Boolean);
var
   MNode    : TMNode;
begin
   FilesToDrop := TTntStringList.Create;
   FilesToDropExpanded := False;
   try
      MNode := TntTreeView1.Selected.Data;
      FilesToDrop.AddObject(MNode.Utf16, MNode);

      DragDrop := TDropSource.Create(self);
      try
         DragDrop.fOnCreateFileGroup    := CreateFileGroupDescriptor;
         DragDrop.fOnCreateFileContents := CreateFileContents;
         if Copy then
         begin
            DragDrop.CopyToClipboard;
         end
         else
         begin
            DragDrop.Execute;
         end;
      finally
         if not Copy then
         begin
            DragDrop.Free;
            DragDrop := nil;
         end;
      end;
   finally
      if not Copy then
      begin
         FilesToDrop.Free;
         FilesToDrop := nil;
      end;
   end;
end;
// End Drag & Drop

procedure TExploreForm.TntListView1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
   if ssLeft in Shift then
   begin
      if StartDragging then
      begin
         if TntListView1.SelCount > 0 then
         begin
            StartListDrag(False);
         end;
      end
      else
      begin
         StartDragging := True;
      end;
   end
   else
   begin
      StartDragging := False;
   end;
end;

// TMirrorData is used to send data from vv to a droptarget

constructor TMirrorData.Create(aowner: TComponent);
begin
   inherited;

   EndPoint := nil;
   MNode := nil;
end;

destructor TMirrorData.Destroy;
begin
   MNode.CloseData;
   inherited;
end;
procedure TMirrorData.SetData(fEndpoint : TNetEndpoint; fMNode : TMNode);
begin
   Endpoint := fEndpoint;
   MNode    := fMNode;
   Size     := MNode.Size;

   MNode.OpenData(Endpoint);
end;

function TMirrorData.ReadData(Buffer : PChar; Offset : Int64; Length : Int64) : Boolean;
begin
   if Length = 0 then
   begin
      Result := True;
      exit;
   end;

   Result := MNode.ReadData(Buffer, Offset, Length);
end;

procedure TExploreForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
   if ShutdownServer then
   begin
      EndPoint.SendStringN('shutdown');
   end
   else
   begin
      EndPoint.SendStringN('exit');
   end;
   EndPoint.Shutdown(SD_BOTH);
end;

procedure TExploreForm.TntTreeView1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
   if ssLeft in Shift then
   begin
      if StartDragging then
      begin
         if Assigned(TntTreeView1.Selected) then
         begin
            StartTreeDrag(False);
         end;
      end
      else
      begin
         StartDragging := True;
      end;
   end
   else
   begin
      StartDragging := False;
   end;
end;

procedure TExploreForm.TntListView1CustomDrawSubItem(
  Sender: TCustomListView; Item: TListItem; SubItem: Integer;
  State: TCustomDrawState; var DefaultDraw: Boolean);
begin
   if SubItem = 6 then
   begin
      DefaultDraw := True;
      Sender.Canvas.Font.Name := 'Courier';
   end
   else
   begin
      Sender.Canvas.Font.Name := Font.Name;
      DefaultDraw := True;
   end;
end;

procedure TExploreForm.TntListView1ColumnDragged(Sender: TObject);
begin
   TntListView1.Invalidate;
end;

procedure TExploreForm.SendMountRequest(MNode : TMNode; Mount : String);
var
   S     : String;
   Error : String;
begin
   EndPoint.SendStringN('mount');
   Log(EndPoint.RecvStringN);
   EndPoint.SendStringN(MNode.GetPath);
   //Log(EndPoint.RecvStringN);
   EndPoint.SendStringN(Mount);
   S := EndPoint.RecvStringN;
   Log(S);
   if StartsWith(S, 'E=', Error) then
   begin
      if Error = 'Done' then
      begin
         // We need to invalidate the tree below MNode...
         MNode.PruneChildren;
         MNode.TreeNode.DeleteChildren;
         MNode.TreeNode.HasChildren := True;
         MNode.DataLoaded := False;
         MNode.FileType := 'M';
         MNode.TreeNode.ImageIndex := 0;
         MNode.TreeNode.SelectedIndex := 0;

         try
            MNode.PasswdFileName := SaveFile(MNode.GetPath + '/etc/passwd');
            MNode.GroupFileName  := SaveFile(MNode.GetPath + '/etc/group');
            MNode.LoadPasswd;
         except
            MNode.Passwd := nil;
         end;

         ReSyncListView;
      end;
   end;
end;

procedure TExploreForm.SendLoopbackRequest(MNode : TMNode);
var
   S     : String;
   Error : String;
begin
   EndPoint.SendStringN('load');
   Log(EndPoint.RecvStringN);
   EndPoint.SendStringN(MNode.GetPath);
   S := EndPoint.RecvStringN;
   Log(S);
   if StartsWith(S, 'E=', Error) then
   begin
      if Error = 'Done' then
      begin
         MNode.FileType := 'B';
         ReSyncListView;
      end;
   end;
end;

procedure TExploreForm.Mount1Click(Sender: TObject);
var
   MNode : TMNode;
   MountForm : TMountForm;
begin
   if Assigned(TntTreeView1.Selected) then
   begin
      MNode := TntTreeView1.Selected.Data;

      MountForm := TMountForm.Create(self);
      try
         MountForm.Endpoint := EndPoint;
         MountForm.MountPointEdit.Text := MNode.GetPath + '/';
         if MountForm.ShowModal = mrOK then
         begin
            SendMountRequest(MNode, MountForm.GetMountString);
         end;
      finally
         MountForm.Free;
      end;

   end;
end;

procedure TExploreForm.Copy1Click(Sender: TObject);
begin
   if Assigned(TntTreeView1.Selected) then
   begin
      StartTreeDrag(True);
   end;
end;

function TExploreForm.SupportClipboardFormat : Boolean;
var
   DataObject  : IDataObject;
   Ret         : HRESULT;
begin
   Result := False;

   Ret := OleGetClipboard(DataObject);
   if Ret = S_OK then
   begin
      if Assigned(DataObject) then
      begin
         Log('Got DataObject');

         if DropTarget.PrivateFormat(DataObject) then
         begin
            Result := False;
         end
         else
         begin
            Result := DropTarget.SupportedFormat(DataObject);
         end;
      end;
   end;
end;

procedure TExploreForm.TreePopupMenuPopup(Sender: TObject);
begin
   if SupportClipboardFormat then
   begin
      Paste1.Enabled := True;
   end
   else
   begin
      Paste1.Enabled := False;
   end;
end;

procedure TExploreForm.Paste1Click(Sender: TObject);
var
   DataObject  : IDataObject;
   Ret         : HRESULT;
begin

   Ret := OleGetClipboard(DataObject);
   if Ret = S_OK then
   begin
      DropTarget.ProcessData(TntTreeView1.Handle, DataObject);
   end;
end;

procedure TExploreForm.ConvertDropfilesToList(DropFiles : PDropFiles; List : TTntStringList);
var
   Start : PChar;
   WS : WideString;
   S : String;
begin
   Start := PChar(DropFiles) + DropFiles.pFiles;
   while Start[0] <> #0 do
   begin
      if DropFiles.fWide then
      begin
         WS := WideString((PWideChar(Start)));
         List.Add(WS);
         Start := Start + ((Length(WS) + 1) * sizeof(WideChar));
      end
      else
      begin
         S := String(Start);
         List.Add(S);
         Start := Start + Length(S) + 1;
      end;
   end;
end;

function TExploreForm.ImportFileContents(Window : HWND; Name : WideString; Length : Int64; Data : IStream) : Boolean;
var
   Buffer   : String;
   Done     : Int64;
   i        : Integer;
   Actual   : Integer;
   Pos      : Largeint;
   PMNode   : TMNode;
   MNode    : TMNode;
   FileName : String;
begin
   if Window = TntTreeView1.Handle then
   begin
      PMNode := TntTreeView1.Selected.Data;
   end
   else if Window = TntListView1.Handle then
   begin
      PMNode := CurrentListedMNode;
   end
   else
   begin
      Log('unknown window');
      Result := False;
      exit;
   end;


   Done := 0;
   if Assigned(Data) then
   begin
      FileName := PMNode.GetPath + '/' + WideStringToUTF8(Name);
      Log('import ' + FileName);
      if PMNode.CreateFile(Endpoint, FileName) then
      begin
         // we will do a new ls later
         // for now just hack in the new MNode
         MNode := TMNode.Create;
         MNode.Utf8 := WideStringToUTF8(Name);
         MNode.Utf16 := Name;
         MNode.Size := 0;
         MNode.FileType := 'F';

         PMNode.AddChild(MNode);
         MNode.Parent := PMNode;

         if MNode.OpenData(Endpoint) then
         begin
            Data.Seek(0, STREAM_SEEK_SET, Pos);
            while Done < Length do
            begin
               i := 4096;
               if Done + i > Length then
               begin
                  i := Length - Done;
               end;
               SetLength(Buffer, i);
               Data.Read(PChar(Buffer), i, @Actual);
               if Actual = 0 then
               begin
                  break;
               end
               else
               begin
                  MNode.WriteData(PChar(Buffer), Done, Actual);
                  Done := Done + Actual;
               end;
            end;
            MNode.CloseData;

            CurrentListedMNode := nil;
            SyncListView(PMNode);
         end;
      end;
   end
   else
   begin
      Log('mkdir ' + Name);
   end;
   Result := True;
end;

procedure TExploreForm.SyncViewMenu;
var
   i : Integer;
begin
   ReverseOrder1.Checked := SortBackwards;
   CaseSensitive1.Checked := SortCase;

   for i := 0 to ArrangeIconsby1.Count - 1 do
   begin
      with ArrangeIconsby1.Items[i] do
      begin
         if RadioItem then
         begin
            if Tag = SortColumn then
            begin
               Checked := True;
            end
            else
            begin
               Checked := False;
            end;
         end;
      end;
   end;

   Toolbar1.Checked   := ShowToolbar;
   Statusbar1.Checked := ShowStatusbar;
   Debug1.Checked     := ShowDebug;
   
end;

procedure TExploreForm.TntListView1ColumnClick(Sender: TObject;
  Column: TListColumn);
begin
   Log('Change sort type to ' + IntToStr(Column.Index));

   if SortColumn = Column.Index then
   begin
      SortBackwards := not SortBackwards;
   end
   else
   begin
      SortColumn := Column.Index;
   end;

   TntListView1.SortType := stNone;
   TntListView1.SortType := stData;

   SyncViewMenu;
end;

procedure TExploreForm.TntListView1Compare(Sender: TObject; Item1,
  Item2: TListItem; Data: Integer; var Compare: Integer);
begin
   if SortColumn = 0 then
   begin
      if SortCase then
      begin
         Compare := CompareStr(Item1.Caption, Item2.Caption);
      end
      else
      begin
         Compare := CompareText(Item1.Caption, Item2.Caption);
      end;
   end
   else if SortColumn = 1 then // size
   begin
      Compare := TMNode(Item1.Data).Size - TMNode(Item2.Data).Size;
   end
   else
   begin
      if SortCase then
      begin
         Compare := CompareStr(Item1.SubItems[SortColumn - 1], Item2.SubItems[SortColumn - 1]);
      end
      else
      begin
         Compare := CompareText(Item1.SubItems[SortColumn - 1], Item2.SubItems[SortColumn - 1]);
      end;
   end;

   if (SortColumn > 0) and (Compare = 0) then
   begin
      // secondary sort on name
      if SortCase then
      begin
         Compare := CompareStr(Item1.Caption, Item2.Caption);
      end
      else
      begin
         Compare := CompareText(Item1.Caption, Item2.Caption);
      end;
   end;

   if SortBackwards then
   begin
      Compare := - Compare;
   end;
end;

procedure TExploreForm.TntListView1ColumnRightClick(Sender: TObject;
  Column: TListColumn; Point: TPoint);
begin
   SortCase := not SortCase;

   TntListView1.SortType := stNone;
   TntListView1.SortType := stData;

   SyncViewMenu;
end;

procedure TExploreForm.L1Click(Sender: TObject);
var
   Item  : TTntListItem;
   MNode : TMNode;
begin
   if TntListView1.SelCount = 1 then
   begin
      Item := TntListView1.Selected;
      MNode := Item.Data;
      if MNode.FileType = 'F' then
      begin
         SendLoopbackRequest(MNode);
      end;
   end;
end;

procedure TExploreForm.ReverseOrder1Click(Sender: TObject);
begin
   SortBackwards := not SortBackwards;

   TntListView1.SortType := stNone;
   TntListView1.SortType := stData;

   SyncViewMenu;
end;

procedure TExploreForm.CaseSensitive1Click(Sender: TObject);
begin
   SortCase := not SortCase;

   TntListView1.SortType := stNone;
   TntListView1.SortType := stData;

   SyncViewMenu;
end;

procedure TExploreForm.Name1Click(Sender: TObject);
begin
   with Sender as TMenuItem do
   begin
      SortColumn := Tag;
   end;

   TntListView1.SortType := stNone;
   TntListView1.SortType := stData;

   SyncViewMenu;
end;

procedure TExploreForm.LargeIcons1Click(Sender: TObject);
begin
   TntListView1.ViewStyle := vsIcon;
   LargeIcons1.Checked := True;
end;

procedure TExploreForm.SmallIcons1Click(Sender: TObject);
begin
   TntListView1.ViewStyle := vsSmallIcon;
   SmallIcons1.Checked := True;
end;

procedure TExploreForm.List1Click(Sender: TObject);
begin
   TntListView1.ViewStyle := vsList;
   List1.Checked := True;
end;

procedure TExploreForm.Details1Click(Sender: TObject);
begin
   TntListView1.ViewStyle := vsReport;
   Details1.Checked := True;
end;

procedure TExploreForm.Toolbar1Click(Sender: TObject);
begin
   ShowToolbar := not ShowToolbar;
   ToolbarPanel.Visible := ShowToolbar;

   SyncViewMenu;
end;

procedure TExploreForm.Statusbar1Click(Sender: TObject);
begin
   ShowStatusbar := not ShowStatusbar;
   TntStatusBar1.Visible := ShowStatusbar;

   SyncViewMenu;
end;

procedure TExploreForm.Debug1Click(Sender: TObject);
begin
   ShowDebug := not ShowDebug;

   if ShowDebug then
   begin
      Memo1.Align := alBottom;
      Memo1.Visible := True;
      DebugSplitter := TSplitter.Create(Panel2);
      DebugSplitter.Parent := Panel2;
      DebugSplitter.Align := alBottom;
   end
   else
   begin
      Memo1.Align := alNone;
      Memo1.Visible := False;
      DebugSplitter.Free;
      DebugSplitter := nil;
   end;

   SyncViewMenu;
end;

procedure TExploreForm.About1Click(Sender: TObject);
var
   About : TAboutBox;
begin
   About := TAboutBox.Create(self);
   About.ShowModal;
   About.Free;
end;

end.
