unit mount;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, NetLib;

type
  TMountForm = class(TForm)
    Label1: TLabel;
    MountPointEdit: TEdit;
    Label2: TLabel;
    MountComboBox: TComboBox;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    Button1: TButton;
    Button2: TButton;
    Label3: TLabel;
    Label4: TLabel;
    PlinkEdit: TEdit;
    Button3: TButton;
    OpenDialog1: TOpenDialog;
    SftpSession: TComboBox;
    TabSheet2: TTabSheet;
    Label5: TLabel;
    Ext2volume: TComboBox;
    ext2writeable: TCheckBox;
    Label6: TLabel;
    Label7: TLabel;
    Ext2UUID: TComboBox;
    Ext2Label: TComboBox;
    Label8: TLabel;
    SftpUsername: TEdit;
    Label9: TLabel;
    SftpPassword: TEdit;
    TabSheet3: TTabSheet;
    Label10: TLabel;
    Label11: TLabel;
    Label12: TLabel;
    ReiserVolume: TComboBox;
    ReiserUUID: TComboBox;
    ReiserLabel: TComboBox;
    TabSheet4: TTabSheet;
    Label13: TLabel;
    Label14: TLabel;
    Label15: TLabel;
    fatvolume: TComboBox;
    fatuuid: TComboBox;
    fatlabel: TComboBox;
    fatwritable: TCheckBox;
    TabSheet5: TTabSheet;
    Label16: TLabel;
    Win32PathEdit: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure TabSheet1Show(Sender: TObject);
    procedure MountComboBoxChange(Sender: TObject);
    procedure TabSheet2Show(Sender: TObject);
    procedure Ext2volumeChange(Sender: TObject);
    procedure Ext2UUIDChange(Sender: TObject);
    procedure PageControl1Change(Sender: TObject);
    procedure TabSheet3Show(Sender: TObject);
    procedure ReiserVolumeChange(Sender: TObject);
    procedure ReiserUUIDChange(Sender: TObject);
    procedure TabSheet4Show(Sender: TObject);
    procedure fatvolumeChange(Sender: TObject);
    procedure fatuuidChange(Sender: TObject);
  private
    { Private declarations }
    procedure LoadPuttySessions;
    procedure LoadExt2Filesystems;
    procedure SelectExt2Filesystem(Index : Integer);
    procedure LoadReiserFilesystems;
    procedure SelectReiserFilesystem(Index : Integer);
    procedure LoadFatFilesystems;
    procedure SelectFatFilesystem(Index : Integer);
  public
    { Public declarations }
    Endpoint : TNetEndpoint;   

    function GetMountString : String;
  end;

var
  MountForm: TMountForm;

implementation

uses debug, strings;

{$R *.DFM}

function TMountForm.GetMountString : String;
begin
   if MountComboBox.Text = 'SFTP' then
   begin
      Result := 'sftp ' + SftpSession.Text;
      if PlinkEdit.Text <> 'Default' then
      begin
         Result := Result + ' plink="' + PlinkEdit.Text + '"';
      end;
      if Length(SftpUsername.Text) > 0 then
      begin
         Result := Result + ' username="' + SftpUsername.Text + '"';
      end;
      if Length(SftpPassword.Text) > 0 then
      begin
         Result := Result + ' password="' + SftpPassword.Text + '"';
      end;
   end
   else if MountComboBox.Text = 'Ext2/Ext3' then
   begin
      Result := 'ext2 ' + Ext2volume.Text;
      if ext2writeable.Checked then
      begin
         Result := Result + ' rw';
      end;
   end
   else if MountComboBox.Text = 'ReiserFS' then
   begin
      Result := 'reiserfs ' + ReiserVolume.Text;
   end
   else if MountComboBox.Text = 'FAT' then
   begin
      Result := 'fat ' + FatVolume.Text;
   end
   else if MountComboBox.Text = 'Win32' then
   begin
      Result := 'win32 ' + Win32PathEdit.Text;
   end
   else
   begin
      Log('Unimplemented mount type');
   end;
end;

procedure TMountForm.FormCreate(Sender: TObject);
begin
   MountComboBox.ItemIndex := 0;
   PageControl1.ActivePageIndex := 0;
end;

procedure TMountForm.LoadPuttySessions;
var
   Reg      : hkey;
   KeyPath  : String;
   i        : Integer;
   Buffer   : String;
begin
   KeyPath := 'SOFTWARE\SimonTatham\PuTTY\Sessions';
   if RegOpenKeyEx(HKEY_CURRENT_USER, PChar(KeyPath), 0, KEY_READ, Reg) = ERROR_SUCCESS then
   begin
      i := 0;
      while true do
      begin
         SetLength(Buffer, MAX_PATH + 1);
         if RegEnumKey(Reg, i, PChar(Buffer), Length(Buffer)) = ERROR_SUCCESS then
         begin
            SetLength(Buffer, strlen(PChar(Buffer)));
            SftpSession.Items.Add(Buffer);
            Log(Buffer);
            i := i + 1;
         end
         else
         begin
            break;
         end;
      end;
      RegCloseKey(Reg);
   end;
end;

procedure TMountForm.Button3Click(Sender: TObject);
begin
   if OpenDialog1.Execute then
   begin
      PLinkEdit.Text := OpenDialog1.FileName;
   end;
end;

procedure TMountForm.TabSheet1Show(Sender: TObject);
begin
   SftpSession.SetFocus;

   if SftpSession.Items.Count = 0 then
   begin
      LoadPuttySessions;
   end;
end;

procedure TMountForm.MountComboBoxChange(Sender: TObject);
var
   i : Integer;
begin
   for i := 0 to PageControl1.PageCount - 1 do
   begin
      if PageControl1.Pages[i].Caption = MountComboBox.Text then
      begin
         PageControl1.ActivePageIndex := i;
         break;
      end;
   end;
end;

procedure TMountForm.TabSheet2Show(Sender: TObject);
begin
   ext2volume.SetFocus;

   LoadExt2Filesystems;
   SelectExt2Filesystem(0);
end;

procedure TMountForm.LoadExt2Filesystems;
var
   S     : String;
   Error : String;
   i     : Integer;
begin
   Ext2volume.Items.Clear;
   Ext2UUID.Items.Clear;
   Ext2Label.Items.Clear;

   i := 0;

   // I am breaking the rules here and storing the data in the interface
   // I will come back and fix it later

   Endpoint.SendStringN('show');
   Log(EndPoint.RecvStringN);
   EndPoint.SendStringN('FS=ext2');
   while true do
   begin
      S := EndPoint.RecvStringN;
      Log(S);
      if StartsWith(S, 'E=', Error) then
      begin
         break;
      end
      else if StartsWith(S, 'V=', Error) then
      begin
         i := Ext2volume.Items.Count;
         Ext2volume.Items.Add(Error);
      end
      else if StartsWith(S, 'M=', Error) then
      begin
         if Error = 'rw' then
         begin
            Ext2volume.Items.Objects[i] := TObject(1);
         end;
      end
      else if StartsWith(S, 'A=UUID=', Error) then
      begin
         Ext2UUID.Items.AddObject(Error, TObject(i));
      end
      else if StartsWith(S, 'A=LABEL=', Error) then
      begin
         Ext2Label.Items.AddObject(Error, TObject(i));
      end;
   end;

end;

procedure TMountForm.LoadReiserFilesystems;
var
   S     : String;
   Error : String;
   i     : Integer;
begin
   ReiserVolume.Items.Clear;
   ReiserUUID.Items.Clear;
   ReiserLabel.Items.Clear;

   i := 0;

   // I am breaking the rules here and storing the data in the interface
   // I will come back and fix it later

   Endpoint.SendStringN('show');
   Log(EndPoint.RecvStringN);
   EndPoint.SendStringN('FS=reiserfs');
   while true do
   begin
      S := EndPoint.RecvStringN;
      Log(S);
      if StartsWith(S, 'E=', Error) then
      begin
         break;
      end
      else if StartsWith(S, 'V=', Error) then
      begin
         i := ReiserVolume.Items.Count;
         ReiserVolume.Items.Add(Error);
      end
      else if StartsWith(S, 'A=UUID=', Error) then
      begin
         ReiserUUID.Items.AddObject(Error, TObject(i));
      end
      else if StartsWith(S, 'A=LABEL=', Error) then
      begin
         ReiserLabel.Items.AddObject(Error, TObject(i));
      end;
   end;

end;

procedure TMountForm.LoadFatFilesystems;
var
   S     : String;
   Error : String;
   i     : Integer;
begin
   FatVolume.Items.Clear;
   FatUUID.Items.Clear;
   FatLabel.Items.Clear;

   i := 0;

   // I am breaking the rules here and storing the data in the interface
   // I will come back and fix it later

   Endpoint.SendStringN('show');
   Log(EndPoint.RecvStringN);
   EndPoint.SendStringN('FS=fat');
   while true do
   begin
      S := EndPoint.RecvStringN;
      Log(S);
      if StartsWith(S, 'E=', Error) then
      begin
         break;
      end
      else if StartsWith(S, 'V=', Error) then
      begin
         i := FatVolume.Items.Count;
         FatVolume.Items.Add(Error);
      end
      else if StartsWith(S, 'A=UUID=', Error) then
      begin
         FatUUID.Items.AddObject(Error, TObject(i));
      end
      else if StartsWith(S, 'A=LABEL=', Error) then
      begin
         FatLabel.Items.AddObject(Error, TObject(i));
      end;
   end;

end;

procedure TMountForm.SelectExt2Filesystem(Index : Integer);
var
   i : Integer;
begin
   Ext2volume.ItemIndex := Index;

   for i := 0 to Ext2UUID.Items.Count - 1 do
   begin
      if Integer(Ext2UUID.Items.Objects[i]) = Index then
      begin
         Ext2UUID.ItemIndex := i;
         break;
      end;
   end;

   for i := 0 to Ext2Label.Items.Count - 1 do
   begin
      if Integer(Ext2Label.Items.Objects[i]) = Index then
      begin
         Ext2Label.ItemIndex := i;
         break;
      end;
   end;

   if Ext2volume.Items.Count > 0 then
   begin
      if Integer(Ext2volume.Items.Objects[Index]) = 1 then
      begin
         ext2writeable.Enabled := True;
      end
      else
      begin
         ext2writeable.Checked := False;
         ext2writeable.Enabled := False;
      end;
   end;
end;

procedure TMountForm.SelectReiserFilesystem(Index : Integer);
var
   i : Integer;
begin
   ReiserVolume.ItemIndex := Index;

   for i := 0 to ReiserUUID.Items.Count - 1 do
   begin
      if Integer(ReiserUUID.Items.Objects[i]) = Index then
      begin
         ReiserUUID.ItemIndex := i;
         break;
      end;
   end;

   for i := 0 to ReiserLabel.Items.Count - 1 do
   begin
      if Integer(ReiserLabel.Items.Objects[i]) = Index then
      begin
         ReiserLabel.ItemIndex := i;
         break;
      end;
   end;
end;

procedure TMountForm.SelectFatFilesystem(Index : Integer);
var
   i : Integer;
begin
   FatVolume.ItemIndex := Index;

   FatUUID.Text := '';
   for i := 0 to FatUUID.Items.Count - 1 do
   begin
      if Integer(FatUUID.Items.Objects[i]) = Index then
      begin
         FatUUID.ItemIndex := i;
         break;
      end;
   end;

   FatLabel.Text := '';
   for i := 0 to FatLabel.Items.Count - 1 do
   begin
      if Integer(FatLabel.Items.Objects[i]) = Index then
      begin
         FatLabel.ItemIndex := i;
         break;
      end;
   end;
end;

procedure TMountForm.Ext2volumeChange(Sender: TObject);
begin
   SelectExt2Filesystem(Ext2volume.ItemIndex);
end;

procedure TMountForm.Ext2UUIDChange(Sender: TObject);
begin
   with Sender as TComboBox do
   begin
      SelectExt2Filesystem(Integer(Items.Objects[ItemIndex]));
   end;
end;

procedure TMountForm.PageControl1Change(Sender: TObject);
begin
   MountComboBox.Text := PageControl1.ActivePage.Caption;
end;

procedure TMountForm.TabSheet3Show(Sender: TObject);
begin
   reiservolume.SetFocus;

   LoadReiserFilesystems;
   SelectReiserFilesystem(0);
end;

procedure TMountForm.ReiserVolumeChange(Sender: TObject);
begin
   SelectReiserFilesystem(ReiserVolume.ItemIndex);
end;

procedure TMountForm.ReiserUUIDChange(Sender: TObject);
begin
   with Sender as TComboBox do
   begin
      SelectReiserFilesystem(Integer(Items.Objects[ItemIndex]));
   end;
end;

procedure TMountForm.TabSheet4Show(Sender: TObject);
begin
   fatvolume.SetFocus;

   LoadFatFilesystems;
   SelectFatFilesystem(0);
end;

procedure TMountForm.fatvolumeChange(Sender: TObject);
begin
   SelectFatFilesystem(Fatvolume.ItemIndex);
end;

procedure TMountForm.fatuuidChange(Sender: TObject);
begin
   with Sender as TComboBox do
   begin
      SelectFatFilesystem(Integer(Items.Objects[ItemIndex]));
   end;
end;

end.
