unit ls;

interface

const
   S_IFMT   = $F000; // format mask
   S_IFSOCK = $C000; // socket
   S_IFLNK  = $A000; // symbolic link
   S_IFREG  = $8000; // regular file
   S_IFBLK  = $6000; // block device
   S_IFDIR  = $4000; // directory
   S_IFCHR  = $2000; // character device
   S_IFIFO  = $1000; // fifo

   S_ISUID  = $0800; // SUID
   S_ISGID  = $0400; // SGID
   S_ISVTX  = $0200; // sticky bit

   S_IRWXU  = $01C0; // user mask
   S_IRUSR  = $0100;
   S_IWUSR  = $0080;
   S_IXUSR  = $0040;

   S_IRWXG  = $0038; // group mask
   S_IRGRP  = $0020;
   S_IWGRP  = $0010;
   S_IXGRP  = $0008;

   S_IRWXO  = $0007; // other mask
   S_IROTH  = $0004;
   S_IWOTH  = $0002;
   S_IXOTH  = $0001;


function GetModeString(Mode : Integer) : String;

implementation

function GetModeString(Mode : Integer) : String;
begin
   // we will create a string which looks like
   // srwxrwxrwx
   // depending of course on the mode of the file.....

   // what ever this bit is called
   if (mode and S_IFMT) = S_IFSOCK then
   begin
      Result := 's';
   end
   else if (mode and S_IFMT) = S_IFLNK then
   begin
      Result := 'l';
   end
   else if (mode and S_IFMT) = S_IFBLK then
   begin
      Result := 'b';
   end
   else if (mode and S_IFMT) = S_IFDIR then
   begin
      Result := 'd';
   end
   else if (mode and S_IFMT) = S_IFCHR then
   begin
      Result := 'c';
   end
   else if (mode and S_IFMT) = S_IFIFO then
   begin
      Result := 'p'; // I think (p for pipe?)
   end
   else
   begin
      Result := '-';
   end;

   // the user attributes
   if (mode and S_IRUSR) = S_IRUSR then
   begin
      Result := Result + 'r';
   end
   else
   begin
      Result := Result + '-';
   end;

   if (mode and S_IWUSR) = S_IWUSR then
   begin
      Result := Result + 'w';
   end
   else
   begin
      Result := Result + '-';
   end;

   if (mode and S_IXUSR) = S_IXUSR then
   begin
      if (mode and S_ISUID) = S_ISUID then
      begin
         Result := Result + 's';
      end
      else
      begin
         Result := Result + 'x';
      end;
   end
   else
   begin
      if (mode and S_ISUID) = S_ISUID then
      begin
         Result := Result + 'S';
      end
      else
      begin
         Result := Result + '-';
      end;
   end;

   // group attributes
   if (mode and S_IRGRP) = S_IRGRP then
   begin
      Result := Result + 'r';
   end
   else
   begin
      Result := Result + '-';
   end;

   if (mode and S_IWGRP) = S_IWGRP then
   begin
      Result := Result + 'w';
   end
   else
   begin
      Result := Result + '-';
   end;

   if (mode and S_IXGRP) = S_IXGRP then
   begin
      if (mode and S_ISGID) = S_ISGID then
      begin
         Result := Result + 's';
      end
      else
      begin
         Result := Result + 'x';
      end
   end
   else
   begin
      if (mode and S_ISGID) = S_ISGID then
      begin
         Result := Result + 'S';
      end
      else
      begin
         Result := Result + '-';
      end
   end;

   // other attributes
   if (mode and S_IROTH) = S_IROTH then
   begin
      Result := Result + 'r';
   end
   else
   begin
      Result := Result + '-';
   end;

   if (mode and S_IWOTH) = S_IWOTH then
   begin
      Result := Result + 'w';
   end
   else
   begin
      Result := Result + '-';
   end;

   if (mode and S_IXOTH) = S_IXOTH then
   begin
      if (mode and S_ISVTX) = S_ISVTX then
      begin
         Result := Result + 't';
      end
      else
      begin
         Result := Result + 'x';
      end;
   end
   else
   begin
      if (mode and S_ISVTX) = S_ISVTX then
      begin
         Result := Result + 'T';
      end
      else
      begin
         Result := Result + '-';
      end;
   end;
end;

end.
