unit vfs_mem;

interface

uses vfs, classes;

type
   TMemFs = class(TFs)
   public
      function Mount(Options : TMountOptions) : TVNode; override;
      function UMount : Boolean; override;
   end;

   TMemNode = class(TVNode)
   public
      Entries : TStringList;

      function HasChildren : Boolean; override;
      function HasData : Boolean; override;
      function GetLen : Int64; override;

      function Mkdir(Name : String) : Boolean; override;

      procedure ListChildren(Children : TStringList); override;
      function GetChild(Name : String) : TVNode; override;
   end;

implementation

// FS Specific
// MemFS is the in memory only root filesystem

function TMemFs.Mount(Options : TMountOptions) : TVNode;
var
   Node : TMemNode;
begin
   Node := TMemNode.Create;
   Node.fs := self;
   Node.Name := '/';
   Node.Entries := TStringList.Create;
   Node.Entries.Add('.');
   Node.Entries.Add('..');
   Result := Node.Open;
end;

function TMemFs.UMount : Boolean;
begin
   Result := False;
end;

function TMemNode.HasChildren : Boolean;
begin
   Result := False;
   if Entries.Count > 0 then
   begin
      Result := True;
   end;
end;

function TMemNode.HasData : Boolean;
begin
   Result := False;
end;

function TMemNode.GetLen : Int64;
begin
   Result := 0;
end;

function TMemNode.Mkdir(Name : String) : Boolean;
begin
   if Entries.Count = 0 then
   begin
      // we must not be deleted now...
      Open;
   end;
   Entries.Add(Name);
   Result := True;
end;


procedure TMemNode.ListChildren(Children : TStringList);
begin
{   if Assigned(Overlay) then
   begin
      Overlay.ListChildren(Children);
   end
   else}
   begin
      Children.SetText(Entries.GetText);
   end;
end;

function TMemNode.GetChild(Name : String) : TVNode;
var
   Node : TMemNode;
begin
{   if Assigned(Overlay) then
   begin
      Result := Overlay.GetChild(Name);
   end
   else}
   begin
      Result := nil;
      if Entries.IndexOf(Name) >= 0 then
      begin
         Node := TMemNode.Create;
         Node.fs := fs;
         Node.Name := Name;
         Node.Entries := TStringList.Create;
         Node.Entries.Add('.');
         Node.Entries.Add('..');
         Result := Node;
      end;
   end;
end;

initialization
   RegisterFileSystem('memfs', TMemFs);

end.
