unit vfs_sftp;

interface
uses types, vfs, classes, execute, winsock;

// Implementation based on http://www.ietf.org/internet-drafts/draft-ietf-secsh-filexfer-13.txt
// but my server is only version 3 so that is the version that this program talks.
// inspection of the putty code shows the differences (putty is also version 3)

const
       SSH_FXP_INIT             =   1;
       SSH_FXP_VERSION          =   2;
       SSH_FXP_OPEN             =   3;
       SSH_FXP_CLOSE            =   4;
       SSH_FXP_READ             =   5;
       SSH_FXP_WRITE            =   6;
       SSH_FXP_LSTAT            =   7;
       SSH_FXP_FSTAT            =   8;
       SSH_FXP_SETSTAT          =   9;
       SSH_FXP_FSETSTAT         =  10;
       SSH_FXP_OPENDIR          =  11;
       SSH_FXP_READDIR          =  12;
       SSH_FXP_REMOVE           =  13;
       SSH_FXP_MKDIR            =  14;
       SSH_FXP_RMDIR            =  15;
       SSH_FXP_REALPATH         =  16;
       SSH_FXP_STAT             =  17;
       SSH_FXP_RENAME           =  18;
       SSH_FXP_READLINK         =  19;
       SSH_FXP_LINK             =  21;
       SSH_FXP_BLOCK            =  22;
       SSH_FXP_UNBLOCK          =  23;

       SSH_FXP_STATUS           = 101;
       SSH_FXP_HANDLE           = 102;
       SSH_FXP_DATA             = 103;
       SSH_FXP_NAME             = 104;
       SSH_FXP_ATTRS            = 105;

       SSH_FXP_EXTENDED         = 200;
       SSH_FXP_EXTENDED_REPLY   = 201;


       SSH_FILEXFER_ATTR_SIZE              = $00000001;
       SSH_FILEXFER_ATTR_UIDGID            = $00000002; // depricated since v3?
       SSH_FILEXFER_ATTR_PERMISSIONS       = $00000004;
       SSH_FILEXFER_ATTR_ACCESSTIME        = $00000008;
       SSH_FILEXFER_ATTR_CREATETIME        = $00000010;
       SSH_FILEXFER_ATTR_MODIFYTIME        = $00000020;
       SSH_FILEXFER_ATTR_ACL               = $00000040;
       SSH_FILEXFER_ATTR_OWNERGROUP        = $00000080;
       SSH_FILEXFER_ATTR_SUBSECOND_TIMES   = $00000100;
       SSH_FILEXFER_ATTR_BITS              = $00000200;
       SSH_FILEXFER_ATTR_ALLOCATION_SIZE   = $00000400;
       SSH_FILEXFER_ATTR_TEXT_HINT         = $00000800;
       SSH_FILEXFER_ATTR_MIME_TYPE         = $00001000;
       SSH_FILEXFER_ATTR_LINK_COUNT        = $00002000;
       SSH_FILEXFER_ATTR_UNTRANSLATED_NAME = $00004000;
       SSH_FILEXFER_ATTR_CTIME             = $00008000;
       SSH_FILEXFER_ATTR_EXTENDED          = $80000000;

       {
       SSH_FILEXFER_ATTR_FLAGS_READONLY         0x00000001
       SSH_FILEXFER_ATTR_FLAGS_SYSTEM           0x00000002
       SSH_FILEXFER_ATTR_FLAGS_HIDDEN           0x00000004
       SSH_FILEXFER_ATTR_FLAGS_CASE_INSENSITIVE 0x00000008
       SSH_FILEXFER_ATTR_FLAGS_ARCHIVE          0x00000010
       SSH_FILEXFER_ATTR_FLAGS_ENCRYPTED        0x00000020
       SSH_FILEXFER_ATTR_FLAGS_COMPRESSED       0x00000040
       SSH_FILEXFER_ATTR_FLAGS_SPARSE           0x00000080
       SSH_FILEXFER_ATTR_FLAGS_APPEND_ONLY      0x00000100
       SSH_FILEXFER_ATTR_FLAGS_IMMUTABLE        0x00000200
       SSH_FILEXFER_ATTR_FLAGS_SYNC             0x00000400
       SSH_FILEXFER_ATTR_FLAGS_TRANSLATION_ERR  0x00000800
       }
       SSH_FILEXFER_TYPE_REGULAR         = 1;
       SSH_FILEXFER_TYPE_DIRECTORY       = 2;
       SSH_FILEXFER_TYPE_SYMLINK         = 3;
       SSH_FILEXFER_TYPE_SPECIAL         = 4;
       SSH_FILEXFER_TYPE_UNKNOWN         = 5;
       SSH_FILEXFER_TYPE_SOCKET          = 6;
       SSH_FILEXFER_TYPE_CHAR_DEVICE     = 7;
       SSH_FILEXFER_TYPE_BLOCK_DEVICE    = 8;
       SSH_FILEXFER_TYPE_FIFO            = 9;

       SSH_FX_OK                           = 0;
       SSH_FX_EOF                          = 1;
       SSH_FX_NO_SUCH_FILE                 = 2;
       SSH_FX_PERMISSION_DENIED            = 3;
       SSH_FX_FAILURE                      = 4;
       SSH_FX_BAD_MESSAGE                  = 5;
       SSH_FX_NO_CONNECTION                = 6;
       SSH_FX_CONNECTION_LOST              = 7;
       SSH_FX_OP_UNSUPPORTED               = 8;
       SSH_FX_INVALID_HANDLE               = 9;
       SSH_FX_NO_SUCH_PATH                 = 10;
       SSH_FX_FILE_ALREADY_EXISTS          = 11;
       SSH_FX_WRITE_PROTECT                = 12;
       SSH_FX_NO_MEDIA                     = 13;
       SSH_FX_NO_SPACE_ON_FILESYSTEM       = 14;
       SSH_FX_QUOTA_EXCEEDED               = 15;
       SSH_FX_UNKNOWN_PRINCIPAL            = 16;
       SSH_FX_LOCK_CONFLICT                = 17;
       SSH_FX_DIR_NOT_EMPTY                = 18;
       SSH_FX_NOT_A_DIRECTORY              = 19;
       SSH_FX_INVALID_FILENAME             = 20;
       SSH_FX_LINK_LOOP                    = 21;
       SSH_FX_CANNOT_DELETE                = 22;
       SSH_FX_INVALID_PARAMETER            = 23;
       SSH_FX_FILE_IS_A_DIRECTORY          = 24;
       SSH_FX_BYTE_RANGE_LOCK_CONFLICT     = 25;
       SSH_FX_BYTE_RANGE_LOCK_REFUSED      = 26;
       SSH_FX_DELETE_PENDING               = 27;
       SSH_FX_FILE_CORRUPT                 = 28;
       SSH_FX_OWNER_INVALID                = 29;
       SSH_FX_GROUP_INVALID                = 30;
       SSH_FX_NO_MATCHING_BYTE_RANGE_LOCK  = 31;


           SSH_FXF_ACCESS_DISPOSITION      = $00000007;
               SSH_FXF_CREATE_NEW          = $00000000;
               SSH_FXF_CREATE_TRUNCATE     = $00000001;
               SSH_FXF_OPEN_EXISTING       = $00000002;
               SSH_FXF_OPEN_OR_CREATE      = $00000003;
               SSH_FXF_TRUNCATE_EXISTING   = $00000004;
           SSH_FXF_APPEND_DATA             = $00000008;
           SSH_FXF_APPEND_DATA_ATOMIC      = $00000010;
           SSH_FXF_TEXT_MODE               = $00000020;
           SSH_FXF_BLOCK_READ              = $00000040;
           SSH_FXF_BLOCK_WRITE             = $00000080;
           SSH_FXF_BLOCK_DELETE            = $00000100;
           SSH_FXF_BLOCK_ADVISORY          = $00000200;
           SSH_FXF_NOFOLLOW                = $00000400;
           SSH_FXF_DELETE_ON_CLOSE         = $00000800;
           SSH_FXF_ACCESS_AUDIT_ALARM_INFO = $00001000;
           SSH_FXF_ACCESS_BACKUP           = $00002000;
           SSH_FXF_BACKUP_STREAM           = $00004000;
           SSH_FXF_OVERRIDE_OWNER          = $00008000;

       ACE4_READ_DATA         = $00000001;
       ACE4_LIST_DIRECTORY    = $00000001;
       ACE4_WRITE_DATA        = $00000002;
       ACE4_ADD_FILE          = $00000002;
       ACE4_APPEND_DATA       = $00000004;
       ACE4_ADD_SUBDIRECTORY  = $00000004;
       ACE4_READ_NAMED_ATTRS  = $00000008;
       ACE4_WRITE_NAMED_ATTRS = $00000010;
       ACE4_EXECUTE           = $00000020;
       ACE4_DELETE_CHILD      = $00000040;
       ACE4_READ_ATTRIBUTES   = $00000080;
       ACE4_WRITE_ATTRIBUTES  = $00000100;
       ACE4_DELETE            = $00010000;
       ACE4_READ_ACL          = $00020000;
       ACE4_WRITE_ACL         = $00040000;
       ACE4_WRITE_OWNER       = $00080000;
       ACE4_SYNCHRONIZE       = $00100000;


    // open for version 3
    SSH_FXF_READ                            = $00000001;
    SSH_FXF_WRITE                           = $00000002;
    SSH_FXF_APPEND                          = $00000004;
    SSH_FXF_CREAT                           = $00000008;
    SSH_FXF_TRUNC                           = $00000010;
    SSH_FXF_EXCL                            = $00000020;

type

    uint32 = U32; // Network Byte Order
    uint64 = U64; // Network Byte Order

   TSftpNode = class;

   TSftpPacket = class
   public
      Offset : Integer;
      Data : String;

      procedure SetType(b : Byte);
      procedure AddByte(b : Byte);
      procedure Adduint32(n : uint32);
      procedure Adduint64(n : uint64);
      procedure AddData(Buffer : Pointer; Len : Integer);
      procedure AddString(S : String);

      function GetByte : Byte;
      function Getuint32 : uint32;
      function Getuint64 : uint64;
      function GetData(Buffer : Pointer; Len : Integer) : Boolean;
      function GetString : String;
      function GetAttrs(Node : TSftpNode) : String;
      function GetBool : Boolean;

      function Remaining : Integer;

      function Send(c : TChildProcess) : Boolean;
      function Recv(c : TChildProcess; Expected : Integer) : Boolean;
   end;

   TSftpFs = class(TFs)
   private
      WriteAccess : Boolean;
      Child : TChildProcess;
      RequestID : Integer;
   protected
      function InitProtocol : Boolean;

   public
      function Mount(Options : TMountOptions) : TVNode; override;
      function UMount : Boolean; override;
   end;

   TSftpFsClass = class of TSftpFs;

   TSftpRead = class(TObject)
   public
      Offset  : Int64;
      Len     : Int64;
      Ready   : Boolean;
      Buffer  : String;
      id      : Integer;
   end;

   TSftpNode = class(TVNode)
   private
      Path        : String;
      AttrMode    : Integer;
      AttrSize    : Int64;
      AttrUser    : Int64;
      AttrGroup   : Int64;
      Handle      : String;

      // Read ahead...
      src               : Integer; // sequential read count
      last_read_tail    : Int64;
      ReadAhead         : TList;
      ReadAheadTail     : Int64;

      DelayWrite        : TList;

      function MakeRequestID : Integer;

      procedure SendReadahead(Offset : Int64; Len : Int64);
      procedure GetReadahead;
   public
      constructor Create; override;
      destructor Destroy; override;

      function HasChildren : Boolean; override;
      function HasData : Boolean; override;
      function GetLen : U64; override;
      function GetAttribute(Attribute : String) : String; override;


      procedure ListChildren(Children : TStringList); override;
      function Mkdir(Name : String) : Boolean; override;
      function CreateFile(Name : String) : Boolean; override;
      function Delete(Name : String) : Boolean; override;


      function OpenData(WriteAccess : Boolean) : Boolean; override;
      function CloseData : Boolean; override;
      function ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
      function WriteData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
      function SetLen(Len : Int64) : Boolean; override;


   protected

      function GetChild(Name : String) : TVNode; override;

   end;

implementation

uses windows, sysutils, debug, unicode, vv_io, time, vv, ls;

const Facility = 'vfs_sftp';

procedure TSftpPacket.SetType(b : Byte);
begin
   Offset := 0;
   Data := '';
   AddByte(b);
end;

procedure TSftpPacket.AddByte(b : Byte);
var
   c : Char;
begin
   c := Chr(b);
   Data := Data + c;
end;

procedure TSftpPacket.Adduint32(n : uint32);
var
   S : String;
   p : PU32;
begin
   SetLength(S, sizeof(U32));
   p := PU32(PChar(S));
   p^ := HTONL(n);
   Data := Data + S;
end;

procedure TSftpPacket.Adduint64(n : uint64);
var
   Temp : LARGE_INTEGER;
begin
   Temp.QuadPart := n;
   Adduint32(Temp.HighPart);
   Adduint32(Temp.LowPart);
end;

procedure TSftpPacket.AddString(S : String);
begin
   Adduint32(Length(S));
   Data := Data + S;
end;

procedure TSftpPacket.AddData(Buffer : Pointer; Len : Integer);
var
   OldLen : Integer;
   p : PChar;
begin
   Adduint32(Len);
   OldLen := Length(Data);
   SetLength(Data, OldLen + Len);
   CopyMemory(PChar(Data) + OldLen, Buffer, Len);
end;

function TSftpPacket.GetByte : Byte;
begin
   Result := Ord(PChar(Data)[Offset]);
   Offset := Offset + 1;
end;

function TSftpPacket.Getuint32 : uint32;
var
   p : PU32;
begin
   p := Pointer(PChar(Data) + Offset);
   Result := NTOHL(p^);
   Offset := Offset + sizeof(PU32);
end;

function TSftpPacket.Getuint64 : uint64;
var
   Temp : LARGE_INTEGER;
begin
   Temp.HighPart := Getuint32;
   Temp.LowPart := Getuint32;
   Result := Temp.QuadPart;
end;

function TSftpPacket.GetData(Buffer : Pointer; Len : Integer) : Boolean;
begin
   CopyMemory(Buffer, PChar(Data) + Offset, Len);
   Offset := Offset + Len;
   Result := True;
end;

function TSftpPacket.GetString : String;
var
   Len : Integer;
begin
   Len := Getuint32;
   if Offset + Len > Length(Data) then
   begin
      Log('Bogus string length', Facility, LOG_ERROR);
      Result := '';
   end
   else
   begin
      Result := Copy(Data, Offset + 1, Len);
      Offset := Offset + Len;
   end;
end;

function TSftpPacket.GetAttrs(Node : TSftpNode) : String;
var
   Flags : Integer;
   Count : Integer;
   i : Integer;
begin
   Flags := Getuint32;
//   GetByte; // type - only version 6?
   if Flags and SSH_FILEXFER_ATTR_SIZE = SSH_FILEXFER_ATTR_SIZE then
   begin
      Node.AttrSize := Getuint64;
   end;
   if Flags and SSH_FILEXFER_ATTR_ALLOCATION_SIZE = SSH_FILEXFER_ATTR_ALLOCATION_SIZE then
   begin
      Getuint64;
   end;
   if Flags and SSH_FILEXFER_ATTR_UIDGID = SSH_FILEXFER_ATTR_UIDGID then
   begin
      Node.AttrUser := Getuint32;
      Node.AttrGroup := Getuint32;
   end;
   if Flags and SSH_FILEXFER_ATTR_OWNERGROUP = SSH_FILEXFER_ATTR_OWNERGROUP then
   begin
      GetString;
      GetString;
   end;
   if Flags and SSH_FILEXFER_ATTR_PERMISSIONS = SSH_FILEXFER_ATTR_PERMISSIONS then
   begin
      Node.AttrMode := Getuint32;
   end;

   if Flags and SSH_FILEXFER_ATTR_ACCESSTIME = SSH_FILEXFER_ATTR_ACCESSTIME then
   begin
      Getuint64;
      if Flags and SSH_FILEXFER_ATTR_SUBSECOND_TIMES = SSH_FILEXFER_ATTR_SUBSECOND_TIMES then
      begin
         Getuint32;
      end;
   end;
   if Flags and SSH_FILEXFER_ATTR_CREATETIME = SSH_FILEXFER_ATTR_CREATETIME then
   begin
      Getuint64;
      if Flags and SSH_FILEXFER_ATTR_SUBSECOND_TIMES = SSH_FILEXFER_ATTR_SUBSECOND_TIMES then
      begin
         Getuint32;
      end;
   end;
   if Flags and SSH_FILEXFER_ATTR_MODIFYTIME = SSH_FILEXFER_ATTR_MODIFYTIME then
   begin
      Getuint64;
      if Flags and SSH_FILEXFER_ATTR_SUBSECOND_TIMES = SSH_FILEXFER_ATTR_SUBSECOND_TIMES then
      begin
         Getuint32;
      end;
   end;
   if Flags and SSH_FILEXFER_ATTR_CTIME = SSH_FILEXFER_ATTR_CTIME then
   begin
      Getuint64;
      if Flags and SSH_FILEXFER_ATTR_SUBSECOND_TIMES = SSH_FILEXFER_ATTR_SUBSECOND_TIMES then
      begin
         Getuint32;
      end;
   end;

   if Flags and SSH_FILEXFER_ATTR_ACL = SSH_FILEXFER_ATTR_ACL then
   begin
      GetString;
   end;

   if Flags and SSH_FILEXFER_ATTR_BITS = SSH_FILEXFER_ATTR_BITS then
   begin
      Getuint32;
      Getuint32;
   end;

   if Flags and SSH_FILEXFER_ATTR_TEXT_HINT = SSH_FILEXFER_ATTR_TEXT_HINT then
   begin
      GetByte;
   end;

   if Flags and SSH_FILEXFER_ATTR_MIME_TYPE = SSH_FILEXFER_ATTR_MIME_TYPE then
   begin
      GetString;
   end;

   if Flags and SSH_FILEXFER_ATTR_LINK_COUNT = SSH_FILEXFER_ATTR_LINK_COUNT then
   begin
      Getuint32;
   end;

   if Flags and SSH_FILEXFER_ATTR_UNTRANSLATED_NAME = SSH_FILEXFER_ATTR_UNTRANSLATED_NAME then
   begin
      GetString;
   end;

   if Flags and SSH_FILEXFER_ATTR_EXTENDED = SSH_FILEXFER_ATTR_EXTENDED then
   begin
      Count := Getuint32;
      for i := 0 to Count - 1 do
      begin
         GetString;
         GetString;
      end;
   end;


   Result := '';
end;

function TSftpPacket.GetBool : Boolean;
begin
   if Getuint32 > 0 then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TSftpPacket.Remaining : Integer;
begin
   Result := Length(Data) - Offset;
end;

function TSftpPacket.Send(c : TChildProcess) : Boolean;
var
   Len : uint32;
begin
   Len := HTONL(Length(Data));
   c.Write(@Len, sizeof(len), nil);
   c.Write(PChar(Data), Length(Data), nil);
   Result := True;
end;

function TSftpPacket.Recv(c : TChildProcess; Expected : Integer) : Boolean;
var
   Len  : uint32;
   Msg  : String;
   Done : Integer;
   ID   : pu32;
begin
   while True do
   begin
      Result := False;
      Offset := 0;
      if c.ReadFull(@Len, sizeof(len)) then
      begin
         Len := NTOHL(Len);
         if Len > 65536 then
         begin
            Log('Packet seems too big', Facility, LOG_ERROR);
            SetLength(Msg, 1024);
            Len := HTONL(Len);
            CopyMemory(PChar(Msg), @Len, 4);
            c.read(PChar(Msg) + 4, Length(Msg) - 4, @Done);
            SetLength(Msg, Done);
            Log(Msg, Facility, LOG_ERROR);
         end
         else
         begin
            SetLength(Data, Len);
            if c.ReadFull(PChar(Data), Length(Data)) then
            begin
               Result := True;
            end;
         end;
      end;
      if Result and (Expected <> 0) then
      begin
         id := pu32(PChar(data) + 1);
         if NTOHL(id^) = Expected then
         begin
            break;
         end
         else
         begin
            Log('Unexpected ID ' + IntToStr(Expected) + ' != ' + IntToStr(NTOHL(id^)) + ', skipping');
            Result := False;
         end;
      end
      else
      begin
         break;
      end;
   end;
end;

function TSftpFs.Mount(Options : TMountOptions) : TVNode;
var
   Command  : String;
   Node     : TSftpNode;
   PLink    : String;
   Session  : String;
begin
   Result := nil;
   // start plink...
   Session := Options.GetParameter(0);
   PLink := 'C:\Program Files\PuTTY\plink';
   if Options.HasOption('plink') then
   begin
      PLink := Options.GetOption('plink');
   end;
   if Options.HasOption('username') then
   begin
      PLink := PLink + ' -l ' + Options.GetOption('username');
   end;
   if Options.HasOption('password') then
   begin
      PLink := PLink + ' -pw ' + Options.GetOption('password');
   end;
   Command := PLink + ' -batch -s ' + Session + ' sftp';
   //Log('Command = ' + Command);
   Child := TChildProcess.Create;
   if Child.StartChild(Command) then
   begin
      if InitProtocol then
      begin
         Node := TSftpNode.Create;
         Node.Name := '/';
         Node.Path := '/';
         Node.fs := self;

         // we must be a directory...
         Node.AttrMode := S_IFDIR;
         Result := Node;
      end
      else
      begin
         // Get exit code...
         Log('Exit code ' + IntToStr(Child.GetExitCode));
      end;
   end;
end;

function TSftpFs.UMount : Boolean;
begin
   // disconnect somehow...
   Result := True;
end;

function TSftpFS.InitProtocol : Boolean;
var
   Packet   : TSftpPacket;
   Version  : Integer;
begin
   Packet := TSftpPacket.Create;
   Packet.AddByte(SSH_FXP_INIT);
   Packet.Adduint32(3);
   Packet.Send(Child);
   Packet.Recv(Child, 0);

   if Packet.GetByte = SSH_FXP_VERSION then
   begin
      Version := Packet.Getuint32;
      Log('Using version ' + IntToStr(Version), Facility, LOG_VERBOSE);
      Result := True;
      RequestID := 1;
   end
   else
   begin
      Result := False;
   end;
   Packet.Free;
end;

constructor TSftpNode.Create;
begin
   inherited
end;

destructor TSftpNode.Destroy;
begin
   CloseData;

   inherited;
end;

function TSftpNode.MakeRequestID : Integer;
begin
   Result := TSftpFs(fs).RequestID;
   Inc(TSftpFs(fs).RequestID);
   if Result = 0 then
   begin
      Result := MakeRequestID;
   end;
end;

function TSftpNode.GetChild(Name : String) : TVNode;
var
   Search   : String;
   Packet   : TSftpPacket;
   ID       : Integer;
   Code     : Byte;

begin
   Result := nil;

   if Path[Length(Path)] = '/' then
   begin
      Search := Path + Name;
   end
   else
   begin
      Search := Path + '/' + Name;
   end;

   //Log('sending stat for ' + Search);

   Packet := TSftpPacket.Create;
   Packet.SetType(SSH_FXP_STAT);
   ID := MakeRequestID;
   Packet.Adduint32(ID);
   Packet.AddString(Search);
   Packet.Send(TSftpFs(fs).Child);
   Packet.Recv(TSftpFs(fs).Child, ID);
   Code := Packet.GetByte;
   Packet.Getuint32;// requestid
   if Code = SSH_FXP_ATTRS then
   begin
      Result := TSftpNode.Create;
      Result.fs := fs;
      TSftpNode(Result).Path := Search;
      Result.Name := Name;
      Packet.GetAttrs(TSftpNode(Result));
   end
   else if Code = SSH_FXP_STATUS then
   begin
      Log('SSH_FXP_STAT Status ' + IntToStr(Packet.Getuint32) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
   end
   else
   begin
      Log('SSH_FXP_STAT Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
   end;

   Packet.Free;
end;

function TSftpNode.HasChildren : Boolean;
begin
   if (AttrMode and S_IFMT) = S_IFDIR then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TSftpNode.HasData : Boolean;
begin
   if (AttrMode and S_IFMT) = S_IFREG then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TSftpNode.GetLen : U64;
begin
   Result := AttrSize;
end;

function TSftpNode.GetAttribute(Attribute : String) : String;
begin
   if Attribute = 'user' then
   begin
      Result := IntToStr(AttrUser);
   end
   else if Attribute = 'group' then
   begin
      Result := IntToStr(AttrGroup);
   end
   else if Attribute = 'ctime' then
   begin
      Result := '';
   end
   else if Attribute = 'mtime' then
   begin
      Result := '';
   end
   else if Attribute = 'atime' then
   begin
      Result := '';
   end
   else if Attribute = 'mode' then
   begin
      Result := GetModeString(AttrMode);
   end
end;


procedure TSftpNode.ListChildren(Children : TStringList);
var
   Packet   : TSftpPacket;
   Code     : Byte;
   ID       : Integer;
   Handle   : String;
   Count    : Integer;
   i        : Integer;
   S : String;
   Child : TSftpNode;
   Search : String;
begin
   if Path[Length(Path)] = '/' then
   begin
      Search := Path;
   end
   else
   begin
      Search := Path + '/';
   end;

   Packet := TSftpPacket.Create;
   Packet.SetType(SSH_FXP_OPENDIR);
   ID := MakeRequestID;
   Packet.Adduint32(ID);
   Packet.AddString(Path);
   Packet.Send(TSftpFs(fs).Child);
   Packet.Recv(TSftpFs(fs).Child, ID);
   Code := Packet.GetByte;
   Packet.Getuint32;// requestid
   if Code = SSH_FXP_HANDLE then
   begin
      Handle := Packet.GetString;
      Log('Opened directory', Facility, LOG_VERBOSE);
      while True do
      begin
         Packet.SetType(SSH_FXP_READDIR);
         ID := MakeRequestID;
         Packet.Adduint32(ID);
         Packet.AddString(Handle);
         Packet.Send(TSftpFs(fs).Child);
         Packet.Recv(TSftpFs(fs).Child, ID);
         Code := Packet.GetByte;
         Packet.Getuint32;// requestid
         if Code = SSH_FXP_NAME then
         begin
            Count := Packet.Getuint32;
            for i := 0 to Count - 1 do
            begin
               //Log('There are ' + IntToStr(Packet.Remaining) + ' bytes left');
               Child := TSftpNode.Create;
               Child.fs := fs;
               S := Packet.GetString;
               Child.Name := S;
               Child.Path := Search + S;
               Children.Add(S);
               S := Packet.GetString; // depricated after v3?
               Packet.GetAttrs(Child);

               Child.Open;
               AddChildNode(Child);
               Child.Close;
               // we need a RegisterChild method to insert a child into the dir cache
               // for now, we do a second lookup per node (GetChild)
               //Child.Close;
            end;
            if Packet.Remaining > 0 then
            begin
               if Packet.GetBool then
               begin
                  break;
               end;
            end;
         end
         else if Code = SSH_FXP_STATUS then
         begin
            Log('SSH_FXP_READDIR Status ' + IntToStr(Packet.Getuint32) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
            break;
         end
         else
         begin
            Log('Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
            break;
         end;
      end;
      // Close the handle
      Packet.SetType(SSH_FXP_CLOSE);
      ID := MakeRequestID;
      Packet.Adduint32(ID);
      Packet.AddString(Handle);
      Packet.Send(TSftpFs(fs).Child);
      Packet.Recv(TSftpFs(fs).Child, ID);
      Code := Packet.GetByte;
      if Code = SSH_FXP_STATUS then
      begin
         Log('SSH_FXP_CLOSE Status ' + IntToStr(Packet.Getuint32) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
      end;
   end
   else if Code = SSH_FXP_STATUS then
   begin
      Log('SSH_FXP_OPENDIR Status ' + IntToStr(Packet.Getuint32) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
   end
   else
   begin
      Log('SSH_FXP_OPENDIR Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
   end;
   Packet.Free;
end;

function TSftpNode.Mkdir(Name : String) : Boolean;
var
   Packet   : TSftpPacket;
   Code     : Byte;
   ID       : Integer;
   Status   : Integer;
   Search : String;
begin
   Result := False;

   if Path[Length(Path)] = '/' then
   begin
      Search := Path;
   end
   else
   begin
      Search := Path + '/';
   end;

   Search := Search + Name;


   Packet := TSftpPacket.Create;
   Packet.SetType(SSH_FXP_MKDIR);
   ID := MakeRequestID;
   Packet.Adduint32(ID);
   Packet.AddString(Search);
   Packet.Adduint32(0); // no attributes specified
   Packet.Send(TSftpFs(fs).Child);
   Packet.Recv(TSftpFs(fs).Child, ID);
   Code := Packet.GetByte;
   Packet.Getuint32;// requestid
   if Code = SSH_FXP_STATUS then
   begin
      Status := Packet.Getuint32;
      if Status = SSH_FX_OK then
      begin
         Result := True;
      end
      else
      begin
         Log('SSH_FXP_MKDIR ' + IntToStr(Status) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
      end;
   end
   else
   begin
      Log('SSH_FXP_MKDIR Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
   end;
   Packet.Free;
end;

function TSftpNode.CreateFile(Name : String) : Boolean;
var
   Packet   : TSftpPacket;
   ID       : Integer;
   Access   : Integer;
   //Flags    : Integer;
   Code     : Byte;
   Search   : String;
   Handle   : String;
begin
   Result := False;
   // do an open with SSH_FXF_CREAT
   Access := SSH_FXF_WRITE or SSH_FXF_CREAT;

   if Path[Length(Path)] = '/' then
   begin
      Search := Path;
   end
   else
   begin
      Search := Path + '/';
   end;

   Search := Search + Name;


   Packet := TSftpPacket.Create;
   Packet.SetType(SSH_FXP_OPEN);
   ID := MakeRequestID;
   Packet.Adduint32(ID);
   Packet.AddString(Search);
   Packet.Adduint32(Access);
   Packet.Adduint32(0); // attr mask ?
   Packet.Send(TSftpFs(fs).Child);
   Packet.Recv(TSftpFs(fs).Child, ID);
   Code := Packet.GetByte;
   Packet.Getuint32;// requestid
   if Code = SSH_FXP_HANDLE then
   begin
      Handle := Packet.GetString;
      Log('Created file ' + Path, Facility, LOG_VERBOSE);
      Result := True;

      // close the handle
      Packet.SetType(SSH_FXP_CLOSE);
      ID := MakeRequestID;
      Packet.Adduint32(ID);
      Packet.AddString(Handle);
      Packet.Send(TSftpFs(fs).Child);
      Packet.Recv(TSftpFs(fs).Child, ID);
      Code := Packet.GetByte;
      if Code = SSH_FXP_STATUS then
      begin
         Log('SSH_FXP_CLOSE Status ' + IntToStr(Packet.Getuint32) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
      end
      else
      begin
         Log('SSH_FXP_CLOSE Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
      end;
   end
   else if Code = SSH_FXP_STATUS then
   begin
      Log('SSH_FXP_OPEN Status ' + IntToStr(Packet.Getuint32) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
   end
   else
   begin
      Log('SSH_FXP_OPEN Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
   end;
   Packet.Free;
end;

function TSftpNode.Delete(Name : String) : Boolean;
begin
   Result := False;
end;

function TSftpNode.OpenData(WriteAccess : Boolean) : Boolean;
var
   Packet   : TSftpPacket;
   ID       : Integer;
   Access   : Integer;
   //Flags    : Integer;
   Code     : Byte;
begin
   Result := False;
   if WriteAccess then
   begin
      //Access := ACE4_READ_DATA or ACE4_READ_ATTRIBUTES or ACE4_WRITE_DATA or ACE4_WRITE_ATTRIBUTES;
      Access := SSH_FXF_READ or SSH_FXF_WRITE;
   end
   else
   begin
      //Access := ACE4_READ_DATA or ACE4_READ_ATTRIBUTES;
      Access := SSH_FXF_READ;
   end;

   //Flags := SSH_FXF_OPEN_EXISTING;



   Packet := TSftpPacket.Create;
   Packet.SetType(SSH_FXP_OPEN);
   ID := MakeRequestID;
   Packet.Adduint32(ID);
   Packet.AddString(Path);
   Packet.Adduint32(Access);
//   Packet.Adduint32(Flags); v6 only?
   Packet.Adduint32(0); // attr mask ?
   Packet.Send(TSftpFs(fs).Child);
   Packet.Recv(TSftpFs(fs).Child, ID);
   Code := Packet.GetByte;
   Packet.Getuint32;// requestid
   if Code = SSH_FXP_HANDLE then
   begin
      Handle := Packet.GetString;
      Log('Opened file ' + Path, Facility, LOG_VERBOSE);
      Result := True;

      last_read_tail := 0;
   end
   else if Code = SSH_FXP_STATUS then
   begin
      Log('SSH_FXP_OPEN Status ' + IntToStr(Packet.Getuint32) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
   end
   else
   begin
      Log('SSH_FXP_OPEN Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
   end;
   Packet.Free;

   ReadAhead := TList.Create;

end;

function TSftpNode.CloseData : Boolean;
var
   Packet   : TSftpPacket;
   ID       : Integer;
   Code     : Byte;
   R        : TSftpRead;
begin
   if Length(Handle) > 0 then
   begin
      // Close the handle
      Packet := TSftpPacket.Create;
      Packet.SetType(SSH_FXP_CLOSE);
      ID := MakeRequestID;
      Packet.Adduint32(ID);
      Packet.AddString(Handle);
      Packet.Send(TSftpFs(fs).Child);
      Packet.Recv(TSftpFs(fs).Child, ID);
      Code := Packet.GetByte;
      if Code = SSH_FXP_STATUS then
      begin
         Log('SSH_FXP_CLOSE Status ' + IntToStr(Packet.Getuint32) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
      end
      else
      begin
         Log('SSH_FXP_CLOSE Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
      end;
      Packet.Free;
      Handle := '';

      while ReadAhead.Count > 0 do
      begin
         R := TSftpRead(ReadAhead[0]);
         R.Free;
         ReadAhead.Delete(0);
      end;
      ReadAhead.Free;

   end;
   Result := True;
end;


// There are still lots of bugs in this code.
// mixing other requests in between reads can result in
// discarded replies which never get removed.
procedure TSftpNode.SendReadahead(Offset : Int64; Len : Int64);
var
   Packet   : TSftpPacket;
   R : TSftpRead;
begin
//   Len := 32768; // should we have a sliding window?


   R := TSftpRead.Create;
   R.Offset := Offset;
   R.Len    := Len;
   R.id     := MakeRequestID;
   R.Ready  := False;

   ReadAheadTail := Offset + Len;

   Packet := TSftpPacket.Create;

   Packet.SetType(SSH_FXP_READ);
   Packet.Adduint32(R.id);
   Packet.AddString(Handle);
   Packet.Adduint64(R.Offset);
   Packet.Adduint32(R.Len);
   Packet.Send(TSftpFs(fs).Child);

   // we will read the reply next time we use
   // Packet.Recv or ReadData
   Packet.Free;

   ReadAhead.Add(R);
end;

// can the client re-order the replies here ?
procedure TSftpNode.GetReadahead;
var
   Packet   : TSftpPacket;
   Actual   : Integer;
   Code     : Byte;
   Id       : Integer;
   R        : TSftpRead;
   i        : Integer;
begin
   // are there any outstanding requests???
   if ReadAhead.Count > 0 then
   begin
      if not TSftpRead(ReadAhead[0]).Ready then
      begin
         Packet := TSftpPacket.Create;
         Packet.Recv(TSftpFs(fs).Child, 0);
         Code := Packet.GetByte;
         Id := Packet.Getuint32;// requestid
         if Code = SSH_FXP_DATA then
         begin
            R := nil;
            for i := 0 to ReadAhead.Count - 1 do
            begin
               R := TSftpRead(ReadAhead[i]);
               if R.Id = Id then
               begin
                  Actual := Packet.Getuint32;
                  R.Len := Actual;
                  //Log('Got read ahead data ' + IntToStr(Actual) + ' bytes', Facility, LOG_VERBOSE);
                  // copy the data
                  SetLength(R.Buffer, Actual);
                  Packet.GetData(PChar(R.Buffer), Actual);
                  R.Ready := True;
                  break;
               end;
            end;
            if not Assigned(R) then
            begin
               Log('Could not find the request for data id = ' + IntToStr(Id));
            end;
         end
         else if Code = SSH_FXP_STATUS then
         begin
            Log('SSH_FXP_READ Status ' + IntToStr(Packet.Getuint32) + ' ' + Packet.GetString, Facility, LOG_ERROR);
         end
         else
         begin
            Log('SSH_FXP_READ Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
         end;
         Packet.Free;
      end;
   end;
end;

function TSftpNode.ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean;
var
   Packet   : TSftpPacket;
   Actual   : Integer;
   Code     : Byte;
   Start    : Integer;
   R        : TSftpRead;
   i        : Integer;
begin
   Result := False;

   if Handle = '' then
   begin
      OpenData(False);
   end;

   // is there read ahead data available???
//   if readahead_len > 0 then
 //  begin
   GetReadahead;

   i := 0;
   while i < ReadAhead.Count do
   begin
      R := TSftpRead(ReadAhead[i]);
      if (Offset >= R.Offset) and ((Offset + Len) <= (R.Offset + R.Len)) then
      begin
         Start := Offset - R.Offset;
         CopyMemory(Buffer, PChar(R.Buffer) + Start, len);
         last_read_tail := Offset + Len;
         if (Offset + Len) = (R.Offset + R.Len) then
         begin
            // delete this request
            R.Free;
            ReadAhead.Delete(i);

            // read more data?
            //SendReadahead(ReadAheadTail, Len);
         end;
         Result := True;
         //Log('read ahead');

         exit;
      end
      else if Offset + Len > R.Offset then
      begin
         // We have seeked over this data.. discard it
         log('discarding read ahead data');
         src := 0;
         R.Free;
         ReadAhead.Delete(i);
      end
      else
      begin
         i := i + 1;
      end;
   end;

   // no read ahead data...
   if ReadAhead.Count > 0 then
   begin
      // I'm backwards
      // delete all the data?
      Log('can''t use this read ahead data');
      src := 0;
      while ReadAhead.Count > 0 do
      begin
         R := TSftpRead(ReadAhead[0]);
         R.Free;
         ReadAhead.Delete(0);
      end;
   end;
   // do the read read, but put it in the queue
   // this is an extra memory copy though...
   SendReadahead(Offset, Len);
   GetReadahead;
   R := TSftpRead(ReadAhead[0]);

   if R.Ready then
   begin
      if R.Len = len then
      begin
         CopyMemory(Buffer, PChar(R.Buffer), len);
         Result := True;
      end
      else
      begin
         Log('Short read');
      end;
      R.Free;
      ReadAhead.Delete(0);
   end
   else
   begin
      Log('some kind of error');
   end;

   if Result then
   begin
      if Offset = last_read_tail then
      begin
         src := src + 1;
      end
      else
      begin
         src := 0;
      end;
      last_read_tail := Offset + Len;

      if src >= 2 then
      begin
         // we should start read ahead
         if ReadAhead.Count < 8 then  // low water mark
         begin
            ReadAheadTail := last_read_tail;
            while ReadAhead.Count < 128 do // high water mark
            begin
               SendReadahead(ReadAheadTail, Len);//8 * 1024 ?32768?);
               if Len < 32768 then
               begin
                  Len := 32768;
               end;
            end;
         end;
      end;
   end;

end;

function TSftpNode.WriteData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean;
var
   Packet   : TSftpPacket;
   Code     : Byte;
   ID       : Integer;
   Status   : Integer;
begin
   Result := False;

   if Handle = '' then
   begin
      OpenData(True);
   end;


   Packet := TSftpPacket.Create;
   Packet.SetType(SSH_FXP_WRITE);
   ID := MakeRequestID;
   Packet.Adduint32(ID);
   Packet.AddString(Handle);
   Packet.Adduint64(Offset);
   Packet.AddData(Buffer, Len);
   Packet.Send(TSftpFs(fs).Child);

   if Assigned(DelayWrite) then
   begin
      DelayWrite.Add(Pointer(ID));
      if DelayWrite.Count < 16 then
      begin
         ID := 0;
         Result := True; // we have to assume
      end
      else
      begin
         ID := Integer(DelayWrite[0]);
         DelayWrite.Delete(0);
      end;
   end;

   if ID > 0 then
   begin
      Packet.Recv(TSftpFs(fs).Child, ID);
      Code := Packet.GetByte;
      Packet.Getuint32;// requestid
      if Code = SSH_FXP_STATUS then
      begin
         Status := Packet.Getuint32;
         if Status = SSH_FX_OK then
         begin
            Result := True;

            // we can enable delayed writes
            if not Assigned(DelayWrite) then
            begin
               DelayWrite := TList.Create;
            end;

         end
         else
         begin
            Log('SSH_FXP_WRITE ' + IntToStr(Status) + ' ' + Packet.GetString, Facility, LOG_VERBOSE);
         end;
      end
      else
      begin
         Log('SSH_FXP_WRITE Unexpected code ' + IntToStr(Code), Facility, LOG_ERROR);
      end;
   end;
   Packet.Free;
end;

function TSftpNode.SetLen(Len : Int64) : Boolean;
begin
   Result := False;
end;

initialization
   RegisterFileSystem('sftp', TSftpFs);


end.
