unit vv_loop;

interface

uses vv, vfs;

  // Allow access to a file as a block device
  // like the linux loop driver
  type TLoopVolume = class(TVirtualVolume)
  private
     Node : TVNode;
  public
    function Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;

    function Open(WriteAccess : Boolean) : Boolean; override;
    function Close : Boolean; override;

    procedure SetNode(Node : TVNode);

    function GetLength : Int64; override;
    function IsWritable : Boolean; override;

  end;


implementation

function TLoopVolume.Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
begin
   Result := Node.ReadData(Offset, Length, Buffer);
end;

function TLoopVolume.Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
begin
   if IsWritable then
   begin
      Result := Node.WriteData(Offset, Length, Buffer);
   end
   else
   begin
      Result := False;
   end;
end;

function TLoopVolume.Open(WriteAccess : Boolean) : Boolean;
begin
   Result := Node.OpenData(WriteAccess);
end;

function TLoopVolume.Close : Boolean;
begin
   Node.CloseData;
   Node.Close;
   Node := nil;
   Result := True;
end;

procedure TLoopVolume.SetNode(Node : TVNode);
begin
   self.Node := Node;
end;

function TLoopVolume.GetLength : Int64;
begin
   Result := Node.GetLen;
end;

function TLoopVolume.IsWritable : Boolean;
begin
   Result := Node.IsWritable;
end;


end.
