unit vv_part;

interface

uses classes, windows, sysutils;

implementation

uses vv, types, vv_win95, vv_map, debug, winioctl;

const FACILITY = 'vv_part';

function FindDosPart(Volume : TVirtualVolume; VolumeList : TList) : Boolean;
var
   PartNo   : Integer;
   ExtendedStart : ULong;

   function ProcessSector(Disk : TVirtualVolume; Sector : ULong; Depth : Integer) : Boolean;
   var
      Size     : DWORD;
      Buffer   : TDiskTable;
      i        : Integer;
      n        : TMapVolume;
      Start    : Int64;
      Len      : Int64;

   begin
      Result := True;
      Size := sizeof(Buffer);
      Log('Looking for table at sector ' + IntToStr(Sector), Facility);

      if not Disk.Read(Sector * 512, size, @Buffer) then
         begin
            Result := False;
            exit;
         end;
         Log('Signature is ' + IntToHex(Buffer.Signature, 4), Facility);
         if (Buffer.Signature = $AA55) then
         begin
            for i := 1 to 4 do
            begin
               Log('Processing partitions', Facility);
               if Buffer.Table[i].sys_ind <> 0 then
               begin
                  Log('Partition           ' + IntToStr(i) + ' /dev/hdx' + IntToStr(PartNo), Facility);
                  Log(' boot indicator      ' + IntToHex(Buffer.Table[i].boot_ind,2), Facility);
                  Log(' starting head       ' + IntToStr(Buffer.Table[i].head), Facility);
                  Log(' starting sector     ' + IntToStr(Buffer.Table[i].sector), Facility);
                  Log(' starting cylinder   ' + IntToStr(Buffer.Table[i].cyl), Facility);
                  Log(' partition type      ' + IntToHex(Buffer.Table[i].sys_ind,2), Facility);
                  case Buffer.Table[i].sys_ind of
                     PARTITION_EXTENDED         : Log('Extended... (dos)', Facility);
                     PARTITION_EXTENDED_LINUX   : Log('Extended... (linux)', Facility);
                     PARTITION_EXTENDED_WIN98   : Log('Extended... (win98)', Facility);
                     PARTITION_FAT_16           : Log('FAT 16', Facility);
                     PARTITION_HUGE             : Log('VFAT', Facility);
                     PARTITION_IFS              : Log('NTFS', Facility);
                     PARTITION_LINUX            : Log('Linux', Facility);
                     PARTITION_LINUX_SWAP       : Log('Linux Swap', Facility);
                  else
                     Log('Unknown', Facility);
                  end;
                  Log('end head            ' + IntToStr(Buffer.Table[i].end_head), Facility);
                  Log('end sector          ' + IntToStr(Buffer.Table[i].end_sector), Facility);
                  Log('end cylinder        ' + IntToStr(Buffer.Table[i].end_cyl), Facility);
                  Log(' starting sector     ' + IntToStr(Buffer.Table[i].start_sect), Facility);
                  Log(' number of sectors   ' + IntToStr(Buffer.Table[i].nr_sects), Facility);
                  Log('', Facility);

                  n := TMapVolume.Create;
                  n.AddName(Volume.GetName + '/' + IntToStr(PartNo));
                  Start := Buffer.Table[i].nr_sects;
                  Start := Start * 512;
                  Len := Buffer.Table[i].start_sect + Sector;
                  Len := Len * 512;
                  n.AddMap(0, Start, Volume, Len);
                  n.SetLayer(Facility, LEVEL_PROCESSED);
                  VolumeList.Add(n);

                  Inc(PartNo);
               end;
            end;
            for i := 1 to 4 do
            begin
               if Buffer.Table[i].sys_ind in EXTENDED_PARTITIONS then
               begin
                  Log('Processing extended partition....', Facility);
                  if Depth < 32 then
                  begin
                     if PartNo < 5 then
                     begin
                        PartNo := 5;
                     end;
                     if Depth = 1 then
                     begin
                        ExtendedStart := Sector;
                     end;
                     // There is a new partition table at offset start_sect...
                     if Depth >= 2 then
                     begin
                     if not ProcessSector(Disk, ExtendedStart + Buffer.Table[i].start_sect, Depth + 1) then
                        begin
                           Log('Did not seem to be a valid extended partition', Facility);
                        end;
                     end
                     else
                     begin
                        if not ProcessSector(Disk, Sector + Buffer.Table[i].start_sect, Depth + 1) then
                        begin
                           Log('Did not seem to be a valid extended partition', Facility);
                        end;
                     end;
                  end
                  else
                  begin
                     Log('More than 32 partitions detected! This is most likely a bug!', Facility);
                  end;
               end;
            end;
         end
         else
         begin
            Result := False;
         end;
      end;
begin
   PartNo := 1;
   ExtendedStart := 0;
   Result := ProcessSector(Volume, 0, 0);
end;

function BuildDosPartitionMap(VolumeList : TList; Options : TStringList) : Boolean;
var
   i : Integer;
   v : TVirtualVolume;

   Buffer   : String;
   Magic    : String;
   UUID     : String;
   Name     : String;
   Offset   : Integer;
begin
   Result := False;

   SetLength(Buffer, 512);

   for i := 0 to VolumeList.Count - 1 do
   begin
      v := TVirtualVolume(VolumeList[i]);

      if v.GetLayer(Facility) = LEVEL_NONE then
      begin
         // we don't want to process RAID member disks...
         if v.GetLayer('raid') = LEVEL_QUERIED then
         begin
            // we don't want to scan win32 disks? or maybe we do...?

            // we don't want to scan known fat filesystems
            if v.GetLayer('vfs_fat') = LEVEL_PROCESSED then
            begin
               // do nothing
            end
            else
            begin
               v.SetLayer(Facility, LEVEL_QUERIED);
               if FindDosPart(v, VolumeList) then
               begin
                  v.SetLayer(Facility, LEVEL_PROCESSED);
                  Result := True;
               end;
            end;

         end
         else if v.GetLayer('raid') = LEVEL_NONE then
         begin
            Log('Not scanning unchecked raid member ' + v.GetName, Facility, LOG_LOOP);
            Result := True;
         end
         else if v.GetLayer('raid') = LEVEL_PROCESSED then
         begin
            Log('Not scanning unused raid member ' + v.GetName, Facility, LOG_LOOP);
         end
         else if v.GetLayer('raid') = LEVEL_ACTIVATED then
         begin
            Log('Not scanning used raid member ' + v.GetName, Facility, LOG_LOOP);
         end
         else
         begin
            Log('Not scanning unknown status raid ' + v.GetName + ' at level ' + IntToStr(v.GetLayer('raid')), Facility, LOG_VERBOSE);
         end;
      end;
   end;
end;



initialization
   RegisterVolumeScan(Facility, BuildDosPartitionMap);


end.
